'''
Bealjas coordinator
'''
import logging, types

from bealjas.utils.setupLogging import setupConsoleLogger
from bealjas.network.serializeFunction import serialize

from coordinatorError import CoordinatorError
from slaveError import SlaveError

class BealjasTask:
    '''
    Holder object for task meta-data.
    
    There is one object per plot to visualize
    '''

    def __init__(self, myID):
        '''
        Constructor
        
        @param myID: unique ID assigned to this task. The ID must be the index of this task
        in the taskArguments array passed to the visualize, close and filter functions          
        '''    
        # Also index in coordinator.taskList array
        self.taskID = myID
        
        # Task argument sent to the visualize function. The length of this list is the
        # number of visualize, filter and close function calls for this task
        self.taskArguments = []
        
        # Process object set for currently running slaves
        self.slave = None
        
        # Pipe end for currently running slaves
        self.pipe = None
        
        # True if this task visualization is currently visible
        self.isVisible = False         
        
        # True if this tasks meets the visible filter criteria, False if it should be hidden,
        # and None if there is no visible filter
        self.visibleFilter = None    
        
        # Tuples with pending filter function and argument. This function is run when the task is started,
        # but after all visualization functions 
        self.pendingVisibleFilter = None # if set: (function, argument)
        #self.pendingCloseFilter = None   # if set: (function, argument)
        
        # True if this task is killed
        self.filterKill = False
        
        # Timestamp for when this task was last shown
        self.showTimestamp = None
        
    def getNFunctionCalls(self):
        '''
        Get the number of visualize, close and filter functions called for this task
        
        @return: number of function calls
        '''
        
        # A task object is added for each call
        return(len(self.taskArguments))
    
    def setSlave(self, slaveProc, slavePipe):        
        '''
        Assign a slave to this task
        
        @param slaveProc: slave process object
        @param slavePipe: pipe to slave
        '''                
        self.slave = slaveProc        
        self.pipe = slavePipe
         
    def removeSlave(self):
        '''
        Disconnect this task from a slave
        '''
        self.slave = None
        self.pipe = None
        
    def isRunning(self):
        '''
        @return: True if the task is mapped to a running slave process, False otherwise.
        '''
        if self.slave != None:
            return True
        else:
            return False
        
    def isShowable(self):
        '''
        Check if a task is hidden, that the visible filter is not false for this task, and that 
        the task is not closed.
        
        @return: True if the slave can be shown, False otherwise
        '''
        # Rule out illegal combinations
        if self.isVisible and self.visibleFilter == False:
            raise CoordinatorError('Visible window with visible filter = False')
        if self.isVisible and self.filterKill: 
            raise CoordinatorError('Visible window that is closed')
        
        if self.isVisible == True:
            # already shown
            return False
        elif self.visibleFilter == False:
            # should not be shown
            return False
        elif self.filterKill == True:
            # is closed
            return False
        else:
            return True
      
global loggerCnt  
loggerCnt = 0

class BealjasCoordinator:
    '''
    Bealjas coordinator superclass.
    '''
    
    def __init__(self, stateObject=None):
        '''
        Constructor
        
        @param stateObject: optional object sent as argument to many of the slave functions. Default is None
        '''      
        
        # Note! It may be better here to create a copy of the state object in case it is
        # modified during execution. If so the state on the slaves may differ.
        # However, there may be other state changed by the coordinator that is used by the
        # slaves, so it is probably best to have a programming policy stating that the 
        # coordinator should not change any state
        self.stateObject = stateObject
        
        # Initialized slaveLogger[SetupFunction,Level,OptionalArguments] variables
        global loggerCnt
        loggerCnt += 1
        self.logger, self.logFilename = setupConsoleLogger('coordinator-%d' % (loggerCnt), logging.INFO)
        self.logger.info("Setting slave logger to default (console with log level warning")
        self.slaveLoggerSetup(setupConsoleLogger, logging.INFO, None)
    
    def startSlaves(self, visualizeFunction, tasks):
        """
        Start slaves processes and do the first visualization.
        
        @param vizualizeFunction: visualization function that is run on all slaves. The function
        may be None.
        @param tasks: list of tasks where each element is used as input to one slave.                    
        @return: None
        """
        raise CoordinatorError('Must be implemented by subclass')
 
    def killSlaves(self):
        """
        Kill all slave nodes
        """
        raise CoordinatorError('Must be implemented by subclass')
    
    def visualize(self, vizualizeFunction, tasks):
        """
        Execute a new visualization function using the state on the slave nodes
        
        @param visualizeFunction: user provided visualize function called to do perform the slave task
        @param tasks: list of tasks where each element is used as input to one slave.        
        @return: None
        """
        raise CoordinatorError('Must be implemented by subclass')
        
    def showNext(self, nToShow=None):
        """
        Show next set of views
        
        @param nToShow: optional number of new tasks to show,
        """
        raise CoordinatorError('Must be implemented by subclass')
    
    def showPrevious(self, nToShow=None):
        """
        Show the previous set of views
        
        @param nToShow: optional number of new tasks to show
        """        
        raise CoordinatorError('Must be implemented by subclass')
    
    def showFirst(self):
        """
        Show the first set of views
        """
        self.showIndex(0)
        
    def showIndex(self, startIndex):
        """
        Show views starting from an startIndex
        
        @param startIndex: start startIndex
        """
        raise CoordinatorError('Must be implemented by subclass')
    
    def showRandom(self, nToShow=None):
        """
        Show a random set of views
        
        @param nToShow: optional number of new tasks to show
        """
        raise CoordinatorError('Must be implemented by subclass')
    
    def showIndexes(self, taskIndexes):
        """
        Show specific views
        
        taskIndexes: index of the visualizations to showb in the task list that was passed as
        argument to the visualization function.
        """  
        raise CoordinatorError('Must be implemented by subclass')
    
    def startNext(self, nToStart):
        '''
        Start new tasks that are likely to be shown in the near future.
        
        This function selects the next showable tasks to start. It will also kill the 
        least recently shown tasks.
        
        @param nToStart: number of new tasks to start
        '''
        raise CoordinatorError('Must be implemented by subclass')
    
    def startPrev(self, nToStart):
        '''
        Start new tasks that are likely to be shown in the near future.
        
        This function selects the previous showable tasks to start. It will also kill the 
        least recently shown tasks.
        
        @param nToStart: number of new tasks to start
        '''
        raise CoordinatorError('Must be implemented by subclass')
    
    def filterVisible(self, filterFunction, filterArgs=None):
        """
        Determine views to show using a filter function
        
        @param filterFunction: filter function to run
        @param filterArgs: arguments passed to filter function
        """
        raise CoordinatorError('Must be implemented by subclass')
    
    def pendingFilterVisible(self, filterFunction, filterArgs=None):        
        """
        Determine views to show using a filter function. The filter is pending, meaning that it is only run
        when the task is running.
        
        @param filterFunction: filter function to run
        @param filterArgs: arguments passed to filter function
        """
        raise CoordinatorError('Must be implemented by subclass')
    
    def filterVisibleOff(self):
        """
        Turn off visible filter
        """
        raise CoordinatorError('Must be implemented by subclass')
        
    def filterKill(self, filterFunction, filterArgs=None):
        """
        Use a filter function to kill some of the slaves processes
        
        @param filterFunction: function to run. The return value determines wheter the slave should be
        kept (True) or killed (False)
        @param filterArgs: optional argument sent to the filter function
        """
        raise CoordinatorError('Must be implemented by subclass')
    
    def gather(self, gatherFunction, gatherArgs=None):
        '''
        Gather values from all tasks
        
        @param gatherFunction: function run on each slave that returns the values to gather.
        @param gatherArgs: arguments passed to the gather function
        @return: list of values returned by the close functions on each slave
        '''
        raise CoordinatorError('Must be implemented by subclass') 
    
    def readLogFiles(self):
        '''
        Gather the log file content from all tasks
        
        @return list where each element is a list of log file lines read from a task
        '''
        raise CoordinatorError('Must be implemented by subclass') 
    
    def slaveLoggerSetup(self, setupFunction, level, optionalArguments=None):
        '''
        Specify function to run in order to initialize logging on the slaves. Default is 
        setupLoggerConsole()
        
        @param setupFunction: function that setups logging. It should take three arguments:
        process-ID, logging level, and optional arguments
        @param level: logging level to use
        @param optionalArguments: disctionary with optional argumetns to send to setup
        function.
        '''
        if not isinstance(setupFunction, types.FunctionType):
            raise CoordinatorError('Setup function error: is not a function')
        
        # Level is not checked since it may be application specific
        
        # Optional arguments not checked since they are setup function specific
        
        self.slaveLoggerSetupFunction = setupFunction
        self.slaveLoggerLevel = level
        self.slaveLoggerOptionalArguments = optionalArguments
        
    
    #
    # Helper functions to send a message on a slave's pipe
    #
    def sendCloseMsg(self, pipe):
        '''
        Send a window close message to a task
        
        @param pipe: pipe connected to task
        @return: None
        '''
        
        pipe.send("close")
        #if pipe.recv() == 'killed':
        #    raise SlaveError('Slave is killed')
        
    def sendKillMsg(self, pipe):
        '''
        Send a kill message to a task
        
        @param pipe: pipe connected to task
        @return: None
        '''
        pipe.send("kill")
        if pipe.recv() == 'killed':
            self.logger.info('Slave was killed before the kill message was sent')
        
    def sendShowMsg(self, pipe):
        '''
        Send a show window message to a task
        
        @param pipe: pipe connected to task
        @return: None
        '''
        pipe.send("show")
        #if pipe.recv() == 'killed':
        #    raise SlaveError('Slave is killed')
        
    def sendHideMsg(self, pipe):
        '''
        Send a hide window message to a task
        
        @param pipe: pipe connected to task
        @return: None
        '''
        pipe.send("hide")
        #if pipe.recv() == 'killed':
        #    raise SlaveError('Slave is killed')
        
    def sendVisualizeMsg(self, pipe, visualizeFunction, taskArgument):
        '''
        Send a visualize function to a task
        
        @param pipe: pipe connected to task
        @param visualizeFunction: visualize function to send
        @param taskArgument: task specific argument sent as parameter to the visualizeFunction
        @return: None
        '''
        pipe.send("visualize")
        pipe.send((serialize(visualizeFunction), taskArgument))
        #if pipe.recv() == 'killed':
        #    raise SlaveError('Slave is killed')
        
    def sendGatherMsg(self, pipe, gatherFunction, gatherArg=None):
        '''
        Send a gather message and receive result. The return values must be received 
        separately.
        
        @param pipe: pipe connected to task
        @param gatherFunction: function used to gather values
        @param gatherArg: optional argument sent to the gather function
        @return: None
        '''
        pipe.send("gather")
        pipe.send((serialize(gatherFunction), gatherArg))
        #rv = pipe.recv()
        #if rv == 'killed':
        #    raise SlaveError('Slave is killed')
        
    def sendFilterMsg(self, pipe, filterFunction, filterArg=None):
        '''
        Send a filter function to a task
        
        @param pipe: pipe connected to task
        @param filterFunction: filter function to send
        @param taskArgument: task specific argument sent as parameter to the filterFunction
        @return: None
        '''
        pipe.send("filter")
        pipe.send((serialize(filterFunction), filterArg))
        #if pipe.recv() == 'killed':
        #    raise SlaveError('Slave is killed')
    
    def sendGetPosSizeMsg(self, pipe):
        '''
        Send get window position and size message and return result
        
        @param pipe: pipe connected to task
        @return: position and size 5-tuple
        '''
        pipe.send("getPosSize")
        #if pipe.recv() == 'killed':
        #    raise SlaveError('Slave is killed')
        #else:
        return pipe.recv()
    
    def sendSetPosSizeMsg(self, pipe, posSize5Tuple):
        '''
        Send a set window positino and size message to a task
        
        @param pipe: pipe connected to task
        @param posSize5Tuple: position and size 5-tuple to send
        @return: None        
        '''
        pipe.send("setPosSize")
        pipe.send(posSize5Tuple)
        #if pipe.recv() == 'killed':
        #    raise SlaveError('Slave is killed')
        
    def sendReadLogFilesMsg(self, pipe):
        '''
        Send read log file lines from task. The logfiles must be received from the pipe
        separately.        
        
        @param pipe: pipe connected to task
        @return: None
        
        '''
        pipe.send("readLogfileLines")
        #if pipe.recv() == 'killed':
        #    raise SlaveError('Slave is killed')        