import random, types

import bealjas.network.protSocket as bs

# Superclass
from bealjasCoordinator import BealjasCoordinator
from coordinatorError import CoordinatorError

from bealjas.network.serializeFunction import serialize


class BealjasTileTask:
    '''
    Simplified holder object for task meta-data.
    
    There is one object per plot to visualize. Each tile has in addition an BealjasTask with the
    full meta-data for their tasks.
    '''

    def __init__(self, myID, tileID):
        '''
        Constructor
        
        @param myID: unique ID assigned to this task. The ID must be the index of this task
        in the taskArguments array passed to the visualize, close and filter functions  
        @param tileID: tile ID to which this task is mapped to.
        '''    
        # Also index in coordinator taskList array
        self.taskID = myID
        
        # Task argument sent to the visualize function. The length of this list is the
        # number of visualize, filter and close function calls for this task
        self.taskArguments = []
        
        # Tile ID 
        self.tileID = tileID

class MasterCoordinator(BealjasCoordinator):
    '''
    Superclass for window based plotters
    '''
    
    #
    # Public interface
    #
    
    def __init__(self, nTiles, plotsPerTile, slavesPerTile, 
                 tileIDs, nwThread,
                 requestQueue, rvQueue,
                 stateObject=None):
        '''
        Constructor
        
        @param nTiles: number of tiles running tileCoordinators
        @param plotsPerTile: number of windows to open per tile
        @param slavesPerTile: number of processes to start at once. If None, the default value of 
        nPlots is used. If -1 one slave process is started per task. Otherwise, the specified number 
        of slave processes are started. 
        @param tileIDs: client IDs
        @param nwThread: network server thread
        @param requestQueue: queue used to send requests to tile coordinators
        @param rvQueue: queue used to return requests from tile coordinators                
        @param stateObject: optional object sent as argument to many of the slave functions. Default is None   
        '''        
        BealjasCoordinator.__init__(self, stateObject)
        
        self.logger.info('Coordinating %d tiles with %d windows and %d slaves per tile' % (nTiles, plotsPerTile, slavesPerTile))
        
        if plotsPerTile > slavesPerTile:
            raise CoordinatorError('nPlots > nSlaves')
        
        # Number of windows to open and number of slave processes to start
        self.nTiles = nTiles
        self.plotsPerTile = plotsPerTile
        self.slavesPerTile = slavesPerTile
        self.tileIDs = tileIDs
        self.nwThread = nwThread
        
        # Queues used to send and receive commands to/from the tile coordinators
        self.requestQueue = requestQueue # master -> tiles
        self.rvQueue = rvQueue           # master <- tiles
                            
        # List of executed visualization functions
        #
        # This list is used to playback multi-step visualizations when there are more tasks
        # than slaves.
        self.functionList = []
        
        # Lists of tasks to visualize. The visualization is split into multiple tasks that are visualized 
        # independently. Both lists have references to the same BealjasTask objects. 
        # The lists is initialized in startSlaves
        # 
        # The first list contains all tasks objects in the order they are provided to the visualize function
        self.taskList = []
        # The second list has the task objects ordered according to tile IDs
        self.tasksPerTile = []
    
        # To ensure experiment repeatability
        random.seed(31278)       
                  
    def startSlaves(self, visualizeFunction, taskArgs):
        """
        Start slaves processes and show first set of windows
        
        @param vizualizeFunction: visualization function that is run on all slaves. The function
        may be None.
        @param taskArgs: list of task arguments where each element is used as input to one slave.                    
        @return: None
        """
        
        self.logger.info("Starting slaves")
        
        if visualizeFunction != None and not isinstance(visualizeFunction, types.FunctionType):
            raise CoordinatorError('Visualize function error: not a function')            
        
        nSlaves = self.nTiles * self.slavesPerTile
        if nSlaves != -1 and len(taskArgs) < nSlaves:
            raise CoordinatorError('Number of tasks is less than number of slaves to start (%d < %d)' % (len(taskArgs), nSlaves))                                                              
        
        #
        # Distribute tasks equally among tiles
        #
        self.tasksPerTile = {} # tile ID -> [task IDs]
        taskArgsPerTile   = {} # tile ID -> [taskArgs]
        nTasks = len(taskArgs)
        nTasksPerTile = len(taskArgs) / self.nTiles
        # First M tasks has one more task than others
        tai = 0
        for i in range(nTasks % self.nTiles):
            l = []
            l2 = []
            for j in range(nTasksPerTile + 1):
                # Add task object to the global and per tile tasks lists
                b = BealjasTileTask(tai + j, self.tileIDs[i])
                l.append(b)
                l2.append(taskArgs[tai + j])
                self.taskList.append(b)
            self.tasksPerTile[ self.tileIDs[i] ] = l
            taskArgsPerTile[self.tileIDs[i]] = l2
            self.logger.debug('Tile %s has %d tasks' % (self.tileIDs[i], len(l)))          
            tai = tai + nTasksPerTile + 1
        # Remaining tasks have nTasksPerTile
        for i in range(nTasks % self.nTiles, self.nTiles):
            l = []
            l2 = []
            for j in range(nTasksPerTile):
                # Add task object to the global and per tile tasks lists
                b = BealjasTileTask(tai + j, self.tileIDs[i])
                l.append(b)
                l2.append(taskArgs[tai + j])
                self.taskList.append(b)
            self.tasksPerTile[ self.tileIDs[i] ] = l
            taskArgsPerTile[self.tileIDs[i]] = l2
            self.logger.debug('Tile %s has %d tasks' % (self.tileIDs[i], len(l)))
            tai = tai + nTasksPerTile            
                      
        # Send startSlaves message to all tiles
        self.logger.debug('Send startSlaves message to each slave')
        for tid in self.tileIDs:
            self.sendRequestMsg([tid], ('startSlaves', serialize(visualizeFunction), taskArgsPerTile[tid]))
        # The visualize message is not sent since startSlaves will automatically call the 
        # visualization function
 
 
    def killTileCoordinators(self):
        '''
        Kill all tile coordinators
        '''
        
        # Kill tile coordinators
        self.logger.info('Kill tile coordinators')
        self.sendRequestMsg("all", ('killTileCoordinators',))
        
        # Kill all network layer threads
        self.logger.info('Kill network clients')
        bs.killServerAndClients(self.requestQueue)
    
    def killSlaves(self):
        '''
        Kill all slave processes
        '''    
        self.logger.info('Kill all slaves')
        
        # Kill all currently running slaves
        self.sendRequestMsg("all", ('killSlaves',))        
        self.taskList = []
        self.tasksPerTile = []
        
        
    def visualize(self, visualizeFunction, taskArgs):
        '''
        Distribute visualization tasks and run a visualization function on all slaves.            
        
        @param visualizeFunction: user provided visualize function called to do perform the slave task
        @param args: list of args where each element is used as input to one slave.        
        @return: None
        '''
        
        if not isinstance(visualizeFunction, types.FunctionType):
            raise CoordinatorError('Invalid visualize function: not a function')        
        if self.taskList == None:
            raise CoordinatorError('Slaves have not been started')
        if taskArgs == None:
            raise CoordinatorError('Invalid tasks argument: None')
        # The number of tasks cannot be changed
        if len(taskArgs) != len(self.taskList):
            raise CoordinatorError('Invalid number of entries in taskArgument list (got %d expected %d)' % (len(taskArgs), len(self.taskList)))                    
                                
        # Update list of executed functions and distribute new tasks to visualize
        self.functionList.append(visualizeFunction)      
        for i in range(len(taskArgs)):
            self.taskList[i].taskArguments.append(taskArgs[i])
            assert(len(self.functionList) == len(self.taskList[i].taskArguments))

        argsPerTile = self.args2tiles(taskArgs)
        
        # Send visualization function to all tiles          
        self.logger.debug("Send visualize function to all tiles")
        for tid in self.tileIDs:
            self.sendRequestMsg([tid], ('visualize', serialize(visualizeFunction), argsPerTile[tid]))
             
            
    def filterVisible(self, filterFunction, filterArgs=None):
        """
        Determine views to show using a filter function. This function will run the filter function on 
        all running tasks even if they are not currently running.
        
        @param filterFunction: filter function to run
        @param filterArgs: optional arguments passed to filter function
        """
        
        self.logger.info("Filter visible")
        
        if not isinstance(filterFunction, types.FunctionType):
            raise CoordinatorError('Invalid visualize function: not a function')        
        if self.taskList == None:
            raise CoordinatorError('Slaves have not been started')
        if filterArgs != None and len(filterArgs) != len(self.taskList):
            raise CoordinatorError('Invalid number of entries in taskArgument list (got %d expected %d)' % (len(filterArgs), len(self.taskList)))        
            
        argsPerTile = None    
        if filterArgs != None:
            argsPerTile = self.args2tiles(filterArgs)
        
        # Send filter to function to all tileCoordinators
        for tid in self.tileIDs:
            if filterArgs != None:
                self.sendRequestMsg([tid], ('filterVisible', serialize(filterFunction), argsPerTile[tid]))
            else:
                self.sendRequestMsg([tid], ('filterVisible', serialize(filterFunction), None))
        
        # The filterVisible does not return any return value
                
    def pendingFilterVisible(self, filterFunction, filterArgs=None):
        """
        Determine views to show using a filter function. This function will run the filter function on 
        all running tasks even if they are not currently running.
        
        @param filterFunction: filter function to run
        @param filterArgs: optinal arguments passed to filter function
        """
        
        self.logger.info("Pending filter visible")
        
        if not isinstance(filterFunction, types.FunctionType):
            raise CoordinatorError('Invalid visualize function: not a function')        
        if self.taskList == None:
            raise CoordinatorError('Slaves have not been started')
        if filterArgs != None and len(filterArgs) != len(self.taskList):
            raise CoordinatorError('Invalid number of entries in taskArgument list (got %d expected %d)' % (len(filterArgs), len(self.taskList)))
                
        argsPerTile = None    
        if filterArgs != None:
            argsPerTile = self.args2tiles(filterArgs)
        
        # Send filter to function to all tileCoordinators
        for tid in self.tileIDs:
            if argsPerTile != None:
                self.sendRequestMsg([tid], ('pendingFilterVisible', serialize(filterFunction), argsPerTile[tid]))
            else:
                self.sendRequestMsg([tid], ('pendingFilterVisible', serialize(filterFunction), None))
                    
    def filterVisibleOff(self):
        """
        Turn off visible filter
        """
        self.logger.info("Turn off visible filter")     
        
        # Send message to all tile coordinators
        self.sendRequestMsg('all', ('filterVisibleOff',))
        
    def filterClose(self, filterFunction, filterArgs=None):
        """
        Use a filter function to close some of the tasks
        
        @param filterFunction: function to run. The return value determines wheter the slave should be
        kept (True) or killed (False)
        @param filterArgs: optional argument sent to the filter function
        """
        self.logger.info("Close all windows")
        
        if not isinstance(filterFunction, types.FunctionType):
            raise CoordinatorError('Invalid visualize function: not a function')        
        if self.taskList == None:
            raise CoordinatorError('Slaves have not been started')
        if filterArgs != None and len(filterArgs) != len(self.taskList):
            raise CoordinatorError('Invalid number of entries in taskArgument list (got %d expected %d)' % (len(filterArgs), len(self.taskList)))
        
        argsPerTile = None    
        if filterArgs != None:
            argsPerTile = self.args2tiles(filterArgs)
        
        # Send filter function to all tileCoordinators
        for tid in self.tileIDs:
            if filterArgs != None:
                self.sendRequestMsg([tid], ('filterClose', serialize(filterFunction), argsPerTile[tid]))
            else:
                self.sendRequestMsg([tid], ('filterClose', serialize(filterFunction), None))
                
    def gather(self, gatherFunction, gatherArgs=None):
        '''
        Gather values from all tasks
        
        @param gatherFunction: function run on each slave that returns the values to gather.
        @param gatherArgs: arguments passed to the gather function
        @return: list of values returned by the close functions on each slave
        '''
        
        self.logger.info("Gather return values")
        
        if not isinstance(gatherFunction, types.FunctionType):
            raise CoordinatorError('First argument is not a function')
        if self.taskList == None:
            raise CoordinatorError('Slaves have not been started')
       
        argsPerTile = None    
        if gatherArgs != None:
            argsPerTile = self.args2tiles(gatherArgs)
        
        # Return values
        tile2rv = {} # Tile ID -> rvs
        
        # Send filter to function to all tileCoordinators
        for tid in self.tileIDs:
            rv = None
            if gatherArgs != None:
                rv = self.sendRequestMsg([tid], ('gather', serialize(gatherFunction), argsPerTile[tid]))
            else:
                rv = self.sendRequestMsg([tid], ('gather', serialize(gatherFunction), None))
            tile2rv.update(rv)
            
        # Sort and add return values to a single list
        rvs = []
        for tid in self.tileIDs:
            rvs = rvs + tile2rv[tid]
            
        # Return the list
        return rvs
    
    def showNext(self, nToShow=None):
        """
        Hide all currently visible task and show the show the next set of tasks. The tasks are selected based 
        on their task arguments array index (taskID number). For example if tasks 2, 3 and 13 are currently 
        visible, then these will be replaces with 4, 5 and 14.
        
        @param nToShow: optional number of new tasks to show, if not specified self.nMaxVisible is used instead
        """
                
        self.logger.debug("Show next set of slaves")        
        
        self.sendRequestMsg("all", ('showNext', nToShow))        
        
    def showPrevious(self, nToShow=None):
        """
        Hide all currently visible task and show the show the previous set of tasks. The tasks are selected based 
        on their task arguments array index (taskID number).  For example if 1, 2, and 13 (of 30) tasks are visible, 
        the previous task shown are 0, 12, and 29
        
        @param nToShow: optional number of new tasks to show, if not specified self.nMaxVisible is used instead
        """
        
        self.logger.debug("Show previous set of tasks")        
        self.sendRequestMsg("all", ('showPrevious', nToShow))
        
    def showRandom(self, nToShow=None):
        """
        Show a random set of views
        
        @param nToShow: optional number of new tasks to show,
        """
        
        self.logger.debug("Show random tasks")        
        self.sendRequestMsg("all", ('showRnadom', nToShow))
    
    def showIndex(self, index, nToShow=None):
        """
        Show views starting from an index. 
        For example if 1, 2, and 13 are visible and the index is 1, this will show 1, 2, and 3 
        
        @param index: start index
        @param nToShow: optional number of new tasks to show, if not specified self.nMaxVisible is used instead
        """
        
        if index > len(self.taskList):
            raise CoordinatorError("Invalid index: %d" % (index))
        
        self.logger.debug("Show %d tasks starting from index %d" % (nToShow, index))
        
        # Find tile with given index
        startTile = self.taskList[index].tileID
        # Find the offset within that tiles task list
        offset = 0
        for task in self.tasksPerTile[startTile]:
            if task.taskID == index:
                break
            else:
                offset += 1
        if offset == len(self.tasksPerTile[index]):
            raise 'Could not find task with index %d on tile %d' % (index, startTile)
                
        # Send showIndex message to all tiles
        if offset <= (len(self.taskList) / self.nTiles):
            self.sendRequestMsg("all", ('showIndex', offset, nToShow))        
        # Special case: if offset is the additional 'modolu' task added to each of the first M tiles
        else:
            # First M tasks
            tids = []
            for i in range(len(self.taskList) % self.nTiles):
                tids.append(self.tileIDs[i])
            self.sendRequestMsg(tids, ('showIndex', offset, nToShow))
            
            # For the remaining the offset is one less
            tids = []
            for i in range(len(self.taskList) % self.nTiles, self.nTiles):
                tids.append(self.tileIDs[i])
            self.sendRequestMsg(tids, ('showIndex', offset-1, nToShow))                                
        
    def showIndexes(self, taskIndexes):
        """
        Show specific views
        
        taskIndexes: index of the visualizations to show in the task list that was passed as
        argument to the visualization function.
        """  
        
        if (len(taskIndexes) > self.nMaxVisible):
            raise CoordinatorError("Too many task indexes (got; %d, max is: %d)" % (len(taskIndexes), self.nMaxVisible))
        
        self.logger.info("Set the following tasks to visible: %s" % (taskIndexes))
              
        showPerTile = {}                
        # Map task indexes to tiles
        for ti in taskIndexes:
            if ti > len(self.taskList):
                raise CoordinatorError('Invalid task ID: %d' % (ti))
            tid = self.taskList[ti].tileID
            if tid not in showPerTile:
                showPerTile[tid] = []
            showPerTile[tid].append(ti)
        
        for tid in showPerTile:
            self.sendRequestMsg([tid], ('showIndexes', showPerTile[tid]))
                             
    def closeWindows(self):
        '''
        Close all open windows on slaves. Note that close differs from hide in that window resources
        are freed, and it is possible for the slaves to return values that are gathered by the coordinator.
        
        @return: None
        '''

        self.logger.debug("Close all windows")        

        # Clear list of executed functions and arguments
        self.functionList = []         
        for i in range(len(self.taskList)):
            self.taskList[i].tasks = []

        self.sendRequestMsg("all", ('closeWindows',))
            
    def resetVisible(self):
        """
        Resend hide and show messages. This function is useful if there is a glitch in the setVisible
        functions which sometimes may happen (especially at startup)
        """
        self.logger.info("Resending show and hide messages")
        self.sendRequestMsg("all", ('resetVisible',))

    def showMore(self):
        """
        Show additional visualizations. This function only has an effect if currently visible is less 
        than nMaxVisible.
        """
        self.logger.debug("Show more")
        self.sendRequestMsg("all", ('showMore',))
        
        
    #
    # Helper functions
    #        
        
    def args2tiles(self, taskArgs):
        '''
        Distribute task arguments among tiles
        
        @param tasks: 1D list of task arguments
        @return tasksPerTile dict, that maps tile IDs to a list of task arguments for the tile's tasks
        '''
        
        assert(len(taskArgs) == len(self.taskList))
        
        perTile = {}
        # One list per tile
        for tid in self.tileIDs:
            perTile[tid] = []
        
        for i in range(len(taskArgs)):
            perTile[ self.taskList[i].tileID ].append(taskArgs[i])
            
        return perTile
    
    def sendRequestMsg(self, clientIDs, tupleData):
        '''
        Send a request tuple to the specified tiles
                
        @param clientIDs: either a list of client IDs, or "all" if the message should be sent
        to all.
        @param tupleData: tuple with data to send
        @param rvs: received return values
        '''
        self.requestQueue.put((clientIDs, tupleData))
        return self.rvQueue.get()
    
    def consoleDriver(self):
        '''
        Simple console driver that reads window navigation input from the console
        '''
        while 1:
            inputStr = raw_input("(N)ext, (P)rev, (R)andom, (F)irst, (I)ndex(Number), (5) Refresh, (E)xit: ")
            inputStr.lower()
            
            if inputStr == 'e' or inputStr == 'exit':
                break
            elif inputStr == '5':
                self.resetVisible()
            elif inputStr == 'n' or inputStr == 'next':
                self.showNext()
            elif inputStr == 'p' or inputStr == 'prev':
                self.showPrevious()
            elif inputStr == 'r' or inputStr == 'random':
                self.showRandom()
            elif inputStr == 'f' or inputStr == 'first':
                self.showFirst()
            elif inputStr.startswith('index'):
                indexNum = int(inputStr[1:])
                self.showIndex(indexNum)
            elif inputStr.startswith('i'):
                indexNum = int(inputStr[5:])
                self.showIndex(indexNum)
                
def startMasterCoordinator(nTiles, plotsPerTile, slavesPerTile, hostname, listenPort, stateObject=None):
    '''
    Setup network layer and start master coordinator
    
    @param nTiles: number of tiles running tileCoordinators
    @param plotsPerTile: number of windows to open per tile
    @param slavesPerTile: number of processes to start at once. If None, the default value of 
    nPlots is used. If -1 one slave process is started per task. Otherwise, the specified number 
    of slave processes are started. 
    @param hostname: network interface to connect to
    @param listePort: listen port to open
    @param stateObject: optional state object
    '''
    
    print '[startMasterCoordinator] start network layer server'
    nwThread, requestQueue, rvQueue = bs.runServer(nTiles, hostname, listenPort)    
        
    # The network layer has put a message with all client ids in the rv queue
    print '[startMasterCoordinator] wait for client IDs'
    tileIDs = rvQueue.get()
    if len(tileIDs) != nTiles:
        raise CoordinatorError('Invalid number of tile IDs received: %d (expected %d)' % (len(tileIDs), nTiles))
        
    # Send parameters to all clients
    print '[startMasterCoordinator] send parameters to all clients'
    requestQueue.put(("all", (plotsPerTile, slavesPerTile)))
    # Receive and discard return values
    rvQueue.get()
    
    # Setup master
    print '[startMasterCoordinator] setup and return master coordinator object'
    coordinator = MasterCoordinator(nTiles, plotsPerTile, slavesPerTile, 
                                    tileIDs, nwThread,
                                    requestQueue, rvQueue,
                                    stateObject)
    
    return coordinator
    
    
            