import bealjas.network.protSocket as bs

from bealjas.network.serializeFunction import deserialize

# Superclass
from windowCoordinator import WindowCoordinator
from coordinatorError import CoordinatorError

class TileCoordinator(WindowCoordinator):
    '''
    Superclass for window based plotters
    '''
    
    #
    # Public interface
    #
    
    def __init__(self, 
                 nPlots, nSlaves,
                 requestQueue, rvQueue,
                 createWindowFunc, guiLoopFunc, closeFunc,
                 setVisibleFunc, getPosSizeFunc, setPosSizeFunc,
                 stateObject=None):
        '''
        Constructor
               
        @param nPlots: number of windows to open
        @param nSlaves: number of processes to start at once. If None, the default value of 
        nPlots is used. If -1 one slave process is started per task. Otherwise, the specified number 
        of slave processes are started. 
        @param requestQueue: queue used to send requests to tile coordinators
        @param rvQueue: queue used to return requests from tile coordinators 
        @param createWindowFunction: function called in the slave constructur to create window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param guiLoopFunction: function called to execute the GUI loop. (functions for various backends may
        be found in modules bealjas2X modules the bealjas.utils package)
        @param closeFunc: function called to close the window and free any resource allocated
        by the visualization. (functions for various backends may be found in modules bealjas2X modules the 
        bealjas.utils package)
        @param setVisibleFunc: function called to either show or hide window (functions for various
        backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param getPosSizeFunc: function called to get the position and size of a window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param setPosSizeFunc: function called to set the position and size of a window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)     
        @param stateObject: optional object sent as argument to many of the slave functions. Default is None      
        '''        
        
        # nPlots and nSlaves are bot set to 1 since these are later received from the master
        WindowCoordinator.__init__(self,                                   
                                   nPlots, nSlaves,
                                   createWindowFunc, guiLoopFunc, closeFunc,
                                   setVisibleFunc, getPosSizeFunc, setPosSizeFunc,
                                   stateObject)
        
        # Queues used to send and receive commands to/from the tile coordinators
        self.requestQueue = requestQueue # master -> tiles
        self.rvQueue = rvQueue           # master <- tiles          
    
    def mainLoop(self):
        '''
        Main loop to process requests received from the master
        '''
        
        funcNum = 0
        
        self.logger.info("Tile coordinator entering main loop")
        while 1:
            # Recv cmd from master 
            reqTuple = self.requestQueue.get()
            #print reqTuple
            cmd = reqTuple[0]            
            
            # Always send return value, even if it just a None value            
            rv = None
            
            # Unique function names
            funcName = 'func%d' % (funcNum)
            funcNum += 1
            
            # Process cmd, and if needed, send a return value to the master
            if cmd == 'startSlaves':
                vizFunc = deserialize(reqTuple[1], funcName)
                taskArgs = reqTuple[2]
                self.startSlaves(vizFunc, taskArgs)
            elif cmd == 'killSlaves':
                self.killSlaves()
            elif cmd == 'visualize':
                vizFunc = deserialize(reqTuple[1], funcName)
                taskArgs = reqTuple[2]
                self.visualize(vizFunc, taskArgs)
            elif cmd == 'filterVisible':
                vizFunc = deserialize(reqTuple[1], funcName)
                taskArgs = reqTuple[2]
                self.filterVisible(vizFunc, taskArgs)
            elif cmd == 'pendingFilterVisible':
                vizFunc = deserialize(reqTuple[1], funcName)
                taskArgs = reqTuple[2]
                self.pendingFilterVisible(vizFunc, taskArgs)
            elif cmd == 'filterVisibleOff':
                self.filterVisibleOff()
            elif cmd == 'filterClose':
                vizFunc = deserialize(reqTuple[1], funcName)
                taskArgs = reqTuple[2]
                self.filterClose(vizFunc, taskArgs)
            elif cmd == 'gather':
                vizFunc = deserialize(reqTuple[1], funcName)
                taskArgs = reqTuple[2]
                rv = self.gather(vizFunc, taskArgs)                
            elif cmd == 'showNext':
                nToShow = reqTuple[1]
                self.showNext(nToShow)
            elif cmd == 'showPrevious':
                nToShow = reqTuple[1]
                self.showPrevious(nToShow)
            elif cmd == 'showRandom':
                nToShow = reqTuple[1]
                self.showRandom(nToShow)
            elif cmd == 'showIndex':                
                taskIndex = reqTuple[1]
                nToShow = reqTuple[2]                
                self.showIndex(taskIndex, nToShow)
            elif cmd == 'showIndexes':
                taskIndexes = reqTuple[1]
                self.showIndexes(taskIndexes)
            elif cmd == 'closeWindows':
                self.closeWindows()
            elif cmd == 'resetVisible':
                self.resetVisible()
            elif cmd == 'showMore':
                self.showMore()
            elif cmd == 'killTileCoordinators':                         
                self.rvQueue.put('ack')
                self.logger.info('Tile coordinator exiting')
                break
            else:
                raise CoordinatorError('Invalid cmd: %s' % (cmd)) 
                    
            # Send return value
            self.rvQueue.put(rv)            
            
        
            
def startTileCoordinator(clientID, host, port,
                         createWindowFunc, guiLoopFunc, closeFunc,
                         setVisibleFunc, getPosSizeFunc, setPosSizeFunc,
                         stateObject=None):
    '''
    Connect to master coordinator and initialize a TileCoordinator object and call its main loop.
            
    @param clientID: unique ID assigned to this client
    @param masterHostname: host running master
    @param masterPort: port master listens on
    @param createWindowFunction: function called in the slave constructur to create window (functions 
    for various backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param guiLoopFunction: function called to execute the GUI loop. (functions for various backends may
    be found in modules bealjas2X modules the bealjas.utils package)
    @param closeFunc: function called to close the window and free any resource allocated
    by the visualization. (functions for various backends may be found in modules bealjas2X modules the 
    bealjas.utils package)
    @param setVisibleFunc: function called to either show or hide window (functions for various
    backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param getPosSizeFunc: function called to get the position and size of a window (functions 
    for various backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param setPosSizeFunc: function called to set the position and size of a window (functions 
    for various backends may be found in modules bealjas2X modules the bealjas.utils package)     
    @param stateObject: optional object sent as argument to many of the slave functions. Default is None 
    @return None
    '''
        
    # Connect socket to master and setup request and return value queues
    print '[startTileCoordinator] setup network client'
    nwThread, requestQueue, rvQueue = bs.runClient(clientID, host, port)
     
    # First message sent by master is a tuple with the number of slaves and windows to open
    print '[startTileCoordinator] get parameters'
    tupleData = requestQueue.get()    
    nPlots, nSlaves = tupleData
    
    print '[startTileCoordinator] create and run tile coordinator'
    tc = TileCoordinator(nPlots, nSlaves,
                         requestQueue, rvQueue,
                         createWindowFunc, guiLoopFunc, closeFunc,
                         setVisibleFunc, getPosSizeFunc, setPosSizeFunc,
                         stateObject)
    
    # Loop until told to exit
    tc.mainLoop()
    
    # Join with previously started network thread
    print '[startTileCoordinator] join with network thread'
    nwThread.join()