import random, types, time

from multiprocessing import Process, Pipe
# Superclass
from bealjasCoordinator import BealjasCoordinator, BealjasTask

from coordinatorError import CoordinatorError
from slaveError import SlaveError

# Entry point for slave process
from bealjas.plotter.windowPlotter2 import plotterEntry

class WindowCoordinator(BealjasCoordinator):
    '''
    Superclass for window based plotters
    '''
    
    #
    # Public interface
    #    
    def __init__(self,                  
                 nWindows, nSlaves,
                 createWindowFunc, guiLoopFunc, closeFunc,
                 setVisibleFunc, getPosSizeFunc, setPosSizeFunc,
                 stateObject=None):
        '''
        Constructor
                
        @param nWindows: number of windows to open
        @param nSlaves: number of processes to start at once. If None, the default value of 
        nPlots is used. If 0 one slave process is started per task. Otherwise, the specified number 
        of slave processes are started. 
        @param createWindowFunction: function called in the slave constructur to create window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param guiLoopFunction: function called to execute the GUI loop. (functions for various backends may
        be found in modules bealjas2X modules the bealjas.utils package)
        @param closeFunc: function called to close the window and free any resource allocated
        by the visualization. (functions for various backends may be found in modules bealjas2X modules the 
        bealjas.utils package)
        @param setVisibleFunc: function called to either show or hide window (functions for various
        backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param getPosSizeFunc: function called to get the position and size of a window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param setPosSizeFunc: function called to set the position and size of a window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param stateObject: optional object sent as argument to many of the slave functions. Default is None        
        '''        
        BealjasCoordinator.__init__(self, stateObject)
        
        if nSlaves == None:
            self.logger.info('Creating windows coordinator with %d windows and %d slaves' % (nWindows, nWindows))
        elif nSlaves == 0:
            self.logger.info('Creating windows coordinator with %d windows and number-of-tasks slaves' % (nWindows))
        elif nSlaves < 0:
            raise CoordinatorError('Invalid number of slaves: %d' % (nSlaves))
        else:
            self.logger.info('Creating windows coordinator with %d windows and %d slaves' % (nWindows, nSlaves))
            if nWindows > nSlaves:
                raise CoordinatorError('nWindows > nSlaves')        
            
        if createWindowFunc != None and not isinstance(createWindowFunc, types.FunctionType):
            raise CoordinatorError('createWindowFunc: not a function')  
        if guiLoopFunc != None and not isinstance(guiLoopFunc, types.FunctionType):
            raise CoordinatorError('guiLoopFunc: not a function')
        if closeFunc != None and not isinstance(closeFunc, types.FunctionType):
            raise CoordinatorError('closeFunc: not a function')
        if getPosSizeFunc != None and not isinstance(getPosSizeFunc, types.FunctionType):
            raise CoordinatorError('getPosSizeFunc: not a function')
        if setPosSizeFunc != None and not isinstance(setPosSizeFunc, types.FunctionType):
            raise CoordinatorError('setPosSizeFunc: not a function')
          
        
        # Number of windows to open and number of slave processes to start
        self.nWindows = nWindows
        # The above may be changed if the number of tasks is reduces to less than nWindows
        # The below is never changed
        self.maxWindows = nWindows
        if nSlaves == None:
            self.nSlaves = nWindows
        else:
            self.nSlaves = nSlaves # 0 or the specified number
                    
        self.createWindowFunc = createWindowFunc
        self.guiLoopFunc = guiLoopFunc
        self.closeWindowFunc = closeFunc
        self.setWindowVisibleFunc = setVisibleFunc
        self.getWindowPosSizeFunc = getPosSizeFunc
        self.setWindowPosSizeFunc = setPosSizeFunc 
        
        # Queue of executed visualization functions
        #
        # This queue is used to playback multi-step visualizations when there are more tasks
        # than slaves.
        self.functionList = []
        
        # List of list of tasks to visualize. The visualization is split into multiple tasks
        # that are visualized independently.
        #
        # The list is initialized in startSlaves
        self.taskList = []
        
        # Currently running slaves.
        #
        # The list is initialized in startSlaves
        self.currentlyRunning = []        
        
        # Index in currentlyRunning[] and currentlyRunningPipes[] for visible windows
        self.currentlyVisible = []
        
        # Position and size of closed windows. These are used to place new windows
        # Each closed window is represented by a 5-tuple (m, x, y, w, h), where m is True if the window
        # is maximized, and where x, y, w, and h are either all None if m is True, or they contain 
        # respectively the window x position, y position, width and height. 
        self.closedWindowsPosSize = []
        
        # To ensure experiment repeatability
        random.seed(31278)       
            
    def startSlaves(self, visualizeFunction, tasks):
        """
        Start slaves processes and show first set of windows
        
        @param vizualizeFunction: visualization function that is run on all slaves. The function
        may be None.
        @param tasks: list of tasks where each element is used as input to one slave.                       
        @return: None
        """
        
        if visualizeFunction != None and not isinstance(visualizeFunction, types.FunctionType):
            raise CoordinatorError('Visualize function error: not a function')  
        if tasks == None or len(tasks) < 1:
            raise CoordinatorError('No tasks specified')          
        
        if self.nSlaves != 0 and len(tasks) < self.nSlaves:
            raise CoordinatorError('Number of tasks is less than number of slaves to start (%d < %d)' % (len(tasks), self.nSlaves))                    
        
        if len(self.currentlyRunning) > 0:
            raise CoordinatorError('Slaves already running, please kill these before starting new slaves')        
        
        # Create tasks object
        nTasks = len(tasks)
        for i in range(nTasks):                                
            self.taskList.append( BealjasTask(i) )
                                
        if self.nSlaves == 0:
            # The number of tasks to start is now known
            self.logger.info("Setting nSlaves = %d" % (nTasks))
            self.nSlaves = nTasks

        # Check if nWindows have been changed (in case of restart of slaves)
        if self.nWindows != self.maxWindows:
            self.nWindows = self.maxWindows
            
        self.logger.info('Starting %d slaves' % (self.nSlaves))
        # 
        # Number of processes to start in the background too precompute the visualization to show
        # This number is set such that processes are aggressively pre-started. However, if there are
        # many available hidden processes some are reserved for recently shown visualizations.            
        if self.nSlaves == self.nWindows: # No room for buffering tasks        
            # Map slaves to the first N tasks
            for i in range(self.nSlaves):
                self.startSlave(self.taskList[i])
            self.startAhead = 0            
        elif self.nSlaves >= self.nWindows * 3:
            # Enough slaves that some can be used to cache previously shown tasks
            startBehind = self.nWindows
            self.startAhead = (self.nSlaves - self.nWindows) - startBehind
            # Start visible and prestarted tasks first
            for i in range(0, self.nWindows + self.startAhead):
                self.startSlave(self.taskList[i])                           
            # The previsouly shown tasks
            for i in range(nTasks - startBehind, nTasks):
                self.startSlave(self.taskList[i])
                # A low value such that these will be selected by the LRU algorithm 
                self.taskList[i].showTimestamp = 1                                          
        else:
            # Use all extra slaves for precomputing visualizations predicted to be shown soon
            self.startAhead = self.nSlaves - self.nWindows # Always less than self.nWindwos            
            for i in range(0, self.nWindows + self.startAhead):
                self.startSlave(self.taskList[i])
        # The movement direction is used to determine how many new processes to start
        self.lastMovement = "next" # tasks set up for showing next
        
        self.logger.info("nPreStart (process started for pre visualization): %d" % (self.startAhead))
                
        assert(len(self.currentlyRunning) == self.nSlaves)
        
        if visualizeFunction == None:
            return
        
        self.logger.debug('Executing first visualize function on %d of %d slaves' % (self.nWindows, self.nSlaves))            
        self.visualize(visualizeFunction, tasks)            
 
    def killSlaves(self):
        '''
        Kill all slave processes
        '''    
        #assert(len(self.currentlyRunning) == len(self.taskList)), "Assuming 1-1 mapping between tasks and slaves"
        
        self.logger.info("Killing %d currently running slaves" % (len(self.currentlyRunning)))
        # Must iterate over a copy of currenlyRunning since killSlave() removes elements from the list
        for task in self.currentlyRunning[:]:
            self.killSlave(task)            
            
        assert(len(self.currentlyRunning) == 0), '%d slaves still running' % (len(self.currentlyRunning))
        self.taskList = []
        self.functionList = []
        
    def visualize(self, visualizeFunction, tasks):
        '''
        Distribute visualization tasks and run a visualization function on all slaves.
        
        The slaves to set as visible are determined as follows:
        - nWindows - currentlyVisible = N, new windows to show
        - The N new windows to show are from the N first hidden slaves
        - The remaining shown windows are updated with the new visualize function
        
        @param visualizeFunction: user provided visualize function called to do perform the slave task
        @param tasks: list of tasks where each element is used as input to one slave.        
        @return: None
        
        TODO: may want to add a label/name to the function for debugging, logging and possible future
        playback mechanisms
        '''
        
        if not isinstance(visualizeFunction, types.FunctionType):
            raise CoordinatorError('Invalid visualize function: not a function')        
        if len(self.taskList) == 0:
            raise CoordinatorError('Slaves have not been started')
        if tasks == None:
            raise CoordinatorError('Invalid tasks argument: None')
        # The number of tasks cannot be changed
        if len(tasks) != len(self.taskList):
            raise CoordinatorError('Invalid number of entries in taskArgument list (got %d expected %d)' % (len(tasks), len(self.taskList)))
        #assert(len(self.currentlyRunning) == len(self.taskList)), "Assuming 1-1 mapping between tasks and slaves (was %d != %d)" % (len(self.taskList), len(self.currentlyRunning))        
        
        self.logger.debug("Execute visualize function: %s" % (visualizeFunction))
                        
        # Update currently visible slaves
        self.showMore()
        
        # Update list of executed functions and distribute new tasks to visualize
        self.functionList.append(visualizeFunction)         
        for i in range(len(tasks)):
            self.taskList[i].taskArguments.append(tasks[i])
            assert(len(self.functionList) == len(self.taskList[i].taskArguments))
        
        # Send new visualization function to all running slaves, first to the visible then to the hidden visualizations
        self.logger.info("Send visualize function to %d currently visible slaves" % (len(self.currentlyVisible)))
        failedSlaves = []
        for task in self.currentlyVisible:
            try:
                self.sendVisualizeMsg(task.pipe, visualizeFunction, task.taskArguments[-1])
            except SlaveError:
                self.logger.warn('Task has been killed')
                failedSlaves.append(task)    
                        
        self.logger.info("Send visualize function to %d currently running, but hidden, slaves" % (len(self.taskList) - len(self.currentlyVisible)))
        for task in self.currentlyRunning:
            if task.isVisible == False: # Cmd not yet sent
                try:
                    self.sendVisualizeMsg(task.pipe, visualizeFunction, task.taskArguments[-1])
                except SlaveError:
                    self.logger.warn('Task has been killed')
                    self.failedSlaves.append(task)   
                    
        if len(failedSlaves) > 0:
            # Some slaves were killed by the user
            for f in failedSlaves:
                self.killSlave(f, True)
            self.showMore()
            
    def filterVisible(self, filterFunction, filterArgs=None):
        """
        Determine views to show using a filter function. This function will run the filter function on 
        all running tasks even if they are not currently running.
        
        @param filterFunction: filter function to run
        @param filterArgs: arguments passed to filter function
        @return: number of task visible after filter
        """
        
        if not isinstance(filterFunction, types.FunctionType):
            raise CoordinatorError('Invalid visualize function: not a function')        
        if len(self.taskList) == 0:
            raise CoordinatorError('Slaves have not been started')
        if filterArgs != None and len(filterArgs) != len(self.taskList):
            raise CoordinatorError('Invalid number of entries in taskArgument list (got %d expected %d)' % (len(filterArgs), len(self.taskList)))        
        
        self.logger.info("Sending filter function to %d tasks" % (len(self.taskList)))
        
        # task2rv is a mapping of task objects to filter function return values
        task2rv = self.executeFunctionAll('filter', filterFunction, filterArgs)
              
        # Update filter status on all tasks
        nVisibleAfterFilter = 0
        for task in self.taskList:
            rv = task2rv[task]
            if rv == True:
                task.visibleFilter = True
                nVisibleAfterFilter += 1
            else:
                self.logger.debug('Filtered to be hidden: %d' % (task.taskID))
                task.visibleFilter = False
                if task.isVisible == True:
                    # Save coordinates for later use
                    self.logger.debug('Hide open window: %d' % (task.taskID))
                    self.hideTask(task)
                    # hideTask return value is ignored since there is nothing to do for killed tasks                                        
                
        self.logger.info("%d of %d tasks visible after filter" % (nVisibleAfterFilter, len(self.taskList)))
        
        # This will show additional windows if some were hidden
        self.showMore()
        
        return nVisibleAfterFilter
                
    def pendingFilterVisible(self, filterFunction, filterArgs=None):
        """
        Determine views to show using a filter function. This function will run the filter function on 
        all running tasks even if they are not currently running.
        
        @param filterFunction: filter function to run
        @param filterArgs: arguments passed to filter function
        """
        if not isinstance(filterFunction, types.FunctionType):
            raise CoordinatorError('Invalid visualize function: not a function')        
        if len(self.taskList) == 0:
            raise CoordinatorError('Slaves have not been started')
        if filterArgs != None and len(filterArgs) != len(self.taskList):
            raise CoordinatorError('Invalid number of entries in taskArgument list (got %d expected %d)' % (len(filterArgs), len(self.taskList)))
        
        # Set filter to pending for all tasks, the currently running will be executed below
        for task in self.taskList:            
            if filterArgs != None:
                task.pendingVisibleFilter = (filterFunction, filterArgs[task.taskID])
            else:
                task.pendingVisibleFilter = (filterFunction, None)
            # Make sure to clear the visibleFilter field
            task.visibleFilter = None
        
        # Execute filter function on currently running slaves. The return values are gathered below
        tasks = self.currentlyRunning[:] # Need a copy since below function may remove elements from the list
        for t in tasks:
            self.executeVisibleFilterTask(t)
            
    def filterVisibleOff(self):
        """
        Turn off visible filter
        """
        self.logger.info("Turn off visible filter")
        for task in self.taskList:
            task.visibleFilter = None        
        
        self.showMore()     
        
    def filterKill(self, filterFunction, filterArgs=None):
        """
        Use a filter function to close some of the tasks. The close is pending so non-running tasks are not closed
        before they are scheduled to run.
        
        @param filterFunction: function to run. The return value determines wheter the slave should be
        kept (True) or killed (False)
        @param filterArgs: optional argument sent to the filter function
        @return: number of non-closed tasks
        """
        if not isinstance(filterFunction, types.FunctionType):
            raise CoordinatorError('Invalid visualize function: not a function')        
        if len(self.taskList) == 0:
            raise CoordinatorError('Slaves have not been started')
        if filterArgs != None and len(filterArgs) != len(self.taskList):
            raise CoordinatorError('Invalid number of entries in taskArgument list (got %d expected %d)' % (len(filterArgs), len(self.taskList)))
        #assert(len(self.currentlyRunning) == len(self.taskList)), "Assuming 1-1 mapping between tasks and slaves (was %d != %d)" % (len(self.taskList), len(self.currentlyRunning))
        
        self.logger.debug("Filter task to kill")
        
        # Run filter function
        task2rv = self.executeFunctionAll('filter', filterFunction, filterArgs)
        
        toKill = []
        nRemaining = 0
        for task in self.taskList:
            if task not in task2rv: # Closed by user or already closed
                continue
            rv = task2rv[task]
            if rv == False:                                
                task.filterKill = True
                if task in self.currentlyRunning:
                    toKill.append(task)
            elif rv == True:
                nRemaining += 1
            else:
                raise CoordinatorError('Unexpected task2rv value: %s' % (rv))
                
        self.logger.info("%d of %d tasks will be killed (%d already closed)" % (len(toKill), len(self.taskList), len(self.taskList) - nRemaining))
        # Kill currently running processes        
        for task in toKill:
            if task.isRunning():
                self.killSlave(task)          
        
        if nRemaining < self.nWindows:
            self.nWindows = nRemaining
        if nRemaining < self.nSlaves:
            self.nSlaves = nRemaining
                    
        # Start more slaves
        self.logger.info("Currently %d of %d slaves are running, so starting more slaves" % (len(self.currentlyRunning), self.nSlaves))
        for task in self.taskList:
            if len(self.currentlyRunning) == self.nSlaves:
                break
            if not task.filterKill:
                self.startSlave(task) 
            
        self.logger.info("%d slaves currently running" % (len(self.currentlyRunning))) 
        self.showMore()
               
        return nRemaining
    
    def gather(self, gatherFunction, gatherArgs=None):
        '''
        Gather values from all tasks
        
        @param gatherFunction: function run on each slave that returns the values to gather.
        @param gatherArgs: arguments passed to the gather function
        @return: list of values returned by the close functions on each slave
        '''
        
        if not isinstance(gatherFunction, types.FunctionType):
            raise CoordinatorError('First argument is not a function')
        if len(self.taskList) == 0:
            raise CoordinatorError('Slaves have not been started')
        if gatherArgs != None and len(gatherArgs) != len(self.taskList):
            raise CoordinatorError('Invalid number of entries in gatherArgs list (got %d expected %d)' % (len(gatherArgs), len(self.taskList)))
        
        self.logger.debug("Gather values")        
        # Save return in a hash map that maps task object to return values
        task2rv = self.executeFunctionAll('gather', gatherFunction, gatherArgs)
        
        rvs = []
        for task in self.taskList:
            if task not in task2rv:
                rvs.append(None)
            else:
                rvs.append( task2rv[task] )
            
        return rvs
    
    def readLogFiles(self):
        '''
        Gather the log file content from all tasks
        
        @return list where each element is a list of log file lines read from a task
        '''
        
        if len(self.taskList) == 0:
            raise CoordinatorError('Slaves have not been started')
        self.logger.info("Gather logfile lines from all running slaves")
        
        # Save return in a hash map that maps task object to return values
        task2rv = self.executeFunctionAll('readLogFiles')
        
        rvs = []
        for task in self.taskList:
            if task not in task2rv:
                rvs.append(None)
            else:
                rvs.append( task2rv[task] )
            
        return rvs
    
    def showNext(self, nToShow=None):
        """
        Hide all currently visible task and show the show the next set of tasks. The tasks are selected based 
        on their task arguments array index (taskID number). For example if tasks 2, 3 and 13 are currently 
        visible, then these will be replaces with 4, 5 and 14.
        
        @param nToShow: optional number of new tasks to show, if not specified self.nWindows is used instead
        """
        
        if nToShow == None:
            nToShow = self.nWindows
        if nToShow > self.nWindows:
            raise CoordinatorError('Cannot show more than %d windows: %d requested' % (self.nWindows, nToShow))
                    
        # Find slaves to hide and show
        self.logger.debug("Show next set of slaves")        
        toHide = self.currentlyVisible[:nToShow] # ok if second index > list length
        
        # First select a task to show for each task that is hidden
        toShow = []
        for task in toHide:
            ss = self.selectNextTaskToShow(task.taskID, toHide + toShow)
            if ss == None:
                # No more tasks that can be shown
                break
            else:
                toShow.append(ss)
                
        # Make sure to also check length of toHide, because if this list is empty then toShow will be empty here
        if len(toShow) == 0 and len(toHide) > 0: 
            # No tasks could be shown
            self.logger.warn("No additional tasks could be shown")
            return
        
        # Second select additional task to show
        while len(toShow) < nToShow:
            # Always use last select task
            if len(toShow) > 0:
                hi = toShow[-1]
            else:
                hi = self.taskList[0]
            ss = self.selectNextTaskToShow(hi.taskID, toHide + toShow)
            if ss == None:
                # No more tasks that can be shown
                break
            else:
                toShow.append(ss)
                
        # Third, if still to few tasks then remove some from the hide list
        if len(toShow) < nToShow:
            # Remove the first task
            if toHide > nToShow - len(toShow):            
                toHide = toHide[(nToShow - len(toShow)):]
            else: # remove all
                toHide = []
                                
        # Hide and show tasks            
        self.showCommon(toHide, toShow)
        
        # Finally, if some tasks could not be shown (pending filter visible), then attempt to find
        # additional tasks to show
        #
        # Use the last selected task to show as starting point
        if len(toShow) > 0:
            hi = toShow[-1]
        else:
            hi = self.taskList[0]
        while len(self.currentlyVisible) < nToShow:
            ss = self.selectNextTaskToShow(hi.taskID, self.currentlyVisible)
            if ss == None:
                # No more showable tasks
                break
                        
            self.showTask(ss)
            # continue independent of return value. If the task was shown, the loop will
            # exit if currentlyVisible == nToShow. If it was not shown another task will
            # be selected since currentlyVisible < nToShow
            
        # Start some processes in the background such that these can quickly be shown
        if self.startAhead > 0:
            if self.lastMovement == 'next':
                # Continuing in same direction, so only start enough task to replace the hidden tasks
                if len(toHide) > 0:                        
                    self.startNext(len(toHide))
            else:
                # Started in new direction so start the maximum allowable task
                self.startNext(self.startAhead)
        self.lastMovement = 'next'
        
    def showPrevious(self, nToShow=None):
        """
        Hide all currently visible task and show the show the previous set of tasks. The tasks are selected based 
        on their task arguments array index (taskID number).  For example if 1, 2, and 13 (of 30) tasks are visible, 
        the previous task shown are 0, 12, and 29
        
        @param nToShow: optional number of new tasks to show, if not specified self.nWindows is used instead
        """
        
        if nToShow == None:
            nToShow = self.nWindows
        if nToShow > self.nWindows:
            raise CoordinatorError('Cannot show more than %d windows: %d requested' % (self.nWindows, nToShow))
        
        # Find slaves to hide and show
        self.logger.debug("Show previous %d tasks" % (nToShow))
        toHide = None
        if len(self.currentlyVisible) > nToShow:
            # Do not hide all windows.
            # THe lest N windows are hidden
            toHide = self.currentlyVisible[(len(self.currentlyVisible) - nToShow):]
        else:
            toHide = self.currentlyVisible[:]
                
        # First select a task to show for each task that is hidden
        toShow = []
        for task in toHide:
            ss = self.selectPrevTaskToShow(task.taskID, toHide + toShow)
            if ss == None:
                # No more tasks that can be shown
                break
            else:
                toShow.append(ss)
                
        if len(toShow) == 0:
            # No tasks could be shown
            self.logger.warn("No additional tasks could be shown")
            return
        
        # Second select additional task to show
        while len(toShow) < nToShow:
            # Always use the first select task            
            hi = toShow[0]
            ss = self.selectPrevTaskToShow(hi.taskID, toHide + toShow)
            if ss == None:
                # No more tasks that can be shown
                break
            else:
                toShow.append(ss)
                
        # Third, if still to few tasks then remove some from the hide list
        if len(toShow) < nToShow:
            # Remove the last tasks
            toShow = toShow[:nToShow]
                                
        # Hide and show tasks            
        self.showCommon(toHide, toShow)
        
        # Finally, if some tasks could not be shown (pending filter visible), then attempt to find
        # additional tasks to show
        #
        # Use the first selected task to show as starting point
        if len(toShow) > 0:
            hi = toShow[0]
        else:
            hi = self.taskList[0]
        while len(self.currentlyVisible) < nToShow:
            ss = self.selectPrevTaskToShow(hi.taskID, self.currentlyVisible)
            if ss == None:
                # No more showable tasks
                break
                        
            self.showTask(ss)
            # continue independent of return value. If the task was shown, the loop will
            # exit if currentlyVisible == nToShow. If it was not shown another task will
            # be selected since currentlyVisible < nToShow
            
        # Start some processes in the background such that these can quickly be shown
        if self.startAhead > 0:
            if self.lastMovement == 'prev':
                # Continuing in same direction, so only start enough task to replace the hidden tasks
                if len(toHide) > 0:                        
                    self.startPrev(len(toHide))
            else:
                # Started in new direction so start the maximum allowable task
                self.startPrev(self.startAhead)
        self.lastMovement = 'prev'
        
    def showRandom(self, nToShow=None):
        """
        Show a random set of views
        
        @param nToShow: optional number of new tasks to show,
        """
        
        if nToShow == None:
            nToShow = self.nWindows
        if nToShow > self.nWindows:
            raise CoordinatorError('Cannot show more than %d windows: %d requested' % (self.nWindows, nToShow))
        
        self.logger.debug("Show %d random tasks" % (nToShow))
        
        toHide = self.currentlyVisible[:]
        
        toShow = []
        # Select n tasks randomly
        while len(toShow) < self.nWindows:
            ti = random.randint(0, len(self.taskList) - 1)
            ss = self.taskList[ti]
            
            if ss in toShow:
                continue            
            # Add even if already shown
            elif ss.isShowable() or ss.isVisible:
                toShow.append(ss)
                
        # Remove tasks to show that are already shown
        toRemove = []
        for ss in toShow:
            if ss in toHide:
                toRemove.append(ss)
        for ss in toRemove:
            toHide.remove(ss)
            toShow.remove(ss)
            
        # Hide and show tasks            
        self.showCommon(toHide, toShow)
                
        # Finally, if some tasks could not be shown (pending filter visible), then attempt to find
        # additional tasks to show
        #
        # Use the last selected task to show as starting point
        checked = toShow + toRemove        
        nTasks = len(self.taskList)
        while len(self.currentlyVisible) < nToShow:
            if len(checked) == len(self.taskList):
                self.logger.warn("No more showable tasks")
            # Select a random start index (the first task selected/checked is ti+1
            ti = random.randint(0, nTasks - 1)            
            # Iterate until a showable or shown task is found
            ss = self.selectNextTaskToShow(ti, checked)
            if ss == None:
                # No more showable tasks
                break
            checked.append(ss)
                        
            self.showTask(ss)
            # continue independent of return value. If the task was shown, the loop will
            # exit if currentlyVisible == nToShow. If it was not shown another task will
            # be selected since currentlyVisible < nToShow
            
        # Do not start any new task since the movement is expected to be random
        self.lastMovement = 'random'
    
    def showIndex(self, startIndex, nToShow=None):
        """
        Show views starting from an startIndex. 
        For example if 1, 2, and 13 are visible and the startIndex is 1, this will show 1, 2, and 3 
        
        @param startIndex: start startIndex
        @param nToShow: optional number of new tasks to show, if not specified self.nWindows is used instead
        """
        
        if startIndex < 0 or startIndex > len(self.taskList):
            raise CoordinatorError('Invalid start index: %d (not in range 0...%d)' % (startIndex, len(self.taskList)))
        
        if nToShow == None:
            nToShow = self.nWindows
        
        self.logger.debug("Show %d tasks starting from startIndex %d" % (nToShow, startIndex))
        
        # Select tasks to show starting from the provided startIndex
        toShow = []    
        while len(toShow) < nToShow:
            # -1 since the first argument specifies the last index to consider
            ss = self.selectNextTaskToShow((startIndex - 1) % len(self.taskList), toShow, True)
            if ss == None:
                # No more tasks to show
                break
            else:
                toShow.append(ss)
                
        # Find tasks to hide
        toHide = []
        for task in self.currentlyVisible:
            if task not in toShow:
                toHide.append(task)
        
        # Remove already shown tasks from the toShow list
        for task in self.currentlyVisible:
            if task in toShow:
                toShow.remove(task)
                
        # Hide and show tasks            
        self.showCommon(toHide, toShow)
        
        # Finally, if some tasks could not be shown (pending filter visible), then attempt to find
        # additional tasks to show
        #
        # Use the last selected task to show as starting point
        if len(toShow) > 0:
            hi = toShow[-1]
        else:
            hi = self.taskList[startIndex]
        while len(self.currentlyVisible) < nToShow:
            ss = self.selectNextTaskToShow(hi.taskID, self.currentlyVisible)
            if ss == None:
                # No more showable tasks
                break
                        
            self.showTask(ss)
            # continue independent of return value. If the task was shown, the loop will
            # exit if currentlyVisible == nToShow. If it was not shown another task will
            # be selected since currentlyVisible < nToShow
            
        # Start some processes in the background such that these can quickly be shown
        # Attempt to start as many task as allowed
        self.startNext(self.startAhead)
        self.lastMovement = 'startIndex'
        
    def showIndexes(self, taskIndexes):
        """
        Show specific views
        
        taskIndexes: index of the visualizations to show in the task list that was passed as
        argument to the visualization function.
        """  
        
        if (len(taskIndexes) > self.nWindows):
            raise CoordinatorError("Too many task indexes (got; %d, max is: %d)" % (len(taskIndexes), self.nWindows))
        for i in taskIndexes:
            if i < 0 or i > len(self.taskList):
                raise CoordinatorError('Invalid index: %d (not in range 0...%d)' % (i, len(self.taskList)))
        
        
        self.logger.info("Set the following tasks to visible: %s" % (taskIndexes))
        
        toHide = self.currentlyVisible[:]
        
        toShow = []
        for ti in taskIndexes:
            ss = self.taskList[ti]
            #assert(ss.slave != None), 'Assuming that nTasks == nSlaves'
            
            if ss in toHide: # Already visible
                toHide.remove(ss)
            elif ss.isShowable():
                toShow.append(ss)
                
        # Send hide and show messages and update currentlyVisible list
        self.showCommon(toHide, toShow)     
        
        # Start some processes in the background such that these can quickly be shown
        # Attempt to start as many task as allowed
        self.startNext(self.startAhead)
        self.lastMovement = 'indexes'
        
    def startNext(self, nToStart):
        '''
        Start new tasks that are likely to be shown in the near future.
        
        This function selects the next 'noToStart' showable tasks to start. It will also kill the 
        least recently shown tasks.
        
        @param nToStart: number of new tasks to start
        '''
        self.logger.debug('Start next set of processes in the background')
        self.startCommon(nToStart, self.selectNextTaskToShow)
        
    def startPrev(self, nToStart):
        '''
        Start new tasks that are likely to be shown in the near future.
        
        This function selects the previous showable tasks to start. It will also kill the 
        least recently shown tasks.
        
        @param nToStart: number of new tasks to start
        '''

        self.logger.debug('Start previous set of processes in the background')
        self.startCommon(nToStart, self.selectPrevTaskToShow)

    #
    # Window specific functinos
    #
            
    def resetVisible(self):
        """
        Resend hide and show messages. This function is useful if there is a glitch in the setVisible
        functions which sometimes may happen (especially at startup)
        """
        self.logger.info("Resending show and hide messages")
        failedTasks = []
        for t in self.currentlyRunning:
            if t in self.currentlyVisible:
                try:
                    self.sendShowMsg(t.pipe)
                except SlaveError:
                    self.logger.warn("Task has been killed by user")
                    failedTasks.append(t)
            else:
                try:
                    self.sendHideMsg(t.pipe)
                except SlaveError:
                    self.logger.warn("Task has been killed by user")
                    failedTasks.append(t)
        
        # Cleanup data structures for failed tasks
        for f in failedTasks:
            self.killSlave(f, True)
        self.showMore()  

    def showMore(self):
        """
        Show additional visualizations. This function only has an effect if currently visible is less 
        than nWindows.
        """
        self.logger.debug("%d of %d visualizations are visible" % (len(self.currentlyVisible), self.nWindows))
        
        while 1: # more to show
            nNew = self.nWindows - len(self.currentlyVisible)       
            if nNew <= 0:
                self.logger.debug('No more visualizations that can be shown')
                return
            else:
                self.logger.debug('Show %d new tasks' % (nNew))
             
            toShow = []
            # First check if there are currently running slaves that are hidden
            for task in self.currentlyRunning:
                if len(toShow) == nNew:
                    break # All found
                elif task.isShowable():
                    toShow.append(task)    
                    
            self.logger.debug('Currently running that can be shown: %d' % (len(toShow)))                
                    
            # Then select from the non-running tasks
            for task in self.taskList:
                if len(toShow) == nNew:
                    break # All found            
                elif task.isRunning():
                    # Checked the currently running tasks above
                    continue
                elif task.isShowable():
                    toShow.append(task)
                    
            self.logger.debug('Showable tasks (running and not running): %d' % (len(toShow)))
                            
            if len(toShow) != nNew:                
                # Most tasks may be filtered hidden
                self.logger.warn("Could not find %d additional tasks to show: %d found" % (nNew, len(toShow)))
            
            # Set visualization state of new slaves to visible
            self.logger.debug("Set %d new tasks to visible" % (len(toShow)))
            nShown = 0
            for task in toShow:                     
                if self.showTask(task):
                    nShown += 1
                    
            if nShown == len(toShow):
                return # All shown
        
    #
    # Helper functions
    #        
        
    def startSlave(self, task):
        '''
        Helper function to start a slave process and update the necessary task data structures.
        
        @param task: task to start a slave for
        @param stateObject: object replicated to all slaves. Must be serializable. Default is None.
        '''                
        print 'C: ', time.time()
        
        if task.isRunning():
            raise CoordinatorError('Task is already running')
        if task.filterKill:
            raise CoordinatorError('Task is killed')
        
        self.logger.debug('Start slave for task: %s' % (task.taskID))   
        
        # Create pipe used to send command to slave and receive return values
        parentEnd, childEnd = Pipe()
        # Fork of a new process using Pythons multiprocessing module
        p = Process(target=plotterEntry, args=('BSV Plotter %d' % (task.taskID),
                                               childEnd, 
                                               self.slaveLoggerSetupFunction,
                                               self.slaveLoggerLevel,
                                               self.createWindowFunc,
                                               self.guiLoopFunc,       
                                               self.closeWindowFunc,                                            
                                               self.setWindowVisibleFunc,
                                               self.getWindowPosSizeFunc,
                                               self.setWindowPosSizeFunc,
                                               self.slaveLoggerOptionalArguments,
                                               self.stateObject))
        p.start()
            
        # Update task data structures
        task.setSlave(p, parentEnd)
        self.currentlyRunning.append(task)
        
        # Send previously executed visualization functions to slaves
        if len(self.functionList) > 0:
            self.logger.debug('Sending previously executed synchronization commands to slave; %s' % (task.taskID))
            for i in range(len(self.functionList)):        
                try:
                    self.sendVisualizeMsg(parentEnd, self.functionList[i], task.taskArguments[i])
                except SlaveError:
                    self.logger.warn('Task has been killed')
                    self.killSlave(task, True)
                    return
                
        # Windows starts hidden
        self.sendHideMsg(task.pipe)
                    
        # Execute pending filter functions
        if task.pendingVisibleFilter != None:                        
            self.executeVisibleFilterTask(task)                                  
          
    def killSlave(self, task, isAlreadyKilled=False):
        '''
        Helper function to kill a slave process and update the necessary task data structures.
        
        @param task: task to kill
        @param isAlreadyKilled: optional argument set to True if the slave process is already killed
        '''
        if not task.isRunning():
            raise CoordinatorError('Task is not running')
                
        if task.isVisible and isAlreadyKilled == False:
            try:
                posSize = self.sendGetPosSizeMsg(task.pipe)
                self.closedWindowsPosSize.append(posSize)
            except SlaveError:
                self.logger.warn('Task was killed on other end')
                isAlreadyKilled = True                            
                      
        if not isAlreadyKilled:
            self.logger.debug('Close task window: %s' % (task.taskID))
            try:
                self.sendCloseMsg(task.pipe)
                self.logger.debug('Kill task: %s' % (task.taskID))        
                self.sendKillMsg(task.pipe)
            except SlaveError:
                self.logger.warn('Task was killed on other end')
        
        self.logger.debug("Join with task: %s" % (task.taskID))                       
        task.slave.join()
        
        task.pipe.close()
        task.pipe = None
        task.slave = None        
        
        # Cleanup global variables
        self.currentlyRunning.remove(task)
        if task in self.currentlyVisible[:]:
            task.isVisible = False
            # The remove will throw a ValueError if task is not in currentlyVisible
            self.currentlyVisible.remove(task)
        
    def selectSlaveToKill(self, ignoreTask):
        '''
        Select a currently running slave to kill
        
        @param ignoreTask: ignore a task that is not to be killed. Typically a task
        that has just been started.
        @return: task to kill
        '''
        
        self.logger.debug('Select slave to kill')
        
        # Select a slave to kill using a simple policy where:
        # 1. The slave to kill is the first running slave that is filtered out, or
        # 2. if visible filter is not on, the first slave that is not visible
        # 3. If all slaves are visible, the first running slave
        for task in self.currentlyRunning:
            if task == ignoreTask:
                continue
            if task.visibleFilter == False:
                self.logger.debug('Selected to kill task filtered out by visible filter: %s' % (task.taskID))
                return task
            
        for task in self.currentlyRunning:
            if task == ignoreTask:
                continue
            if task.isVisible == False:
                self.logger.debug('Selected to kill hidden task: %s' % (task.taskID))
                return task
            
        for task in self.currentlyRunning:
            if task == ignoreTask:
                continue
            else:
                return task            
            
        raise CoordinatorError('Could not select a slave to kill')
        
    
    def showTask(self, task):
        '''
        Helper function to show a task and update all meta-data.
        
        Note! this function will throw an exception if the task is already visible
        
        @param task: task to show
        @param pst: optional position and size 5-tuple used to move and resize the window to show 
        @return: True if task was shown, or False if it could not be shown. The latter may happen
        if the visibleFilter is False or more commonly if a pending visible filter returns False
        '''
        
        if task.filterKill:
            raise CoordinatorError('Task to show has been killed')        
        if task in self.currentlyVisible:
            raise CoordinatorError("Task to show is already visible")
        if task.isVisible == True:
            raise CoordinatorError("Task to show is already visible (and inconsistency between isVisible and currentlyVisible[])")
        if task.visibleFilter == False:
            self.logger.warn("Cannot show task currently filtered out by filter function")
            return False
        
        self.logger.debug("show: %s" % (task.taskID))        

        if task not in self.currentlyRunning:
            # First start a new process. We assume it is better to run an additional process
            # rather than increase the latency for the show task operation
            self.startSlave(task)
            if len(self.currentlyRunning) > self.nSlaves: # Must kill a slave
                # Do not kill recently started task             
                self.killSlave( self.selectSlaveToKill(task) )
            if task.visibleFilter == False:
                # Task had a pending visible filter function that returned False
                # Hence task should not be shown
                return False
                    
        # Always send even if it is already visible
        try:
            self.sendShowMsg(task.pipe)
        except SlaveError:
            self.logger.warn("Task has been killed by user")
            self.killSlave(task, True)
            return False
        
        task.isVisible = True
        self.currentlyVisible.append(task)        
        task.showTimestamp = time.time()
        
        pst = None
        # Check if there are window coordinates and size for a previously closed windows, 
        # if so use these such that the window appears in the "expected spot"
        if len(self.closedWindowsPosSize) > 0:
            pst = self.closedWindowsPosSize.pop()
            
        # Set window position and size to the specified values
        if pst != None:
            try:
                self.sendSetPosSizeMsg(task.pipe, pst)
            except SlaveError:
                self.logger.warn("Task has been killed by user")
                self.killSlave(task, True)
                return False
                
        # Success
        return True
            
    def hideTask(self, task):
        '''
        Helper function to hide a task and update all meta-data
        
        Note! this function may be called for already hidden tasks
        
        @param task: task to hide
        @return: True on success, or False on failure. Failures can occur if the user has closed the window
        to hide. 
        '''
        
        if not task.isRunning():
            raise CoordinatorError('Task to hide is not currently running')
        if task.filterKill:
            raise CoordinatorError('Task to hide has been killed')        
        if task not in self.currentlyVisible:
            raise CoordinatorError("Task to hide is not visible")
        if task.isVisible == False:
            raise CoordinatorError("Task to hide is not visible (and inconsistency between isVisible and currentlyVisible[])")
        
        self.logger.debug("hide: %s" % (task.taskID))
        # Always send even if the task is already hidden
        try:
            self.sendHideMsg(task.pipe)
        except SlaveError:
            self.logger.warn("Task has been killed by user")
            self.killSlave(task, True)
            return False
        
        task.isVisible = False        
        # Will throw an ValueError if task is not in currentlyVisible
        self.currentlyVisible.remove(task)
        
        # Get position and size of window
        pst = self.sendGetPosSizeMsg(task.pipe)
        if pst == None:
            self.logger.warn("Task has been killed by user")
            self.killSlave(task, True)
            return False
        else:
            self.closedWindowsPosSize.append(pst)
            return True
        
    def selectNextTaskToShow(self, startTaskID, excludedTasks, includeVisibleTasks=False):
        '''
        Select next task to show
        
        @param startTaskID: taskID of task for which to show the next task (for example if 0, then 1 is selected).
        Note that this task may be selected if no other tasks can be shown.
        @param excludedTasks: tasks that should not be shown. These are typically the currently visible tasks
        or tasks that have been hidden before calling this function.
        @param includeVisibleTasks: optional parameter. If True a visible task can be selected. 
        @return: task to show
        '''
        nTasks = len(self.taskList)
        # Iterate over all tasks, including the start task which is checked last        
        for i in range(1, nTasks + 1):            
            ss = self.taskList[(startTaskID + i) % nTasks]
            if ss in excludedTasks:
                # Already selected or should not be selected for other reasons
                continue
            if includeVisibleTasks and ss.isVisible:
                return ss
            if not ss.isShowable():
                # Task cannot be shown (including visible tasks if includeVisibleTasks=False)
                continue
            # Task that can be shown has been found
            return ss
        
        # No task could be selected
        return None
    
    def selectPrevTaskToShow(self, startTaskID, excludedTasks):
        '''
        Select previous task to show
        
        @param startTaskID: taskID of task for which to show the prev task (for example if 1, then 0 is selected).
        Note that this task may be selected if no other tasks can be shown.
        @param excludedTasks: tasks that should not be shown. These are typically the currently visible tasks
        or tasks that have been hidden before calling this function.
        @return: task to show
        '''
        nTasks = len(self.taskList)
        # Iterate over all tasks, including the start task which is checked last        
        for i in range(1, nTasks - 1):            
            ss = self.taskList[(startTaskID - i) % nTasks]
            if not ss.isShowable():
                # Task cannot be shown
                continue
            if ss not in excludedTasks:
                # Task that can be shown has been found
                return ss
        
        # No task could be selected
        return None
    
    def selectRandomTaskToShow(self, startTaskID, excludedTasks):
        '''
        Select a random task to show
        
        @param startTaskID: taskID of task for which to show the next task (for example if 0, then 1 is selected).
        Note that this task may be selected.
        @param excludedTasks: tasks that should not be shown. These are typically the currently visible tasks
        or tasks that have been hidden before calling this function.
        @return: task to show
        '''
        nTasks = len(self.taskList)
        # Iterate over all tasks, including the start task which is checked last        
        for i in range(1, nTasks + 1):            
            ss = self.taskList[(startTaskID + i) % nTasks]
            if not ss.isShowable():
                # Task cannot be shown
                continue
            if ss not in excludedTasks:
                # Task that can be shown has been found
                return ss
        
        # No task could be selected
        return None
                
    def showCommon(self, toHide, toShow):
        '''
        Helper function to send hide and show messages, and to update the currently visible list
        
        @param toHide: list of slaves to hide
        @param toShow: list of slave to show        
        '''
        
        # Make sure there are no duplicates
        for t in toHide:
            if toHide.count(t) > 1: 
                raise CoordinatorError('Duplicates in toHide: %s' % (t.taskID))
        for t in toShow:
            if toShow.count(t) > 1: 
                raise CoordinatorError('Duplicates in toShow: %s' % (t.taskID))
        # Check if a task is both in the toHide and toShow list        
        for t in toHide[:]:
            if t in toShow:
                self.logger.warn("Task both in toHide and toShow list: %s" % (t.taskID))                
                toHide.remove(t)                
        
        self.logger.debug("Hide %d of %d tasks" % (len(toHide), len(self.currentlyVisible)))
        for s in toHide:
            if s not in self.currentlyVisible:
                raise CoordinatorError('Slave to hide is not currently visible')
            self.hideTask(s)            
        
        self.logger.debug("Show %d new tasks" % (len(toShow)))
        nShown = 0
        for s in toShow:
            if self.showTask(s):
                nShown += 1    
            
        return nShown
    
    def startCommon(self, nToStart, selectFunction):
        '''
        Helper function for startNext and startPrev. Both call this function to do the actual work
        
        @param nToStart: number of tasks to start
        @param selectFunction: function run to select a task to start (either selectNextTaskToShow(),
        or selectPrevTaskToShow()
        '''
        if nToStart <= 0:
            self.logger.warn('Invalid number of tasks to start: %d' %(nToStart))
            return
        
        # no point stopping and then restarting the same tasks again
        if len(self.taskList) <= len(self.currentlyRunning) + nToStart:
            self.logger.warn('Too few non-running tasks')
            return    
                
        # 
        # Find tasks to stop
        #        
        # The function may return fewer than nToStart tasks        
        toStop = self.getLeastRecentlyShownTasks(nToStart)
        if len(toStop) < nToStart:
            # Attempt to find tasks that have not yet been shown but are running
            stopable = []
            for t in self.currentlyRunning:
                # not visible, not already selected, and is not yet shown
                if t not in self.currentlyVisible and t not in toStop and t.showTimestamp == None:
                    stopable.append(t)
            # Select among the stopable tasks randomly
            while len(stopable) > 0 and len(toStop) < nToStart:
                ri = random.randint(0, len(stopable) - 1)
                t = stopable.pop(ri)
                toStop.append(t)                
        
        if len(toStop) == 0:
            # All tasks are visible
            self.logger.warn("Cannot start next set of tasks: all currently running tasks are visible")
            return
        if len(toStop) < nToStart:
            # Reduce the number of tasks that can be started 
            nToStart = len(toStop)
        
        # Do not stop slaves before ensuring that there is a corresponding task that can be started
        
        #
        # Find tasks to start
        #        
        toStart = []
        for i in range(nToStart):
            # Select the next task relative to the currently visible tasks
            if len(self.currentlyVisible) > 0:
                hi = self.currentlyVisible[i % len(self.currentlyVisible)].taskID
            else:
                # no tasks are visible so we make the assumption that the first tasks are
                # most likely to be started
                hi = 0
            
            # Do not select one of the currently visible, already selected, or recently stopped tasks
            ss = selectFunction(hi, self.currentlyRunning + toStart + toStop)
            if ss == None:
                # No more tasks that can be shown
                break
            else:
                toStart.append(ss)
                
        if len(toStart) != nToStart:
            # This is possible if many tasks are filtered out
            self.logger.warn('Could not find %d tasks to start: only %d found' % (nToStart, len(toStart)))
            # Make sure the same number of slaves are stopped and started
            toStop = toStop[0:len(toStart)]
                
        #
        # Start and stop slaves
        #
        for t in toStop:
            self.killSlave(t)
        for t in toStart:
            self.startSlave(t)
            if t.visibleFilter == False:
                # Set shown timestamp to zero such that this task will be stopped in the next call
                # to this function
                t.showTimestamp = 0
            else:
                # Set shown timestamp to None such that this task is not stopped before it has been shown
                t.showTimestamp = None
        
        
    def getLeastRecentlyShownTasks(self, n):
        '''
        Return the n least recently shown currently running tasks
        
        Note that currently visible tasks are not considered since these all are the most
        recently shown tasks. In addition tasks that have been started but not shown are also
        not considered (these tasks have showTimestamp=None)
        
        @param n: number of tasks to retrun
        @return: list with the n least recently shown tasks. Note that fewer than n tasks may be returned
        if there are fewer than n currently running hidden tasks
        '''
        
        self.logger.debug("Stop the %d least recently shown tasks" % (n))
        
        timestampTask = [] # List with (timestamp, task) tuples
        for t in self.currentlyRunning:
            if t not in self.currentlyVisible and t.showTimestamp != None:
                timestampTask.append( (t.showTimestamp, t) )
        
        if n > len(timestampTask):
            self.logger.warn('Cannot return %d least recently shown tasks since only %d are running and not visible nor un-shown' % (n, len(timestampTask)))
            n = len(timestampTask) 
            
        # Sort the task according to their timestamp
        timestampTask.sort()
        
        # The oldest timestamps are now first
        rv = []
        for i in range(n):
            rv.append(timestampTask[i][1])
        return rv
      
    def executeVisibleFilterTask(self, task):
        '''
        Helper function to execute the visible filter function on a single task.
        
        @param task: currently running task to execute filter for
        @param filterFunction: filter function to run
        @param filterArg: argument passed to filter function
        @return: None, but the tasks visibleFilter variable will be set
        '''
        
        if not task.isRunning():
            raise CoordinatorError('Task is not running: %s' % (task.taskID))
        
        filterFunction, filterArg = task.pendingVisibleFilter
        
        try:
            self.sendFilterMsg(task.pipe, filterFunction, filterArg)
        except SlaveError:
            self.logger.warn('Task has been killed: %s' % (task.taskID))
            self.killSlave(task, True)
            return
            
        rv = task.pipe.recv()
        if rv == False:
            task.visibleFilter = False
            if task in self.currentlyVisible:                
                self.hideTask(task)
            # else: task is already hidden
        else:
            task.visibleFilter = True
        # Clear pending function
        task.pendingVisibleFilter = None
    
    def executeFunctionAll(self, functionType, functionReference=None, functionArgs=None):
        '''
        Helper function to execute the filter function on all tasks, even if they are not currently running.
        
        @param funtionType: one of 'filter', 'gather', or 'readLogFiles'
        @param filterFunction: optional, filter function to run. Required for 'filter' and 'gather'
        @param filterArgs: optional, arguments passed to filter function
        @return: dictionary mapping task (objects) to return values
        '''
        
        if functionType not in ['filter', 'gather', 'close', 'readLogFiles']:
            raise CoordinatorError("Invalid function type to execute on all tasks: %s" % (functionType))
        
        # Save return in a hash map that maps task object to return values
        task2rv = {}
        
        # Note the filter function must be run on all task
        failedTasks = []
        while 1:
            # Execute filter function on currently running slaves. The return values are gathered below            
            for task in self.currentlyRunning:
                if task in task2rv:
                    # Already received value in previous iteration
                    continue
                
                self.logger.debug('Execute function on: %s' % (task.taskID))
                try:
                    if functionArgs != None:
                        if functionType == 'filter':
                            self.sendFilterMsg(task.pipe, functionReference, functionArgs[task.taskID])
                        elif functionType == 'gather':
                            self.sendGatherMsg(task.pipe, functionReference, functionArgs[task.taskID])
                        elif functionType == 'close':
                            self.sendCloseMsg(task.pipe)
                        elif functionType == 'readLogFiles':
                            self.sendReadLogFilesMsg(task.pipe)
                    else:                     
                        if functionType == 'filter':
                            self.sendFilterMsg(task.pipe, functionReference, None)
                        elif functionType == 'gather':
                            self.sendGatherMsg(task.pipe, functionReference, None)
                        elif functionType == 'close':
                            self.sendCloseMsg(task.pipe)
                        elif functionType == 'readLogFiles':
                            self.sendReadLogFilesMsg(task.pipe)
                        
                except SlaveError:
                    self.logger.warn("Task was killed by user: %s" % (task.taskID))
                    failedTasks.append(task)
            
            self.logger.info("Wait for reply values from %d tasks" % (len(self.currentlyRunning)))
                        
            # Gather return values                          
            for task in self.currentlyRunning:
                if task in failedTasks:
                    task2rv[task] = None
                elif task in task2rv: 
                    # Already received value in previous iteration
                    continue
                else:
                    rv = task.pipe.recv()
                    task2rv[task] = rv
                
            # If needed start a new set of slaves
            if len(task2rv) < len(self.taskList):
                nToStart = len(self.taskList) - len(task2rv)                                
                if nToStart > self.nSlaves:
                    nToStart = self.nSlaves
                if nToStart == 0:
                    self.logger.warn("No more task that can be started")
                    break
                    
                # Kill running slaves before starting new ones
                nToKill = nToStart - (self.nSlaves - len(self.currentlyRunning))
                for i in range(nToKill):
                    # Always pick the first element in the currentlyRunning list since elements are removed in killSlave()
                    task = self.currentlyRunning[0]
                    self.killSlave(task)
                    
                # Select new slaves to run
                toStart = []
                for task in self.taskList:
                    if task not in task2rv:
                        toStart.append(task)
                    if task.filterKill:
                        # Do not start closed tasks, but count them as started and set return value to None
                        task2rv[task] = None
                        nToStart -= 1
                        
                    if len(toStart) == nToStart:
                        break                
                    
                # Start new slaves
                for task in toStart:
                    self.startSlave(task)
            else:
                # All tasks has received filter function
                break
        
        # Cleanup data structures for failed tasks
        self.logger.warn("%d tasks failed" % (len(failedTasks)))
        for f in failedTasks:
            self.killSlave(f, True)
            
        return task2rv  
    
    def consoleDriver(self):
        '''
        Simple console driver that reads window navigation input from the console
        '''
        while 1:
            inputStr = raw_input("(N)ext, (P)rev, (R)andom, (F)irst, (I)ndex(Number), (5) Refresh, (E)xit: ")
            inputStr.lower()
            
            if inputStr == 'e' or inputStr == 'exit':
                break
            elif inputStr == '5':
                self.resetVisible()
            elif inputStr == 'n' or inputStr == 'next':
                self.showNext()
            elif inputStr == 'p' or inputStr == 'prev':
                self.showPrevious()
            elif inputStr == 'r' or inputStr == 'random':
                self.showRandom()
            elif inputStr == 'f' or inputStr == 'first':
                self.showFirst()
            elif inputStr.startswith('index'):
                indexNum = int(inputStr[1:])
                self.showIndex(indexNum)
            elif inputStr.startswith('i'):
                indexNum = int(inputStr[5:])
                self.showIndex(indexNum)
                
    def consoleBenchamrk(self):
        '''
        Simple console driver that reads window navigation input from the console
        '''
        for i in range(15):
            time.sleep(3)
            print time.clock()
            print time.time()
            self.showNext()
            
                
                
            