'''
PIL/Pyglet based Mandelbrot animation

PIL is used to hold the pixel data
Pyglet to show the data and do window management

Based on: https://gist.github.com/2860171 by rsheldiii 
'''

from __future__ import division

class MandelbrotScreen:
    def __init__(self, width, height, nZooms):
        self.width = width
        self.height = height
        self.nZooms = nZooms
        
        # Initialized in createWindow
        self.window = None

def vizAnimatedMandelbrot(stateObject, logger, zoomXY):
    '''
    Run Mandelbrot animation
        
    @param stateObject: state holder object that includes window handle
    @param logger: initialized logger 
    @param zoomXY: tuple with x and y position for zoom area    
    '''    
    
    import pyglet
    from PIL import Image, ImageDraw

    # Get arguments for this task
    width = stateObject.width
    height= stateObject.height
    nZooms = stateObject.nZooms
    zoomX, zoomY = zoomXY
    
    def mandelbrot(width, height, xoffset, yoffset, xzoom, yzoom, max_iterations=100, color = (255, 255, 255)):
        '''
        Do the mandelbrot calculation
        
        @param width: width of image to compute
        @param height: height of image to compute
        @param xoffset: ?
        @param yoffset: ?
        @param xzoom: ?
        @param yzoom: ?
        @param max_iteration: optional, maximum number of iterations (default 100)
        @param color: optional, color to use for the mandelbrot visualization (default white)
        @return: PIL Image with Mandelbrot pixels
        '''
        
        # Create new set
        image = Image.new("RGB", (width, height))
        # Calculate each pixel
        for x in range(0, width):                        
            for y in range(0,height):
    
                iterations = 0
                i = 0
                j = 0
                
                xscaled = (x / width) * xzoom + xoffset
                yscaled = (y  /height) * yzoom + yoffset
                                
                while ( i*i + j*j < 4  and  iterations < max_iterations ):
                    xtemp = i*i - j*j + xscaled
                    j = 2*i*j + yscaled
                    i = xtemp
                    iterations +=1
                    
                weight = iterations / max_iterations
                image.putpixel((x, y), (int(weight*color[0]), int(weight*color[1]), int(weight*color[2])))             
        return image
    
    #
    # Setup various parameters 
    #        
    xoffset = -2.5
    yoffset = -1
    xlength =  3.5    
    ylength =  2  
    zoom = 1
    
    # Zoom area
    leftx = zoomX - int(width/4)
    rightx = zoomX + int(width/4)
    topy = zoomY - int(height/4)
    bottomy = zoomY + int(height/4)    
             
    #
    # Do animation
    #
    nDraws = 0
    
    while nDraws < nZooms:
        # Do calculation
        imagePixels = mandelbrot(width, height, xoffset, yoffset, xlength / zoom, ylength / zoom, 50)
        draw = ImageDraw.Draw(imagePixels)
        
        # Paint set
        #pygame.surfarray.blit_array(stateObject.screen, mandelbrotset)
        
        # Set parameters for next zoom                  
        xoffset += (zoomX / width) * xlength / zoom - xlength / zoom / 4
        yoffset += (zoomY / height) * ylength / zoom - ylength / zoom / 4                                    
        zoom *=2                
                
        # Draw rectangle around zoom area    
        #pyglet.graphics.draw(stateObject.window, (255, 0, 0), (leftx, topy), (rightx, topy))
        #pyglet.graphics.draw(stateObject.window, (255, 0, 0), (leftx, bottomy), (rightx, bottomy))
        #pyglet.graphics.draw(stateObject.window, (255, 0, 0), (rightx, topy), (rightx, bottomy))
        #pyglet.graphics.draw(stateObject.window, (255, 0, 0), (leftx, topy), (leftx, bottomy))
        draw.rectangle((leftx, topy, rightx, bottomy), None, (255, 0, 0))                                          
        
        imagePixels.save("mandel.png", "PNG")
        
        image = pyglet.image.ImageData(width, height, 'RGB', imagePixels.tostring(), pitch=-50*3)
        sprite = pyglet.sprite.Sprite(image)
        #sprite.draw()
        # Update screen
        #pygame.display.flip()
        nDraws += 1              
        print nDraws
        
        
    print 'Done: close window to quit'
                         

def runMandelbrot(width, height, nZooms, nTasks, nWindows, nProcs):
    '''
    Start Mandelbrot animation
    
    @param width:  window width
    @param height: window height
    @param nZooms: number of zooms to perform
    @param nTasks: number of tasks to start. Each will do the same amount of work, but the zoom area differs
    @param nWindows: number of windows to show simustaneously
    @param nProcs: number of processes to run simustaneously
    '''
    
    from bealjas.coordinator.windowCoordinator import WindowCoordinator
    from bealjas.utils.bealjas2pyglet import createWindow, guiLoop, closeWindow, setWindowPosSize, getWindowPosSize, setWindowVisible
    
    # Initialize state object and select random area to zoom in for each task
    stateObject = MandelbrotScreen(width, height, nZooms)
    import random
    argvs = []
    for i in range(nTasks):
        x = random.randint(width / 4,  3 * (width / 4))
        y = random.randint(height / 4, 3 * (height / 4))
        argvs.append((x, y))
    
    # Initialize bealjas coordinator and use it to start the animation on each node
    coordinator = WindowCoordinator(nWindows, nProcs, createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize, stateObject)
    coordinator.startSlaves(vizAnimatedMandelbrot, argvs)
    coordinator.consoleDriver()
    coordinator.killSlaves()
        
if __name__ == '__main__':
    import sys
    
    assert(len(sys.argv) == 7), 'Usage: xsize ysize iterations nTask nWindows nProcs'
    xsize = int(sys.argv[1])
    ysize = int(sys.argv[2])
    iterations = int(sys.argv[3])
    nTasks = int(sys.argv[4])
    nWindows = int(sys.argv[5])
    nProcs = int(sys.argv[6])
    
    runMandelbrot(xsize, ysize, iterations, nTasks, nWindows, nProcs)