'''
Visualization of the results of the GeoGSMOverlap results.
'''
import sys, os

"""
This class represents a dataset/series in the overlap graph
"""
class Node:
    """
    @param name: node name
    @param nSamples: number of samples in node
    @param date: date given as a string
    @param organisms: list of organisms
    """
    def __init__(self, name, nSamples, date, organisms):
        self.name = name
        self.nSamples = nSamples
        self.date = date
        self.organisms = organisms
        
"""
This class represents overlapping samples between two dataset/series in the overlap graph
"""
class Edge:
    """
    @param node1: reference to node
    @param node2: reference to another node
    @param samples: list of overlapping samples
    """
    def __init__(self, node1, node2, samples):
        self.node1 = node1
        self.node2 = node2
        self.nOverlap = len(samples)
        
"""
This class represents a set of dataset/series with overlapping samples.
"""
class Cluster:
    """
    @param ids: dataset ids of cluster nodes
    @param overlap: overlap data structure holder
    """
    def __init__(self, ids, overlap):
        # list of nodes
        self.nodes = []
        
        # list of edges (string of the form 'gid1,gid2')
        self.edges = []
        
        for i in ids:
            self.nodes.append(overlap.getNode(i))
            for j in ids:
                if i == j:
                    continue                
               
                key = None
                if overlap.compareIDs(i, j) < 0:
                    key = i + ',' + j                    
                elif overlap.compareIDs(i, j) > 0:
                    key = j + ',' + i
                else:
                    print i, j
                    raise 'Not possible'
                
                if key not in overlap.edges:
                    # Edge does not exist
                    continue
                
                if key not in self.edges:
                    self.edges.append(key)
                
        
"""
Sample overlap data structures
"""
class Overlap:
    """
    Constructor
    
    @param overlapFile: overlap file used as input to the FindGSMOverlap tool
    @param overlapLogFile: log file produced by the FindGSMOverlap tool
    """
    def __init__(self, overlapFile, overlapLogFile):
        # Hashmap that maps dataset/series IDs to Node objects
        self.nodes = {}
        # Hashmap that maps dataset/series IDs to Edge objects (key is: gid1,gid2)
        self.edges = {}
        # List of cluster objects before and after overlap removal
        self.clustersBefore = []
        self.clustersAfter = []
        # List of GEO ids of removed duplicates
        self.duplicates = []
        # List of GEO ids of series/datasets removed due to too few samplse
        self.removedSmall = []
        # List of GEO ids of removed supersets
        self.supersets = []
        # Hashmap of removed samples (key: gid, value: list of sample IDs)
        self.removedSamples = {}
        
        self.readOverlapFile(overlapFile)
        self.readLogFile(overlapLogFile)
        self.sortClusters(self.clustersBefore)
        self.sortClusters(self.clustersAfter)


    """
    Read overlapping pairs file produced by GSMOverlap. This file contains information about the nodes.
    
    The line format for the input file is:
      GID_i<tab>GID_j<tab>overlap count,GID_i sample count,GID_j sample count<tab>[overlapping samples]<tab>meta_i<tab>meta_j<newline>
       
      where [overlapping samples] is the list of overlapping samples in the following format
        GSM_1,GSM_2,..,GSM_N<tab>
         
      and meta_i contains meta data read from the GEO meta data table for GID_i
        updateDate<tab>Organism1,Organism2...OrganismN
    
    @param filename: file to read from.
    """
    def readOverlapFile(self, filename):
        fp = open(filename)
        
        while 1:
            line = fp.readline()
            if line == '': # EOF
                break
            
            #
            # Parse line (format is described above)
            #
            parts = line.split('\t')
            if len(parts) != 8:
                print 'Invalid line: ' + line
                continue
            
            gid1 = parts[0]
            gid2 = parts[1]
            
            if self.compareIDs(gid1, gid2) > 0:
                # Ignore duplicate pairs
                continue
            
            subParts = parts[2].split(',')
            #nOverlapping = int(subParts[0])
            nSamples1 = int(subParts[1])
            nSamples2 = int(subParts[2])
            
            overlappingSamples = parts[3].split(',')
            
            date1 = parts[4]
            organisms1 = parts[5].split(',')
            date2 = parts[6]
            organisms2 = parts[7].split(',')
            
            #
            # Create one node for each dataset if the node has not already been created
            #
            if gid1 not in self.nodes:
                newNode = Node(gid1, nSamples1, date1, organisms1)
                self.nodes[gid1] = newNode
            if gid2 not in self.nodes:
                newNode = Node(gid2, nSamples2, date2, organisms2)
                self.nodes[gid2] = newNode
                
            #
            # Always add link
            #
            self.edges[gid1 + ',' + gid2] = Edge(self.nodes[gid1], self.nodes[gid2], overlappingSamples)
        
        fp.close()
        print 'Overlap file read'
    
    """
    Read log file produced by GeoGSMOverlap
    
    @param filename: file to read from.
    """
    def readLogFile(self, filename):
        fp = open(filename)
            
        while 1:
            line = fp.readline()
            if line == '':
                break
            
            #
            # Find start of before clusters
            #
            if line.find('start: clusters before overlap') == 0:
                while 1:
                    line = fp.readline().strip()
                    assert(line != ''), 'Unexpected EOF'
                    if line.find('end: clusters before overlap') == 0:
                        break
                    
                    ids = line.split('\t')
                    newCluster = Cluster(ids, self)
                    self.clustersBefore.append(newCluster)
                print 'Before clusters read'
                    
            elif line.find('start: duplicates removed') == 0:
                while 1:
                    line = fp.readline().strip()
                    assert(line != ''), 'Unexpected EOF'
                    if line.find('end: duplicates removed') == 0:
                        break
                    
                    self.duplicates.append(line)
                print 'Duplicates read'
                    
            
            elif line.find('start: small datasets removed') == 0:
                while 1:
                    line = fp.readline().strip()
                    assert(line != ''), 'Unexpected EOF'
                    if line.find('end: small datasets removed') == 0:
                        break
                    
                    self.removedSmall.append(line)
                print 'Removed small datasets read'
            
            elif line.find('start: supersets removed') == 0:
                while 1:
                    line = fp.readline().strip()
                    assert(line != ''), 'Unexpected EOF'
                    if line.find('end: supersets removed') == 0:
                        break
                    
                    self.supersets.append(line)
                print 'Supersets read'
            
            elif line.find('start: overlap removed') == 0:
                while 1:
                    line = fp.readline().strip()
                    assert(line != ''), 'Unexpected EOF'
                    if line.find('end: overlap removed') == 0:
                        break
                    
                    self.removedSmall.append(line)
                print 'Overlap removed read'
            
            elif line.find('start: clusters after overlap') == 0:
                while 1:
                    line = fp.readline().strip()
                    assert(line != ''), 'Unexpected EOF'
                    if line.find('end: clusters after overlap') == 0:
                        break
                    
                    ids = line.split('\t')
                    newCluster = Cluster(ids, self)
                    self.clustersAfter.append(newCluster)
                print 'Clusters after read'
                    
            
            elif line.find('start: samples removed') == 0:
                while 1:
                    line = fp.readline().strip()
                    assert(line != ''), 'Unexpected EOF'
                    if line.find('end: samples removed') == 0:
                        break
                    
                    parts = line.split('\t')
                    gid = parts[0]
                    # parts[1] and parts[2] are respectively 'remove' and '(count)'
                    sids = parts[3:]
                    self.removedSamples[gid] = sids
                print 'Removed samples read'
        
        fp.close()
        print 'Log file read'
        
    """
    Get a node reference
    
    @param gid: dataset/series ID
    @return corresponding node, or None if a node could not be found
    """
    def getNode(self, gid):
        return self.nodes[gid]
            
    
    
    """
    Helper function to compare two dataset/series IDs
     
    @param gid1 ID that starts with either GSE or GDS and that may contain a "-" for subsets
    @param gid2 ID of the same type as gid1
    @return -1 if gid1 < gid2, 0 if gid1 == gid2, and 1 if gid1 > gid2
    """
    def compareIDs(self, gid1, gid2):
        gid1 = gid1.replace("GSE", "")
        gid1 = gid1.replace("GDS", "")
        gid1 = gid1.replace("GPL", "")
        gid1Parts = gid1.split('-')
        gidn1 = int(gid1Parts[0])
        
        gid2 = gid2.replace("GSE", "")
        gid2 = gid2.replace("GDS", "")
        gid2 = gid2.replace("GPL", "")
        gid2Parts = gid2.split('-')
        gidn2 = int(gid2Parts[0])
        
        if gidn1 < gidn2:
            return -1
        elif gidn1 > gidn2:
            return 1
        else: 
            if len(gid1Parts) == 2 and len(gid2Parts) == 2:
                gpl1 = int(gid1Parts[1])
                gpl2 = int(gid2Parts[1])
                if gpl1 < gpl2:
                    return -1
                elif gpl1 > gpl2:
                    return 1
                else:
                    return 0
            else:
                return 0
    
    """
    Sort clusters by size, largest to smallest
    """
    def sortClusters(self, clusterList):
        # Create list with (size, cluster) tuples
        sizeRefList = []
        for c in clusterList:
            sizeRefList.append((len(c.nodes), c))
            
        # Sort list on size
        sizeRefList.sort(reverse=True)
        
        # Replace input list references with sorted references
        for i in range(len(clusterList)):        
            clusterList[i] = sizeRefList[i][1] 
            
    """
    Check if dataset is deleted
    
    @param gid: GEO dataset or series ID
    @return: True if dataset is either a duplicate, superset, or has too few samples.
    """
    def isDeleted(self, gid):
        if gid in self.duplicates or gid in self.supersets or gid in self.removedSmall:
            return True
        else:
            return False
    
    """
    Return a dictionary with the default style to use for the graph nodes
    
    @return: dict with styles
    """
    def getDefaultStyles(self):
        styles = {'duplicate': 'invis', 
                  'superset': 'bold',
                  'tooSmall': 'dotted',
                  'notDeleted': 'filled',
                  'deletedEdge': 'dotted'}
        
        return styles
    
    """
    Return a dictionary with the default colors to use for the graphs
    
    @return: dict with colors
    """
    def getDefaultColors(self):
        colors = {'superset': 'red',
                  'tooSmall': 'yellow',
                  'notDeleted': 'white',    
                  'samplesDeleted': 'green',
                  'duplicate': 'blue'}
        
        return colors
    
    """
    Get a style for a cluster node
    
    @param name: name of node to check
    @param styles: node styles to use for the graphs
    @return: the style to use
    """ 
    def getStyle(self, name, styles):
        style = styles['notDeleted']
        if name in self.duplicates:
            style = styles['duplicate']
        elif name in self.supersets:
            style = styles['superset']
        elif name in self.removedSmall:
            style = styles['tooSmall']
        elif name in self.removedSamples:
            style = styles['notDeleted']
            
        return style
    
    """
    Get a color for a cluster node
    
    @param name: name of node to check    
    @param colors: node colors to use for the graphs
    @return: the style to use
    """
    def getColor(self, name, colors):
        color = colors['notDeleted']
        if name in self.duplicates:
            color = colors['duplicate']
        elif name in self.supersets:
            color = colors['superset']
        elif name in self.removedSmall:        
            color = colors['tooSmall']
        elif name in self.removedSamples:        
            color = colors['samplesDeleted']
        return color

"""
Use dot to create a PNG file with the graph for each cluster

@param clusters: clusters to plot
@param overlap: overlap data stucture holder object
@param outputDir: directory where output files are written
@param styles: node styles to use for the graphs
@param colors: node colors to use for the graphs
"""
def createPNGs(clusters, overlap, outputDir, styles, colors):
    cnt = 0
    for c in clusters:
        name = 'before-%d' % (cnt)
        cnt += 1
        
        createPNG(c, overlap, os.path.join(outputDir, name + '.png'), styles, colors)        
        
        
"""
Plot each cluster in maptlotlib

@param clusters: clusters to plot
@param overlap: overlap data stucture holder object
@param styles: node styles to use for the graphs
@param colors: node colors to use for the graphs
"""
def createPlots(clusters, overlap, styles, colors):
    for c in clusters:
        createPlot(c, overlap, styles, colors)
        raw_input("Press enter to continue")
        
"""
Create a dot file for each cluster

@param clusters: clusters to plot
@param overlap: overlap data stucture holder object
@param outputDir: directory where output files are written
@param styles: node styles to use for the graphs
@param colors: node colors to use for the graphs
"""
def createDots(clusters, overlap, outputDir, styles, colors):
    cnt = 0
    for c in clusters:
        name = 'before-%d' % (cnt)
        cnt += 1
        
        createDot(c, overlap, os.path.join(outputDir, name + '.dot'), styles, colors)

        
"""
Use dot to create a PNG file for a cluster

@param clusters: cluster to plot
@param overlap: overlap data stucture holder object
@param outputFilename: output file
@param styles: node styles to use for the graphs
@param colors: node colors to use for the graphs
"""
def createPNG(cluster, overlap, outputFilename, styles, colors):
    import pydot
    G = pydot.Dot(graph_type='graph')
    for n in cluster.nodes:
        style = overlap.getStyle(n.name, styles)
        color = overlap.getColor(n.name, colors)
        
        label = n.name
        if n in overlap.removedSamples:
            label = '%s (-%d)' % (n.name, len(overlap.removedSamples[n]))
                        
        newNode = pydot.Node(n.name.replace('-', '_'), label=label, style=style, color=color)
        G.add_node(newNode)
        
    for e in cluster.edges:
        n1, n2 = e.replace('-', '_').split(',')
        if e not in overlap.edges:
            print 'ERROR: Edge not found: %s' % (e)
            continue                
        edge = overlap.edges[e]            
        newEdge = None
        if overlap.isDeleted(n1) or overlap.isDeleted(n2):
            newEdge = pydot.Edge(n1, n2, label=str(edge.nOverlap), style=styles['deletedEdge'])
        else:
            newEdge = pydot.Edge(n1, n2, label=str(edge.nOverlap))
        G.add_edge(newEdge)
        
    #try:
    G.write_png(outputFilename)
    #except pydot.InvocationException:
    #    print 'Warning: failed to create cluster: %s' % (outputFilename) 
    
    
"""
Create a dot file for a cluster

@param clusters: cluster to plot
@param overlap: overlap data stucture holder object
@param outputFilename: output file
@param styles: node styles to use for the graphs
@param colors: node colors to use for the graphs
"""
def createDot(cluster, overlap, outputFilename, styles, colors):
    #
    # Create PNG file
    #
    fp = open(outputFilename, 'w')
    fp.write('Graph {\n')    
    for n in cluster.nodes:
        style = overlap.getStyle(n.name, styles)
        color = overlap.getColor(n.name, colors)
        
        label = n.name
        nodeName = n.name.replace('-', '_')
        if n in overlap.removedSamples:
            label = '%s (-%d)' % (n.name, len(overlap.removedSamples[n]))
                        
        fp.write('\t%s [label="%s", style="%s", color="%s"];\n' % (nodeName, label, style, color))        
        
    for e in cluster.edges:
        n1, n2 = e.replace('-', '_').split(',')
        if e not in overlap.edges:
            print 'ERROR: Edge not found: %s' % (e)
            continue                
        edge = overlap.edges[e]        
        if overlap.isDeleted(n1) or overlap.isDeleted(n2):
            fp.write('\t%s -- %s [label="%d", style="%s"];\n' % (n1, n2, edge.nOverlap, styles['deletedEdge']))   
        else:
            fp.write('\t%s -- %s [label="%d"];\n' % (n1, n2, edge.nOverlap))            
        
    fp.write('}\n')
    fp.close()    

"""
Use matplotlib to plot a graph for a cluster

@param clusters: clusters to plot
@param overlap: overlap data structure holder object
@param styles: node styles to use for the graphs
@param colors: node colors to use for the graphs
"""       
def createPlot(cluster, overlap, styles, colors):
    #
    # Plot
    #
    import networkx as nx
    import matplotlib.pyplot as plt
    
    # Clear current figure
    plt.clf()
    
    G = nx.cycle_graph(len(cluster.nodes))
    for n in cluster.nodes:
        G.add_node(n.name.replace('-', '_'))
    pos = nx.spring_layout(G, iterations=200)
    
    for n in cluster.nodes:
        style = overlap.getStyle(n.name, styles)
        color = overlap.getColor(n.name, colors)
        label = n.name
        if n in overlap.removedSamples:
            label = '%s (-%d)' % (n.name, len(overlap.removedSamples[n]))                            
        nx.draw_networkx_nodes(G, pos, node_list=[n.name], node_color=color, style=style)
    for e in cluster.edges:            
        n1, n2 = e.replace('-', '_').split(',')            
        if e not in overlap.edges:
            print 'Edge not found: %s' % (e)
            raise 'Runtime exception'
        edge = overlap.edges[e]            
        if overlap.isDeleted(n1) or overlap.isDeleted(n2):              
            G.add_edge(n1, n2)
            nx.draw_networkx_edges(G, pos, edgelist=[(n1, n2)], style='dotted')
            nx.draw_networkx_edge_labels(G, pos, edge_labels={(n1, n2): str(edge.nOverlap)})
        else:
            #print 'Add edge (%s, %s)' % (n1, n2)
            G.add_edge(n1, n2)
            #print 'Add edges'
            nx.draw_networkx_edges(G, pos, edgelist=[(n1, n2)])
            #print 'Add labels'
            nx.draw_networkx_edge_labels(G, pos, edge_labels={(n1, n2): str(edge.nOverlap)})
            
    plt.show()
    
def plotTest():
    o = Overlap(r'C:\Users\larsab\troilkatt\overlap2\overlap.pairs', r'C:\Users\larsab\troilkatt\overlap2\overlap.log')
    createPlots(o.clustersBefore, o, o.getDefaultStyles(), o.getDefaultColors())
    
#
# Beljas application provided functions
#
def vizClustersBefore(overlap, logger, ci):
    '''
    This function is called to visualize a graph.
    
    Note! this function should not block
        
    @param overlap: Overlap object
    @param logger: plotter logger
    @param ci: cluster index, in the overlap objects clustersBefore array
    @return: None    
    '''
    import time
    print 'P2: ', time.time()
    styles = overlap.getDefaultStyles()
    colors = overlap.getDefaultColors()
    cluster = overlap.clustersBefore[ci]
    
    #
    # Plot
    #
    import networkx as nx
    import matplotlib.pyplot as plt
    
    # Clear current figure
    #plt.clf()
    
    G = nx.Graph()    
    
    # First add nodes
    for n in cluster.nodes:
        G.add_node(n.name.replace('-', '_'))
    
    # and calculate position of nodes
    pos = nx.spring_layout(G, iterations=200)
    
    # Then edges
    for e in cluster.edges:            
        n1, n2 = e.replace('-', '_').split(',')
        if e not in overlap.edges:
            print 'Edge not found: %s' % (e)
            raise 'Runtime exception'                    
        G.add_edge(n1, n2)      
        
    # Edge styles
    for e in cluster.edges:            
        n1, n2 = e.replace('-', '_').split(',')
        edge = overlap.edges[e]            
        if overlap.isDeleted(n1) or overlap.isDeleted(n2):                          
            nx.draw_networkx_edges(G, pos, edgelist=[(n1, n2)], style='dotted')
            nx.draw_networkx_edge_labels(G, pos, edge_labels={(n1, n2): str(edge.nOverlap)})
        else:    
            nx.draw_networkx_edges(G, pos, edgelist=[(n1, n2)])
            nx.draw_networkx_edge_labels(G, pos, edge_labels={(n1, n2): str(edge.nOverlap)})

    # and node styles
    nx.draw(G, pos, node_color=[overlap.getColor(v.replace('_', '-'), colors) for v in G], size=3000)
            
    print 'P3: ', time.time()
    plt.draw()
    
def vizClustersAfter(overlap, logger, ci):
    '''
    This function is called to visualize a graph.
    
    Note! this function should not block
        
    @param overlap: Overlap object
    @param logger: plotter logger
    @param ci: cluster index, in the overlap objects clustersBefore array
    @return: None
    '''
    
    #styles = getDefaultStyles()
    colors = overlap.getDefaultColors()   
    
    cluster = overlap.clustersAfter[ci]
    
    #
    # Plot
    #
    import networkx as nx
    import matplotlib.pyplot as plt
    
    # Clear current figure
    #plt.clf()
    
    G = nx.Graph()    
    
    # First add nodes
    for n in cluster.nodes:
        G.add_node(n.name.replace('-', '_'))
    
    # and calculate position of nodes
    pos = nx.spring_layout(G, iterations=200)
    
    # Then edges
    for e in cluster.edges:            
        n1, n2 = e.replace('-', '_').split(',')
        if e not in overlap.edges:
            print 'Edge not found: %s' % (e)
            raise 'Runtime exception'                    
        G.add_edge(n1, n2)      
        
    # Edge styles
    for e in cluster.edges:            
        n1, n2 = e.replace('-', '_').split(',')
        edge = overlap.edges[e]            
        if overlap.isDeleted(n1) or overlap.isDeleted(n2):                          
            nx.draw_networkx_edges(G, pos, edgelist=[(n1, n2)], style='dotted')
            nx.draw_networkx_edge_labels(G, pos, edge_labels={(n1, n2): str(edge.nOverlap)})
        else:    
            nx.draw_networkx_edges(G, pos, edgelist=[(n1, n2)])
            nx.draw_networkx_edge_labels(G, pos, edge_labels={(n1, n2): str(edge.nOverlap)})

    # and node styles
    nx.draw(G, pos, node_color=[overlap.getColor(v.replace('_', '-'), colors) for v in G], size=3000)
            
    plt.draw()
    
def filterSuperset(overlap, logger, ci):
    '''
    This function filters clusters based on whether they have a supersets or not
    
    @param overlap: Overlap object
    @param logger: plotter logger
    @param ci: the cluster index in the overlap objects clustersBefore array
    @return True if this cluster has a superset node, and False if not
    '''
    cluster = overlap.clustersBefore[ci]
    
    for n in cluster.nodes:                
        if n.name in overlap.supersets:        
            return True
        
    return False

def filterDuplicates(overlap, logger, ci):
    '''
    This function filters clusters based on whether they have duplicates or not
    
    @param overlap: Overlap object
    @param logger: plotter logger
    @param ci: the cluster index in the overlap objects clustersBefore array
    @return True if this cluster has a superset node, and False if not
    '''
    cluster = overlap.clustersBefore[ci]
    
    for n in cluster.nodes:                
        if n.name in overlap.duplicates:        
            return True
        
    return False

#
# Drivers
#
def executeBealjas(overlapFile, overlapLogFile, nProcs, nWindows):
    # Debug arguments: 
    # overlapFile = "C:\\Users\\larsab\\troilkatt\\overlap2\\overlap.pairs"
    # overlapLogFile = "C:\\Users\\larsab\\troilkatt\\overlap2\\overlap.log"
    clusters = Overlap(overlapFile, overlapLogFile)
    from bealjas.coordinator.windowCoordinator import WindowCoordinator
    from bealjas.utils.bealjas2pyplot import createWindow, guiLoop, closeWindow, setWindowPosSize, getWindowPosSize, setWindowVisible
    nTasks = len(clusters.clustersBefore)    
    print 'FOUND ---%d--- CLUSTERS' % (nTasks)
    coordinator = WindowCoordinator(nWindows, nProcs, createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize, clusters)
    coordinator.startSlaves(vizClustersBefore, range(nTasks))    
    #coordinator.pendingFilterVisible(filterColor, argTuples)
    coordinator.consoleDriver()
    coordinator.benchamrkDriver()
    #coordinator.closeWindows()
    #coordinator.killSlaves()
    return coordinator
    
if __name__ == '__main__':
    assert(len(sys.argv) == 5), 'Usage: %s overlapFile overlap.log nProcs nWindows' % (sys.argv[0])    
    coordinator = executeBealjas(sys.argv[1], sys.argv[2], int(sys.argv[3]), int(sys.argv[4]))    
    coordinator.killSlaves()
    print 'Done'
    