'''
Cluster communication over sockets

The communication protocol is simple and only consists of two message types:
    
    Initialization:
    1. Client -> server: client ID
    2. Server -> client: (plotsPerTile, slavesPerTile)
    
    Messages:
    1. Server -> client: pickled tuple (that contains the message type and arguments)
    2. Client -> server: pickled 'ack' string, or return value
    
The usage of the server and client is equally simple:

    Initializtion:
    - the server calls runServer() to start a new thread that handles communication with all clients
    - the client calls runClient() to start a new thread that handles communication with the server
    - the server calls killSerevr() to kill the server thread
    - the client calls killClient() to kill the client communication thread
    
    For communication between a master coordinator and a tile coordinator the following steps are done:
    1. The master puts a tuple to send to the tile coordinator on the servers request queue    
    2. The server gets the tuple from the queue
    3. The server pickles the tuple and sends it over a socket to the client
    4. The clients unpickles the tuple and puts it to its request queue
    5. The tile coordinator reads the tuple from its request queue, handle the request and possibly
       puts a return value on its return value queue
    6. The client may pickle and send a return value to the server, or just send an ack message if the
       command does not return any values 
    7. The server unpickles the return value (or ack) and puts any return value on its return value queue.
       If there are only ack messages it does not put a message on the return value queue.
    8. The master coordinator gets the return values from the return value queue (if any)

    The tuple format is: (clientID, (data)), where clientID is either an ID or "all" if the data
    should be sent to all clients. 
'''

import threading, socket, pickle
import sys, struct
import Queue
from networkError import NetworkError

MAX_RECV_SIZE = 8192

class BealjasSocketServer(threading.Thread):
    '''
    Server-side thread that handles communication with client coordinators. 
    '''
    
    def __init__(self, nClients, requestQueue, rvQueue, host='localhost', port=31278):
        '''
        Constructor
        
        @param nClients: number of clients to wait for
        @param requestQueue: queue where to recevie requests tuples and post
        @param rvQueue: queue where to send return values
        @param host: optional host to bind listen socket to. Default is localhost.
        @param port: optional port to listen for connections. Default is 31278.
        '''
        
        threading.Thread.__init__(self)
        
        # Number of clients that will connect to this server
        self.nClients = nClients
        
        # Client id to socket mapping
        self.id2socket = {}
        
        # Coordinator -> server messages
        self.requestQueue = requestQueue
        # Coordinator <- server messages
        self.rvQueue = rvQueue
        
        # Host and port to listen on
        self.host = host
        self.port = port
        
    def run(self):
        '''
        Server's main loop
        '''        
        
        # List of client ids
        clientIDs = []
         
        #
        # Wait for clients to connect
        #
        print '[Network] Wait for %d clients to connect to %s:%d:' % (self.nClients, self.host, self.port)
        listenSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        listenSocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        listenSocket.bind((self.host, self.port))
        for i in range(self.nClients):
            listenSocket.listen(1)
            sock, addr = listenSocket.accept()            
            print '\t', addr
            # First message is always the clientID
            clientID = recvSizeDataMsg(sock)
            clientIDs.append(clientID)            
            # Create mappings
            self.id2socket[clientID] = sock
            
        # Put list of clientIDs on return value queue for coordinator to receive
        self.rvQueue.put(clientIDs)           
                    
        #
        # Enter server loop
        #
        print '[Network] Enter server loop'
        while 1:
            # Get request
            clientIDs, tupleData = self.requestQueue.get()
            rvs = {} # clientID -> return value
            
            # Process request            
            if clientIDs == "server":
                if tupleData == "kill":
                    print '[Network] send kill message to all clients'
                    for c in self.id2socket:  
                        sendSizeDataMsg(self.id2socket[c], pickle.dumps("kill"))                  
                    break
                else:
                    raise NetworkError('Invalid command to server: %s' % (tupleData))
            elif clientIDs == 'all':                
                for c in self.id2socket:
                    sock = self.id2socket[c]
                    # Send data...
                    sendSizeDataMsg(sock, pickle.dumps(tupleData))
                    # ...and wait for ack/ return value
                    rv = pickle.loads(recvSizeDataMsg(sock))
                    rvs[c] = rv                    
            else:
                for c in clientIDs:
                    if c not in self.id2socket:
                        raise NetworkError('Invalid client ID: %s' % (c))
                    sock = self.id2socket[c]
                    # Send data...
                    sendSizeDataMsg(sock, pickle.dumps(tupleData))
                    # ...and wait for ack/ return value
                    rv = recvSizeDataMsg(sock)
                    rvs[c] = rv
                    
            # Send reply
            self.rvQueue.put(rvs)
            
        #
        # Close connections and do cleanup
        #
        print '[Network] Close sockets'
        for c in self.id2socket:
            sock = self.id2socket[c]
            sock.close()
            
        print '[Network] Exiting' 
            
class BealjasSocketClient(threading.Thread):
    '''
    Client-side thread that handles communication with master coordinators. 
    '''
    
    def __init__(self, clientID, requestQueue, rvQueue, host='localhost', port=31278):
        '''
        Constructor
        
        @param clientID: this clients ID
        @param requestQueue: queue where to recevie requests tuples and post
        @param rvQueue: queue where to send return values
        @param host: server hostname. 
        @param port: server listen port.
        '''
        
        threading.Thread.__init__(self)
        
        # Number of clients that will connect to this server
        self.clientID = clientID
        
        # Socket connected to server
        self.socket = None
                
        # Client -> coordinator messages
        self.requestQueue = requestQueue
        # Client <- coordinator messages
        self.rvQueue = rvQueue
        
        # Host and port to connect to
        self.serverHost = host
        self.serverPort = port
        
    def run(self):
        '''
        Client main loop
        '''                 
        #
        # Connect to server
        #       
        print '[Network] connect to %s:%d' % (self.serverHost, self.serverPort)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.serverHost, self.serverPort))
        
        # The connection is initialzied by sending this clients ID
        print '[Network] send client ID'
        sendSizeDataMsg(self.socket, self.clientID)        
        # Wait for parameters messages before waiting for first requirement
        print '[Network] wait for parameters'
        parameterTuple = pickle.loads(recvSizeDataMsg(self.socket))
        self.requestQueue.put(parameterTuple)
        print '[Network] send ack for parameter tuple'
        sendSizeDataMsg(self.socket, pickle.dumps(None))
                                 
        #
        # Enter client loop
        #
        print '[Network] enter main loop'
        while 1:            
            # Get request
            pickledData = recvSizeDataMsg(self.socket)
            tupleData = pickle.loads(pickledData)
            
            # Process request
            if tupleData == 'kill':
                break
            else:
                # Send request to coordinator, wait for reply and send reply to server
                self.requestQueue.put(tupleData)
                rv = self.rvQueue.get()
                sendSizeDataMsg(self.socket, pickle.dumps(rv))            
            
        #
        # Close connection
        #
        print '[Network] close connection'
        self.socket.close()     
        print '[Network] Exiting'                

def runServer(nClients, host='localhost', port=31278):
    '''
    Start a server thread that first waits for nClients to connect and then serves the clients
    by forwarding tuples received on the queue.
    
    @param nClients: number of clients that will connect to this server   
    @param host: optional host to bind listen socket to. Default is localhost.
    @param port: optional port to listen for connections. Default is 31278. 
    @return: tuple with (server thread, requestQueue, rvQueue)
    '''
    
    requestQueue = Queue.Queue()
    rvQueue = Queue.Queue()
    thread = BealjasSocketServer(nClients, requestQueue, rvQueue, host, port)
    thread.start()
    
    return (thread, requestQueue, rvQueue)

def killServerAndClients(requestQueue):
    '''
    Send a kill request to the server, which will foward a kill message to all clients
    
    @param requestQueue: request queue
    '''
    requestQueue.put(("server", "kill"))

def runClient(clientID, host, port):
    '''
    Start a thread that handles communication with the server
    
    @param clientID: client ID
    @param host: server hostname
    @param port: server listen port
    @return: tuple with (server thread, requestQueue, rvQueue)
    '''
    
    requestQueue = Queue.Queue()
    rvQueue = Queue.Queue()
    thread = BealjasSocketClient(clientID, requestQueue, rvQueue, host, port)
    thread.start()
    
    return (thread, requestQueue, rvQueue)    

def recvSizeDataMsg(sock):
    '''
    Do a blocking receive on a (size, data) message. Where size is a 4 byte integer, and data
    is a String. This function will not return until size bytes have been read or the sock
    is closed.
    
    @param sock: sock to receive from
    @return: msg data
    '''    
    # Data length is packed into the 4 first bytes
    
    
    # Number of message data bytes read so far
    msgLen = 0    
    msgData = ''
    # Message size, which is initialized to maxint to ensure that the below loop does not block
    size = sys.maxint
    # String that holds size data (and possibly additional data following the size)
    sizeData = ''     
    # Maximum number of bytes to receive in each recv call. It is changed to msg size once the 
    # size has been read
    recvSize = MAX_RECV_SIZE
    
    while msgLen < size:
        # Attempt to read in up to MAX_RECV_SIZE bytes
        sockData = sock.recv(recvSize)
        if sockData == '': # Socket is closed
            raise NetworkError('Socket was closed')
        # The recv call may return with 1 to MAX_RECV_SIZE bytes
        
        # Parse data read from sock so far
        if len(msgData) == 0:    # Size not yet read
            if len(sockData) > 4: # Must have read at least four bytes
                sizeData += sockData
                
                # Create integer from four first bytes
                size = struct.unpack('>i', sizeData[:4])[0]            
                if size > MAX_RECV_SIZE:
                    raise NetworkError('Invalid message size: %d (max: %d)' % (size, MAX_RECV_SIZE))
                
                recvSize = size
                msgData = sizeData[4:] # remove size
                msgLen = len(sizeData) - 4
            else: # all four bytes not yet read
                sizeData += sockData
                msgLen += len(sockData)
        else:
            msgData += sockData
        
    return msgData

def sendSizeDataMsg(sock, data):
    '''
    Send a (size, data) message over a sock. Size is a four byte integer that specifies the length
    of data encoded as a string
    
    @param sock: sock over which to send data
    @param data: string of data to send
    '''
    sock.sendall(struct.pack('>i', len(data)) + data)    
    
if __name__ == '__main__':
    print socket
    ls = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    
    runClient('tile0', 'localhost', 31278)