import types

import threading
from plotterError import PlotterError

class BealjasPlotter(threading.Thread):
    '''
    Plotter thread that also is the superclass for all plotters)
    '''
    
    def __init__(self, name, pipe, loggingSetupFunction, loggingLevel, loggingOptArgs, stateObject=None):
        '''
        Constructor
        
        @param name: process ID used for logging
        @param pipe: pipe connected to coordinator, used by the Window plotter.        
        @param loggingSetupFunction: function called to setup logging for this plotter
        @param loggingLevel: logging level to use. Must be one of logging.[DEBUG, INFO, WARNING, ERROR, CRITICAL]
        @param loggingOptArgs: optional arguments sent to the logging setup function.
        @param stateObject: object sent as the second argument to the visualize function. Default is None.
        '''
        threading.Thread.__init__(self)
                
        self.name = name

        # Pipe connected to coordinator
        self.pipe = pipe
        
        # Object used to hold various state. This object is passed as argument to
        # the visualize, filter, filterKill, and close functions
        self.stateObject = stateObject
        
        if not isinstance(loggingSetupFunction, types.FunctionType):
            raise PlotterError('Invalid setup logging function: not a function')        
        
        self.logger, self.logFilename = loggingSetupFunction(name, loggingLevel, loggingOptArgs)
        if self.logger == None:
            raise PlotterError('Setup function did not return a logger')
        if self.logFilename == None:
            self.logger.info('Log filename not specified')
        
    def readLogfileLines(self):
        '''
        Read and return logfile content
        
        return: logfile content as a list of strings, or None if logging is not set to a file
        '''
        if self.logFilename == None: # Not logging to file
            self.logger.warn('Not logging to file (or setup slave logging function did not return filename)')
            return None
        else:
            return open(self.logFilename).readlines()
        
    #
    # Functions that must be implemented by subclass
    #
    def visualize(self, visualizeFunction, taskArgument):
        '''
        Function called to visualize the data
        
        @param visualizeFunction: user provided visualize function called to do perform the slave task
        @param taskArgument: task for this process        
        @return: None
        '''
        raise PlotterError('Subclass must implement this function')
        
    def gather(self, gatherFunction, taskArgument=None):
        '''
        Function called to gather values from all tasks
        
        @param gatherFunction: function run to gather values to return
        @param taskArgument: optional argument passed to the gatherFunction
        @return: values returned by gather function
        '''
        raise PlotterError('Subclass must implement this function')
        
    
    def kill(self):
        '''
        Kill this plotter
        '''
        raise PlotterError('Subclass must implement this function')
        
    def filterF(self, filterFunction, filterArgs=None):
        '''
        Execute a filter function that determines whether to show or hide this visualization,
        or wheteher to kill this function
        
        @param filterFunction: function to run that return True for visualizations to show
        or False for visualizations to hide (or True for visualization to keep, and False for 
        visualizations to kill)
        @param filterArgs: arguments passed to the filter function
        @return True if windows was set to visible, False otherwise
        '''
        raise PlotterError('Subclass must implement this function')        
            
    def run(self):
        '''
        Main loop for slave process that receives commands from the coordinator, executes these, and
        send replies to the coordinator.
        '''                                
        raise PlotterError('Subclass must implement this function')