import os, threading, types, time

from bealjasPlotter import BealjasPlotter
from plotterError import PlotterError

from bealjas.network.serializeFunction import deserialize

class WindowStatus:
    '''
    Object used to signal state changes between the GUI thread and the WindowPlotter (wp) thread. 
    It consists of two variables: isWindowClosed and nRedraws.
    
    isWindowClosed is used as follows:
    1. The GUI thread creates the WindowStatus object, which initializes isWindowClosed=None
    2. The GUI thread sets isWindowClosed='open' when it starts
    3a. THe GUI thread sets isWindowClsoed='closed' when the user closes the window (and thereby
    exits the GUI loop)
    3a. The wp thread checks if isWindowsClosed='closed' before executing a command, if it
    is closed the main thread closes its end of the cmd pipe and exits the main loop
    
    nRedraws is used as follows:
    1. The GUI thread creates the WindowStatus object, which initialized nRedraws=0
    2a. The GUI thread waits on the associated condition variable redraw, until nRedraws>0
    2b. The wp thread checks if nRedraw==0, and if so sets nRedraw=1 and signals GUI thread
    using the redraw condition variable. If nRedraw>0, it increase nRedraw by one and calls
    setVisible(true) and then signal the GUI thread using the redraw condition variable.         
    '''
    
    def __init__(self):
        # Status can be: [None, "open", "closed"]
        self.isWindowClosed = None
        self.lock = threading.Lock()
        self.nRedraws = 0
        self.redraw = threading.Condition()
        
    def setStatus(self, status):
        self.lock.acquire()
        self.isWindowClosed = status
        self.lock.release()
        
    def getStatus(self):
        self.lock.acquire()
        rv = self.isWindowClosed
        self.lock.release()
        
        return rv
        

class WindowPlotter(BealjasPlotter):
    '''
    Window plotter thread. This thread receives and executes plot commands.
    
    Note that it does not initialize the plot since it is assumed that the main thread does this.
    '''
    
    def __init__(self, name, pipe, windowStatus,
                 loggingSetupFunction, loggingLevel, loggingOptArgs,
                 closeWindowFunction,
                 setWindowVisibleFunction, getWindowPosSizeFunction, setWindowPosSizeFunction,
                 stateObject=None):
        '''
        Constructor       
        
        @param name: process ID used for logging
        @param pipe: pipe connected to coordinator, used by the Window plotter.   
        @param isWindowClose: variable that is set to True if the window is closed.
        @param loggingSetupFunction: function called to setup logging for this plotter
        @param loggingLevel: logging level to use. Must be one of logging.[DEBUG, INFO, WARNING, ERROR, CRITICAL]
        @param loggingOptArgs: optional arguments sent to the logging setup function.  
        @param closeWindowFunction: functoin called to close the window and free any resource allocated
        by the visualization. (functions for various backends may be found in modules bealjas2X modules the 
        bealjas.utils package)
        @param setWindowVisibleFunction: function called to either show or hide window (functions for various
        backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param geWindowtPosSizeFunction: function called to get the position and size of a window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param setWindowPosSizeFunction: function called to set the position and size of a window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param stateObject: object sent as the second argument to the visualize function. Default is None.                 
        '''
        BealjasPlotter.__init__(self, name, pipe, loggingSetupFunction, loggingLevel, loggingOptArgs, stateObject)
        
        self.logger.info('Started window plotter: %s' % (name))
        
        self.windowStatus = windowStatus
        
        # Visualizer threads. Since the visualizer function may or may not block these threads
        # are put in this list, which is periodically checked for joinable threads
        self.runningThreads = []
        
        # Used for logging and debugging
        self.plotterID = os.getpid()              
        
        # Check and set provided window functions
        if not isinstance(closeWindowFunction, types.FunctionType):
            raise PlotterError('Invalid close window function: not a function')
        self.closeWindow = closeWindowFunction
        if not isinstance(setWindowVisibleFunction, types.FunctionType):
            raise PlotterError('Invalid setWindowVisible function: not a function')
        self.setWindowVisible = setWindowVisibleFunction
        # All plotters are hidden until the coordinator shows these
        #self.setWindowVisible(stateObject, self.logger, False)        
        if not isinstance(getWindowPosSizeFunction, types.FunctionType):
            raise PlotterError('Invalid getWindowPosSize function: not a function')    
        self.getWindowPosSize = getWindowPosSizeFunction
        if not isinstance(setWindowPosSizeFunction, types.FunctionType):
            raise PlotterError('Invalid setWindowPosSize function: not a function')
        self.setWindowPosSize = setWindowPosSizeFunction
                
    def visualize(self, visualizeFunction, taskArgument):
        '''
        Function called to visualize the data
        
        @param visualizeFunction: user provided visualize function called to do perform the slave task
        @param taskArgument: task for this process        
        @return: None
        '''
        
        if not isinstance(visualizeFunction, types.FunctionType):
            raise PlotterError('First argument is not a function')
        
        visualizeFunction(self.stateObject, self.logger, taskArgument)
        
    def gather(self, gatherFunction, taskArgument=None):
        '''
        Function called to gather values from all tasks
        
        @param gatherFunction: function run to gather values to return
        @param taskArgument: optional argument passed to this task
        @return: values returned by gather function
        '''                
        if not isinstance(gatherFunction, types.FunctionType):
            raise PlotterError('Invalid gather function: not a function')
        
        self.logger.info('Gather values')
        #self.logger.info('%s, %s, %s' % (str(self.stateObject), str(self.logger), str(taskArgument)))
        return gatherFunction(self.stateObject, self.logger, taskArgument)
        
    
    def kill(self):
        '''
        Kill this visualization
        '''
        self.pipe.close()
        #sys.exit(0)
        os._exit(0)
        
    def filterF(self, filterFunction, filterArgs=None):
        '''
        Execute a filter function that determines whether to show or hide this visualization,
        or wheteher to kill this function
        
        @param filterFunction: function to run that return True for visualizations to show
        or False for visualizations to hide (or True for visualization to keep, and False for 
        visualizations to kill)
        @param filterArgs: arguments passed to the filter function
        @return True if windows was set to visible, False otherwise
        '''
        
        if not isinstance(filterFunction, types.FunctionType):
            raise PlotterError('First argument is not a function')
        
        return filterFunction(self.stateObject, self.logger, filterArgs)        
            
    def run(self):
        '''
        Main loop for slave process that receives commands from the coordinator, executes these, and
        send replies to the coordinator.
        '''                                
        #time.sleep(10)
        
        # All windows start by being hidden
        #attempts = 0
        #maxAttempts = 5
        #while attempts < maxAttempts:
        #    if self.setWindowVisible(self.stateObject, self.logger, False): # success
        #        # Give backend a chance to execute cmd
        #        time.sleep(1)
        #        break
        #    else:
                # Wait a while then try again
        #        attempts += 1
        #        self.logger.warn("Could not hide window...trying again (attempt %d of %d)" % (attempts, maxAttempts))
        #        time.sleep(0.2)
               
        vizFunctionCounter = 0
        filterFunctionCounter = 0
        gatherFunctionCounter = 0
        while 1: # Loop until kill or invalid command is received
            self.logger.debug('Wait for command')
            cmd = self.pipe.recv()
                        
            # Check if window has closed
            if self.windowStatus.getStatus() == "closed":
                self.pipe.send('killed')
                self.pipe.close()
                break
                                    
            #self.pipe.send('processing')
            
            self.logger.info('CMD = %s' % (cmd))            
            if cmd == 'visualize':                
                funcSer, arg = self.pipe.recv()
                func = deserialize(funcSer, 'visualize%03d' % (vizFunctionCounter))
                vizFunctionCounter += 1
                self.visualize(func, arg)                
            elif cmd == 'show':                                                
                # Signal that window can be drawn                
                #self.windowStatus.redraw.acquire()                
                #if self.windowStatus.nRedraws == 0:                    
                #    self.windowStatus.nRedraws = 1                    
                #    self.windowStatus.redraw.notify() # always send signal                     
                #    self.windowStatus.redraw.release()                        
                self.setWindowVisible(self.stateObject, self.logger, True)                                        
                #else:
                #    self.windowStatus.nRedraws += 1
                #    self.setWindowVisible(self.stateObject, self.logger, True)
                #    self.windowStatus.redraw.notify() # always send signal
                #    self.windowStatus.redraw.release()                                    
            elif cmd == 'hide':                                
                self.setWindowVisible(self.stateObject, self.logger, False)
            elif cmd == 'filter':
                funcSer, args = self.pipe.recv()
                func = deserialize(funcSer, 'filter%03d' % (filterFunctionCounter))
                filterFunctionCounter += 1
                rv = self.filterF(func, args)
                self.pipe.send(rv)
            elif cmd == 'gather':
                funcSer, args = self.pipe.recv()    
                func = deserialize(funcSer, 'gather%03d' % (gatherFunctionCounter))
                gatherFunctionCounter += 1   
                rv = self.gather(func, args)
                self.pipe.send(rv)   
            elif cmd == 'close':                
                rv = self.closeWindow(self.stateObject, self.logger)
                self.pipe.send(rv)
            elif cmd == 'closeDebug':
                # Special case for bealjas2debug used for unit testing and development
                # It will kill the threads in addition
                rv = self.closeWindow(self.stateObject, self.logger)
                self.pipe.send(rv)
                # Kill GUI thread
                self.stateObject['kill'] = True
            elif cmd == 'kill':
                self.kill()
                self.pipe.send('killed')
                break
            elif cmd == 'readLogfileLines':
                rv = self.readLogfileLines()
                self.pipe.send(rv)
            # Window plotter specific commands
            elif cmd == 'getPosSize':
                rv = self.getWindowPosSize(self.stateObject, self.logger)
                self.pipe.send(rv)
            elif cmd == 'setPosSize':
                args = self.pipe.recv()
                self.setWindowPosSize(self.stateObject, self.logger, args)
            else:
                self.logger.warning('Invalid cmd received: %s' % (cmd))
                break
            
            # Yield to give GUI thread a chance to run
            time.sleep(0.1)
            
            # Check if any of the visualizer threads are done
            toRemove = []
            for t in self.runningThreads:
                if not t.isAlive():
                    self.logger.info('Join with thread')
                    t.join()
                    toRemove.append(t)                    
            for t in toRemove:
                self.runningThreads.remove(t)                
            
                
def plotterEntry(name, pipe, 
                 setupLoggerFunction, loggerLevel,
                 createWindowFunction, guiLoopFunction, closeWindowFunction,
                 setWindowVisibleFunction, getWindowPosSizeFunction, setWindowPosSizeFunction,
                 loggerOptArgs=None, stateObject=None):
    '''
    Slave process entry point. All plotters should implement this function with the following parameters:
        
    @param name: slave process name
    @param pipe: pipe connected to coordinator, used by the Window plotter.
    @param setupLoggerFunction: function called to setup logging for this plotter
    @param loggingLevel: logging level to use. Must be one of logging.[DEBUG, INFO, WARNING, ERROR, CRITICAL]    
    @param createWindowFunction: function called to create a window that is used in the subeqent calls    
    (functions for various backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param guiLoopFunction: function called to execute the GUI loop. (functions for various backends may
    be found in modules bealjas2X modules the bealjas.utils package)
    @param closeWindowFunction: function called to close the window and free any resource allocated
    by the visualization. (functions for various backends may be found in modules bealjas2X modules the 
    bealjas.utils package)
    @param setWindowVisibleFunction: function called to either show or hide window (functions for various
    backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param geWindowtPosSizeFunction: function called to get the position and size of a window (functions 
    for various backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param setWindowPosSizeFunction: function called to set the position and size of a window (functions 
    for various backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param loggingOptArgs: optional arguments sent to the logging setup function. 
    @param stateObject: optional object used to hold various state for the visualization. This object is 
    passed as argument to the visualize, filter, filterKill, and gather functions. Default is None.
    
    Note! It is possible to use global variables instead of the state object
    '''
    
    print 'P1: ', time.time()
    
    if not isinstance(createWindowFunction, types.FunctionType):
        raise PlotterError('Invalid createWindow function: not a function')
    
        
    windowStatus = WindowStatus()                
    # Start plotter thread
    wp = WindowPlotter(name, pipe, windowStatus,
                       setupLoggerFunction, loggerLevel, loggerOptArgs, 
                       closeWindowFunction,
                       setWindowVisibleFunction, getWindowPosSizeFunction, setWindowPosSizeFunction,
                       stateObject)               
    # Create window before starting request thread    
    createWindowFunction(stateObject, wp.logger)
    setWindowVisibleFunction(stateObject, wp.logger, False)
    wp.start()
        
    # Set status to open before entering the GUI loop
    windowStatus.setStatus("open")
    # This is a blocking call
    guiLoopFunction(stateObject, wp.logger, windowStatus, name)
    # Signals that the wp thread should exit
    windowStatus.setStatus("closed")

    wp.join()
    
if __name__ == '__main__':
    # Debug
    import logging
    from multiprocessing import Pipe
    from bealjas.utils.bealjas2pyplot import createWindow, guiLoop, setWindowPosSize, getWindowPosSize, setWindowVisible
    from bealjas.utils.setupLogging import setupConsoleLogger
    parentEnd, childEnd = Pipe()
    plotterEntry('debugEntry', 
                 childEnd,
                 setupConsoleLogger,
                 logging.DEBUG,
                 createWindow,
                 guiLoop,
                 setWindowVisible,
                 getWindowPosSize,
                 setWindowPosSize)