import os, threading, types, time

from windowPlotter import WindowPlotter
from plotterError import PlotterError

from bealjas.network.serializeFunction import deserialize

class WindowPlotter2(WindowPlotter):
    '''
    Window plotter thread. This thread receives and executes plot commands.
    
    Note that it does not initialize the plot since it is assumed that the main thread does this.
    '''
    
    def __init__(self, name, pipe, windowStatus,
                 loggingSetupFunction, loggingLevel, loggingOptArgs,
                 closeWindowFunction,
                 setWindowVisibleFunction, getWindowPosSizeFunction, setWindowPosSizeFunction,
                 stateObject=None):
        '''
        Constructor       
        
        @param name: process ID used for logging
        @param pipe: pipe connected to coordinator, used by the Window plotter.   
        @param isWindowClose: variable that is set to True if the window is closed.
        @param loggingSetupFunction: function called to setup logging for this plotter
        @param loggingLevel: logging level to use. Must be one of logging.[DEBUG, INFO, WARNING, ERROR, CRITICAL]
        @param loggingOptArgs: optional arguments sent to the logging setup function.  
        @param closeWindowFunction: functoin called to close the window and free any resource allocated
        by the visualization. (functions for various backends may be found in modules bealjas2X modules the 
        bealjas.utils package)
        @param setWindowVisibleFunction: function called to either show or hide window (functions for various
        backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param geWindowtPosSizeFunction: function called to get the position and size of a window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param setWindowPosSizeFunction: function called to set the position and size of a window (functions 
        for various backends may be found in modules bealjas2X modules the bealjas.utils package)
        @param stateObject: object sent as the second argument to the visualize function. Default is None.                 
        '''
        WindowPlotter.__init__(self, name, pipe, windowStatus,
                 loggingSetupFunction, loggingLevel, loggingOptArgs,
                 closeWindowFunction,
                 setWindowVisibleFunction, getWindowPosSizeFunction, setWindowPosSizeFunction,
                 stateObject)
                
    def run(self):
        '''
        This function is called periodically by the GUI thread
        '''                                        
               
        self.vizFunctionCounter = 0
        self.filterFunctionCounter = 0
        self.gatherFunctionCounter = 0
    
        self.logger.debug('Wait for command')
        # poll can be used (followed by recv) in order to avoid blocking GUI thread
        if self.pipe.poll(1):
            cmd = self.pipe.recv()
        else:
            self.stateObject.root.after(1000, self.run)
            return
                                            
        self.logger.info('CMD = %s' % (cmd))            
        if cmd == 'visualize':                
            funcSer, arg = self.pipe.recv()
            func = deserialize(funcSer, 'visualize%03d' % (self.vizFunctionCounter))
            self.vizFunctionCounter += 1
            self.visualize(func, arg)                
        elif cmd == 'show':                                                
            self.setWindowVisible(self.stateObject, self.logger, True)                                        
        elif cmd == 'hide':                                
            self.setWindowVisible(self.stateObject, self.logger, False)
        elif cmd == 'filter':
            funcSer, args = self.pipe.recv()
            func = deserialize(funcSer, 'filter%03d' % (self.filterFunctionCounter))
            self.filterFunctionCounter += 1
            rv = self.filterF(func, args)
            self.pipe.send(rv)
        elif cmd == 'gather':
            funcSer, args = self.pipe.recv()    
            func = deserialize(funcSer, 'gather%03d' % (self.gatherFunctionCounter))
            self.gatherFunctionCounter += 1   
            rv = self.gather(func, args)
            self.pipe.send(rv)   
        elif cmd == 'close':                
            rv = self.closeWindow(self.stateObject, self.logger)
            self.pipe.send(rv)
        elif cmd == 'closeDebug':
            # Special case for bealjas2debug used for unit testing and development
            # It will kill the threads in addition
            rv = self.closeWindow(self.stateObject, self.logger)
            self.pipe.send(rv)
            # Kill GUI thread
            self.stateObject['kill'] = True
        elif cmd == 'kill':
            self.kill()
            self.pipe.send('killed')            
        elif cmd == 'readLogfileLines':
            rv = self.readLogfileLines()
            self.pipe.send(rv)
        # Window plotter specific commands
        elif cmd == 'getPosSize':
            rv = self.getWindowPosSize(self.stateObject, self.logger)
            self.pipe.send(rv)
        elif cmd == 'setPosSize':
            args = self.pipe.recv()
            self.setWindowPosSize(self.stateObject, self.logger, args)
        else:
            self.logger.warning('Invalid cmd received: %s' % (cmd))
            
        self.stateObject.root.after(0, self.run)
        
                
def plotterEntry(name, pipe, 
                 setupLoggerFunction, loggerLevel,
                 createWindowFunction, guiLoopFunction, closeWindowFunction,
                 setWindowVisibleFunction, getWindowPosSizeFunction, setWindowPosSizeFunction,
                 loggerOptArgs=None, stateObject=None):
    '''
    Slave process entry point. All plotters should implement this function with the following parameters:
        
    @param name: slave process name
    @param pipe: pipe connected to coordinator, used by the Window plotter.
    @param setupLoggerFunction: function called to setup logging for this plotter
    @param loggingLevel: logging level to use. Must be one of logging.[DEBUG, INFO, WARNING, ERROR, CRITICAL]    
    @param createWindowFunction: function called to create a window that is used in the subeqent calls    
    (functions for various backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param guiLoopFunction: function called to execute the GUI loop. (functions for various backends may
    be found in modules bealjas2X modules the bealjas.utils package)
    @param closeWindowFunction: function called to close the window and free any resource allocated
    by the visualization. (functions for various backends may be found in modules bealjas2X modules the 
    bealjas.utils package)
    @param setWindowVisibleFunction: function called to either show or hide window (functions for various
    backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param geWindowtPosSizeFunction: function called to get the position and size of a window (functions 
    for various backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param setWindowPosSizeFunction: function called to set the position and size of a window (functions 
    for various backends may be found in modules bealjas2X modules the bealjas.utils package)
    @param loggingOptArgs: optional arguments sent to the logging setup function. 
    @param stateObject: optional object used to hold various state for the visualization. This object is 
    passed as argument to the visualize, filter, filterKill, and gather functions. Default is None.
    
    Note! It is possible to use global variables instead of the state object
    '''
    
    print 'P1: ', time.time()
    
    if not isinstance(createWindowFunction, types.FunctionType):
        raise PlotterError('Invalid createWindow function: not a function')
    
                
    wp = WindowPlotter2(name, pipe, None,
                       setupLoggerFunction, loggerLevel, loggerOptArgs, 
                       closeWindowFunction,
                       setWindowVisibleFunction, getWindowPosSizeFunction, setWindowPosSizeFunction,
                       stateObject)               
    
    # Create window before starting request thread    
    createWindowFunction(stateObject, wp.logger)
    # Do not start
    #wp.start()
            
    # This will enter the gui loop that will do upcalls to the window plotter
    guiLoopFunction(stateObject, wp.logger, wp.run, name)   
    
if __name__ == '__main__':
    # Debug
    import logging
    from multiprocessing import Pipe
    from bealjas.utils.bealjas2pyplot import createWindow, guiLoop, setWindowPosSize, getWindowPosSize, setWindowVisible
    from bealjas.utils.setupLogging import setupConsoleLogger
    parentEnd, childEnd = Pipe()
    plotterEntry('debugEntry', 
                 childEnd,
                 setupConsoleLogger,
                 logging.DEBUG,
                 createWindow,
                 guiLoop,
                 setWindowVisible,
                 getWindowPosSize,
                 setWindowPosSize)