'''
Implementation of bealjas plotter functions for a simple window simulator where:
-screen: is a 10x10 matrix
-isVisible: is a boolean variable
-x, y, w, h: are integers for respectively (x, y) coordinate, width and height of the window. 
These have no influence of the screen object
-m: is a boolean variable
-kill: is periodically checked, and once set to True the gui thread exits
-isClosed: true if window has been closed

All of these are stored in a dictionary that is passed as the stateObject
'''

def createWindow(stateObject, logger):
    '''
    Open a pyplot window and enter the main GUI loop. 
    
    Note that this function is called by the initializer thread and it should not block.
    For blocking calls, use guiLoop instead which is called after createWindows
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    '''
    
    if logger != None:
        logger.info("Create window")
    
    stateObject['screen'] = []
    for i in range(10):
        l = []
        for j in range(10):
            l.append(None)
        stateObject['screen'].append(l)
        
    stateObject['isVisible'] = False
    stateObject['x'] = 0
    stateObject['y'] = 0
    stateObject['w'] = 10
    stateObject['h'] = 10
    stateObject['m'] = False
    stateObject['kill'] = False
    stateObject['isClosed'] = False
    
def guiLoop(stateObject, logger, windowStatus):
    '''
    Enter the GUI loop. 
    
    This function should block until the window is closed by the user.
    
    Note that this function is called by the initializer thread.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param windowStatus: a WindowStatus object that contains the redraw condition variable
    which the GUI thread should wait on until nRedraws>0 before entering the GUI loop.
    '''
    
    # Wait until first visualization is done     
    windowStatus.redraw.acquire()
    while windowStatus.nRedraws == 0:
        windowStatus.redraw.wait()
    windowStatus.redraw.release()
    
    # Block: enter GUI loop
    logger.info("Enter GUI loop")
    import time      
    # Loop 
    while 1:
        if stateObject['kill']:
            break
        time.sleep(1)

def closeWindow(stateObject, logger):
    '''
    Close window and free window resources.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger    
    @return: None
    '''
               
    logger.info('Closing window')
    stateObject['kill'] = True
    stateObject['isClosed'] = True

def setWindowVisible(stateObject, logger, visible):
    '''
    Show or hide the window
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param visible: True if window should be shown, False if it should be hidden
    @return: True on success, False otherwise
    '''
    
    if visible:
        logger.info('Show frame')
        stateObject['isVisible'] = True
    else:
        logger.info("Hide frame")
        stateObject['isVisible'] = False
    
    # Success
    return True

def getWindowPosSize(stateObject, logger):
    '''
    Get the window position and size. This function is called before the close window function in order
    to get the position for the next window to open.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @return a tuple of the format (m, x, y, w, h), where m is True if the window is maximized, and where
    x, y, w, and h are either all None if m is True, or they contain respectively the window x position,
    y position, width and height. 
    '''
        
    return (stateObject['m'], 
            stateObject['x'], stateObject['y'], 
            stateObject['w'], stateObject['h'])
    
def setWindowPosSize(stateObject, logger, args):
    '''
    Set the window position and size. This function is called after the visualize function to move the
    window to the position of the previous open window.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: a tuple of the format (m, x, y, w, h), where m is True if the window is to be maximized, 
    and where x, y, w, and h are either all None if m is True, or they contain respectively the window 
    x position, y position, width and height.
    @return: True on success, False otherwise
    '''
    
    m, x, y, w, h = args
    
    if m: #maximize
        stateObject['m'] = True
        # No need to set position and size
    else: # resize
        stateObject['m'] = False
        stateObject['x'] = x
        stateObject['y'] = y
        stateObject['w'] = w
        stateObject['h'] = h        

    # Success
    return True