'''
Implementation of bealjas plotter functions for the matplotlib.pyplot library
'''

import sys
import pygame
from bealjas.plotter.plotterError import PlotterError

def createWindow(stateObject, logger):
    '''
    Open a pyplot window and enter the main GUI loop. 
    
    Note that this function is called by the initializer thread and it should not block.
    For blocking calls, use guiLoop instead which is called after createWindows.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    '''
    logger.debug('Create window')
    pygame.init()    
    stateObject.screen = pygame.display.set_mode((stateObject.width, stateObject.height), pygame.RESIZABLE)
    
def guiLoop(stateObject, logger, windowStatus, title=None):
    '''
    Enter the GUI loop. 
    
    This function should block until the window is closed by the user.
    
    Note that this function is called by the initializer thread.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param windowStatus: a WindowStatus object that contains the redraw condition variable
    which the GUI thread should wait on until nRedraws>0 before entereing the GUI loop.
    @param title: optional window title
    '''
    logger.debug('Wait for first draw operation before entering the GUI loop')
    
    # Add title
    if title:
        pygame.display.set_caption(title)
    
    # Block: enter GUI loop       
    logger.debug('Enter GUI loop')
    while True:
        for event in pygame.event.get():            
            if event.type == pygame.QUIT: 
                #sys.exit()
                break

def closeWindow(stateObject, logger):
    '''
    Close window and free window resources.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger    
    @return: None
    '''
               
    logger.info('Closing window')
    pygame.display.quit()

def setWindowVisible(stateObject, logger, visible):
    '''
    Show or hide the window
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param visible: True if window should be shown, False if it should be hidden
    @return: True on success, False otherwise
    '''
    
    if visible:
        logger.debug('Show frame')        
        pygame.display.set_mode((stateObject.width, stateObject.height),  pygame.RESIZABLE)
    else:
        logger.debug("Hide frame")        
        #pygame.display.iconify()
        pygame.display.set_mode((1, 1),  pygame.RESIZABLE)
        
    
    # Success
    return True

def getWindowPosSize(stateObject, logger):
    '''
    Get the window position and size. This function is called before the close window function in order
    to get the position for the next window to open.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @return a tuple of the format (m, x, y, w, h), where m is True if the window is maximized, and where
    x, y, w, and h are either all None if m is True, or they contain respectively the window x position,
    y position, width and height. 
    '''
    
    return (False, 0, 0, 0, 0)
    
def setWindowPosSize(stateObject, logger, args):
    '''
    Set the window position and size. This function is called after the visualize function to move the
    window to the position of the previous open window.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: a tuple of the format (m, x, y, w, h), where m is True if the window is to be maximized, 
    and where x, y, w, and h are either all None if m is True, or they contain respectively the window 
    x position, y position, width and height.
    @return: True on success, False otherwise
    '''

    # Success
    return True