'''
Implementation of bealjas plotter functions for the matplotlib.pyplot library
'''

from Tkinter import *

from bealjas.plotter.plotterError import PlotterError


def createWindow(stateObject, logger):
    '''
    Open a pyplot window and enter the main GUI loop. 
    
    Note that this function is called by the initializer thread and it should not block.
    For blocking calls, use guiLoop instead which is called after createWindows.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @return: None, but the root window handle is stored in stateObject.root
    '''
    logger.debug('Create window')
    stateObject.root = Tk()
    
def guiLoop(stateObject, logger, runFunc, title=None):
    '''
    Enter the GUI loop. 
    
    This function should block until the window is closed by the user.
    
    Note that this function is called by the initializer thread.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param runFunc: pointer to the function which the GUI thread does an up call to
    @param title: optional window title
    '''    
    
    # Add title
    if title:
        stateObject.root.title(title)
    
    stateObject.root.after(0, runFunc)
    # Block: enter GUI loop       
    logger.debug('Enter GUI loop')
    mainloop()

def closeWindow(stateObject, logger):
    '''
    Close window and free window resources.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger    
    @return: None
    '''
               
    logger.info('Closing window')
    stateObject.root.destroy()

def setWindowVisible(stateObject, logger, visible):
    '''
    Show or hide the window
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param visible: True if window should be shown, False if it should be hidden
    @return: True on success, False otherwise
    '''
    
    if visible:
        logger.debug('Show frame')
        stateObject.root.update()        
        stateObject.root.deiconify()        
        stateObject.root.update()
    else:
        logger.debug("Hide frame")
        stateObject.root.withdraw()    
            
    # Success
    logger.debug("done")
    return True

def getWindowPosSize(stateObject, logger):
    '''
    Get the window position and size. This function is called before the close window function in order
    to get the position for the next window to open.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @return a tuple of the format (m, x, y, w, h), where m is True if the window is maximized, and where
    x, y, w, and h are either all None if m is True, or they contain respectively the window x position,
    y position, width and height. 
    '''
    
    geoStr = stateObject.root.geometry()
    # geoStr can be for example '400x400+100+100'
    parts1 = geoStr.split('x')
    w = int(parts1[0])
    parts2 = parts1[1].split('+')
    h = int(parts2[0])
    x = int(parts2[1])
    y = int(parts2[2])
    
    return (False, x, y, w, h)
    
def setWindowPosSize(stateObject, logger, args):
    '''
    Set the window position and size. This function is called after the visualize function to move the
    window to the position of the previous open window.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: a tuple of the format (m, x, y, w, h), where m is True if the window is to be maximized, 
    and where x, y, w, and h are either all None if m is True, or they contain respectively the window 
    x position, y position, width and height.
    @return: True on success, False otherwise
    '''
    
    m, x, y, w, h = args
    
    if m: #maximize
        logger.critical('Maximize is not yet implemented')            
    else: # resize
        stateObject.root.geometry('%dx%d+%d+%d' % (w, h, x, y))
        
    # Success
    return True