'''
Implementation of bealjas plotter functions for the matplotlib.pyplot library
'''

from bealjas.plotter.plotterError import PlotterError
import matplotlib
import matplotlib.pyplot as plt

def createWindow(stateObject, logger):
    '''
    Open a pyplot window and enter the main GUI loop. 
    
    Note that this function is called by the initializer thread and it should not block.
    For blocking calls, use guiLoop instead which is called after createWindows.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    '''
    logger.debug('Create window')
    plt.plot()
    
def guiLoop(stateObject, logger, windowStatus, title=None):
    '''
    Enter the GUI loop. 
    
    This function should block until the window is closed by the user.
    
    Note that this function is called by the initializer thread.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param windowStatus: a WindowStatus object that contains the redraw condition variable
    which the GUI thread should wait on until nRedraws>0 before entereing the GUI loop.
    @param title: optional window title
    '''
    logger.debug('Wait for first draw operation before entering the GUI loop')
    
    # Wait until first visualization is done     
    #windowStatus.redraw.acquire()
    #while windowStatus.nRedraws == 0:
    #    windowStatus.redraw.wait()
    #windowStatus.redraw.release()
    
    # Add title
    if title:
        fig = plt.figure()
        fig.canvas.set_window_title(title)
    
    # Block: enter GUI loop       
    logger.debug('Enter GUI loop')
    plt.show()

def closeWindow(stateObject, logger):
    '''
    Close window and free window resources.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger    
    @return: None
    '''
               
    logger.info('Closing window')
    plt.close()

def setWindowVisible(stateObject, logger, visible):
    '''
    Show or hide the window
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param visible: True if window should be shown, False if it should be hidden
    @return: True on success, False otherwise
    '''
    
    if visible:
        logger.debug('Show frame')
    else:
        logger.debug("Hide frame")
        
    if len(plt.get_fignums()) == 0:
        logger.warn("No windows are currently open")
        return False
    
    # Get window manager
    logger.debug('Get fig man')
    mng = plt.get_current_fig_manager()
    # Each manager must be treated separately since they do not necessarily share a common interface
    if isinstance(mng, matplotlib.backends.backend_wx.FigureManagerWx):
        frame = mng.frame
        
        if visible:
            logger.debug('show frame')
            frame.Show()
            logger.debug('shown')
        else:
            frame.Hide()
    else:
        PlotterError("Unsupported window manager: " % (mng))
    
    # Success
    return True

def getWindowPosSize(stateObject, logger):
    '''
    Get the window position and size. This function is called before the close window function in order
    to get the position for the next window to open.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @return a tuple of the format (m, x, y, w, h), where m is True if the window is maximized, and where
    x, y, w, and h are either all None if m is True, or they contain respectively the window x position,
    y position, width and height. 
    '''
    
    # Get window manager
    mng = plt.get_current_fig_manager()
    # Each manager must be treated separately since they do not necessarily share a common interface
    if isinstance(mng, matplotlib.backends.backend_wx.FigureManagerWx):
        frame = mng.frame
        
        # Is maximized? 
        m = frame.IsMaximized()
        # Get position
        x, y = frame.GetPositionTuple()
        # Get size
        w, h = frame.GetSizeTuple()
        
        return (m, x, y, w, h)
    else:
        PlotterError("Unsupported window manager: " % (mng))
    
def setWindowPosSize(stateObject, logger, args):
    '''
    Set the window position and size. This function is called after the visualize function to move the
    window to the position of the previous open window.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: a tuple of the format (m, x, y, w, h), where m is True if the window is to be maximized, 
    and where x, y, w, and h are either all None if m is True, or they contain respectively the window 
    x position, y position, width and height.
    @return: True on success, False otherwise
    '''
    
    m, x, y, w, h = args
    
    # Get window manager
    mng = plt.get_current_fig_manager()
    # Each manager must be treated separately since they do not necessarily share a common interface
    if isinstance(mng, matplotlib.backends.backend_wx.FigureManagerWx):
        frame = mng.frame
        if m: #maximize
            frame.Maximize(True)
            # No need to set position and size
        else: # resize
            frame.Maximize(False)
            # Move
            frame.MoveXY(x, y)
            # Resize
            frame.SetSizeWH(w, h)
    else:
        PlotterError("Unsupported window manager: " % (mng))

    # Success
    return True