'''
Skeletons for plotter functions. 

These following three must be implemented for each plotting/ graphics library used:
-createWindow
-guiLoop
-closeWindow
-setWindowVisible
-getWindowPosSize
-setWindowPosSize

The following three functions provide a function prototype for the functions that are 
provided by the application to the plotter:
-visualize
-filterF
-gather

The final setupLogging provides a function prototype for logger setup called by the plotter
-setupLogging
'''

def createWindow(logger):
    '''
    Open a window. The return value is passed to all subsequent visualize,
    close, and window manipulation functions.
    
    Note that this function is called by the initializer thread and it should not block.
    For blocking calls, use guiLoop instead which is called after createWindows
    
    @return: None
    '''
    pass
    
def guiLoop():
    '''
    Enter the GUI loop. 
    
    This function should block until the window is closed by the user.
    
    Note that this function is called by the initializer thread.
    '''
        
    # Block until window is closed        
    pass

def closeWindow(stateObject, logger):
    '''
    Close window and free window resources.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger    
    @return: None
    '''
               
    pass


def setWindowVisible(stateObject, logger, visible):
    '''
    Show or hide the window
        
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param visible: True if window should be shown, False if it should be hidden
    @return: True on success, False otherwise
    '''
    return False # fail
    

def getWindowPosSize(stateObject, logger):
    '''
    Get the window position and size. This function is called before the close window function in order
    to get the position for the next window to open.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @return a tuple of the format (m, x, y, w, h), where m is True if the window is maximized, and where
    x, y, w, and h are either all None if m is True, or they contain respectively the window x position,
    y position, width and height. 
    '''

    # return empty tuple
    return (False, 0, 0, 0, 0)
    
def setWindowPosSize(stateObject, logger, args):
    '''
    Set the window position and size. This function is called after the visualize function to move the
    window to the position of the previous open window.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: a tuple of the format (m, x, y, w, h), where m is True if the window is to be maximized, 
    and where x, y, w, and h are either all None if m is True, or they contain respectively the window 
    x position, y position, width and height.
    @return: True on success, False otherwise
    '''
    
    # Unpack args
    #m, x, y, w, h = args 
    
    return False # fail

def visualize(stateObject, logger, args):
    '''
    This function is called to visualize a graph.
    
    Note! this function should not block
        
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: optional application specific arguments
    @return: None    
    '''
    
    return None
    
    
def gather(stateObject, logger, args=None):
    '''
    Return a value to be gathered by the coordinator   
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: optional application specific argument
    @return: an object that is sent to the coordinator. It is also possible to return None
    '''
    
    # The return values are gatehred by the coordinator
    return None

def filterF(stateObject, logger, args=None):
    '''
    Run a filter function to determine wether a task should be hidden or killed. The function should
    return True for task to show and not-kill, and False for tasks to hide and kill.
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: optional application specific argument
    @return: True or False
    '''
    
    # Return either True (show or keep) or False (hide or kill)
    return True

def setupLogger(pid, level, args=None):
    '''
    Setup a logger.
    
    Note that setup loggers for console and log file are provided in bealjas.utils.setupLogging
    
    @param pid: process identifier used in the logger output
    @param level: logging level to use. Must be one of logging.[DEBUG, INFO, WARNING, ERROR, CRITICAL]
    @param args: optional application specific argument
    @return: (logger object, log filename). The log filename can be None if the logger does not log to a file.
    '''
    
    logger = None
    logFilename = None
    
    # Return (logger object, log filename)
    return (logger, logFilename)