import unittest
import logging, tempfile, types

from bealjas.coordinator.bealjasCoordinator import BealjasTask
from bealjas.coordinator.windowCoordinator import WindowCoordinator
from bealjas.coordinator.coordinatorError import CoordinatorError
from bealjas.utils.bealjas2debug import createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize
from bealjas.utils.setupLogging import setupFileLogger, setupConsoleLogger

def visualizeSet(stateObject, logger, args):
    '''
    @param: stateObject: the debug screen object
    @param logger: a logger
    @param args: an integer to set the screen elements to
    '''
    logger.info("Set screen elements")
    for i in range(len(stateObject['screen'])):
        for j in range(len(stateObject['screen'][i])):
            stateObject['screen'][i][j] = args
            
def visualizeAdd(stateObject, logger, args):
    '''
    @param: stateObject: the debug screen object
    @param logger: a logger
    @param args: an integer to add to the screen elements
    '''
    logger.info("Add %d to screen elements" % (args))
    for i in range(len(stateObject['screen'])):
        for j in range(len(stateObject['screen'][i])):
            stateObject['screen'][i][j] = stateObject['screen'][i][j] + args
            
def visualizeSub(stateObject, logger, args):
    '''
    @param: stateObject: the debug screen object
    @param logger: a logger
    @param args: an integer to substract from the screen elements
    '''
    logger.info("Subtract %d from screen elements" % (args))
    for i in range(len(stateObject['screen'])):
        for j in range(len(stateObject['screen'][i])):
            stateObject['screen'][i][j] = stateObject['screen'][i][j] - args
            
def gather1(stateObject, logger, args=None):
    '''
    Return the args value   
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: an integer returned by this function
    @return: an object that is sent to the coordinator. It is also possible to return None
    '''
     
    logger.info("Gather arg: %d" % (args))
    return args

def gather2(stateObject, logger, args=None):
    '''
    Return an element in the screen matrix   
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: ignored
    @return: an object that is sent to the coordinator. It is also possible to return None
    '''
     
    rv = stateObject['screen'][2][3]
    logger.info("Gather screen element: %d" % (rv))
    return rv

def gatherStateObject(stateObject, logger, args=None):
    '''
    Return the stateObject  
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: ignored
    @return: the state Object
    '''    

    logger.info("Gather state object")
    return stateObject

def filter1(stateObject, logger, args=None):
    '''
    Return the boolean in args
    
    param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param taskArg: True or False
    @return: args
    '''
    logger.info("Filter: %s" % (args))
    return args

def filter2(stateObject, logger, args=None):
    '''
    Return True if the screen element values are even and False if they are not (all elements have the
    same value)
    
    param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param taskArg: True or False
    @return: args
    '''    
    v = stateObject['screen'][2][3]
    logger.info("Filter: is %d even" % (v))
    return (v % 2 == 0)
     
class WindowCoordinatorTest(unittest.TestCase):

    def setUp(self):
        # Tmp dir for logfiles
        self.tmpLogDir = tempfile.mkdtemp()
        
        # This dict holds the debug window state
        self.debugWindow = {}
        
        self.nWindows = 3
        self.nSlaves = 15
        self.nTasks = 32
        self.coordinator = WindowCoordinator(self.nWindows, self.nSlaves, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)

    def tearDown(self):
        pass        
            
    # Helper function that returns a state object
    def getStateObjects(self):        
        return self.coordinator.gather(gatherStateObject, None)

    def testConstructor(self):
        self.assertEqual(self.nWindows, self.coordinator.nWindows)
        self.assertEqual(self.nWindows, self.coordinator.maxWindows)
        self.assertEqual(self.nSlaves, self.coordinator.nSlaves)
        
        self.assertEqual(createWindow, self.coordinator.createWindowFunc)
        self.assertEqual(guiLoop, self.coordinator.guiLoopFunc)
        self.assertEqual(closeWindow, self.coordinator.closeWindowFunc)
        self.assertEqual(setWindowVisible, self.coordinator.setWindowVisibleFunc)
        self.assertEqual(getWindowPosSize, self.coordinator.getWindowPosSizeFunc)
        self.assertEqual(setWindowPosSize, self.coordinator.setWindowPosSizeFunc)
        
        self.assertEqual(0, len(self.coordinator.functionList))
        self.assertEqual(0, len(self.coordinator.taskList))
        self.assertEqual(0, len(self.coordinator.currentlyRunning))
        self.assertEqual(0, len(self.coordinator.currentlyVisible))
        self.assertEqual(0, len(self.coordinator.closedWindowsPosSize))
        
        #
        # Errors
        #
        # Too many windows
        self.assertRaises(CoordinatorError, WindowCoordinator, 6, 3, 
                          createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                          self.debugWindow)
        # Invalid functions
        foo = []
        self.assertRaises(CoordinatorError, WindowCoordinator, 6, 3, 
                          foo, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                          self.debugWindow)
        self.assertRaises(CoordinatorError, WindowCoordinator, 6, 3, 
                          createWindow, foo, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                          self.debugWindow)
        self.assertRaises(CoordinatorError, WindowCoordinator, 6, 3, 
                          createWindow, guiLoop, foo, setWindowVisible, getWindowPosSize, setWindowPosSize,
                          self.debugWindow)
        self.assertRaises(CoordinatorError, WindowCoordinator, 6, 3, 
                          createWindow, guiLoop, closeWindow, foo, getWindowPosSize, setWindowPosSize,
                          self.debugWindow)
        self.assertRaises(CoordinatorError, WindowCoordinator, 6, 3, 
                          createWindow, guiLoop, closeWindow, setWindowVisible, foo, setWindowPosSize,
                          self.debugWindow)
        self.assertRaises(CoordinatorError, WindowCoordinator, 6, 3, 
                          createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, foo,
                          self.debugWindow)
    
    def testStartKillSlaves(self):
        # Start slaves
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))        
        self.assertEquals(self.nTasks, len(self.coordinator.taskList))
        self.assertEquals(self.nSlaves, len(self.coordinator.currentlyRunning))
        self.assertEquals(1, len(self.coordinator.functionList))
        self.assertTrue(self.coordinator.startAhead > 0)
        
        for i in range(len(self.coordinator.taskList)):
            t = self.coordinator.taskList[i]
            self.assertEquals(i, t.taskID)
            
        for i in range(len(self.coordinator.currentlyRunning)):
            t = self.coordinator.currentlyRunning[i]
            self.assertTrue(t.isRunning())
            
        self.assertEquals(visualizeSet, self.coordinator.functionList[0])
        
        # Kill slaves
        self.coordinator.killSlaves()
        self.assertEquals(0, len(self.coordinator.taskList))
        self.assertEquals(0, len(self.coordinator.currentlyRunning))
        self.assertEquals(0, len(self.coordinator.functionList))
                            
        # nSlaves not specified (set to nWindows)
        self.coordinator = WindowCoordinator(self.nWindows, None, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.assertEqual(self.nWindows, self.coordinator.nSlaves)
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))        
        self.assertEquals(self.nTasks, len(self.coordinator.taskList))
        self.assertEquals(self.nWindows, len(self.coordinator.currentlyRunning))
        self.assertEquals(1, len(self.coordinator.functionList))
        self.assertTrue(self.coordinator.startAhead == 0)
        self.coordinator.killSlaves()
        
        # nSlaves not specified (set to nTasks)
        self.coordinator = WindowCoordinator(self.nWindows, 0, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.assertEqual(0, self.coordinator.nSlaves)
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))        
        self.assertEquals(self.nTasks, len(self.coordinator.taskList))
        self.assertEquals(self.nTasks, len(self.coordinator.currentlyRunning))
        self.assertEquals(1, len(self.coordinator.functionList))
        self.assertTrue(self.coordinator.startAhead > 0)
        self.coordinator.killSlaves()
        
        # No visualization function
        self.coordinator = WindowCoordinator(self.nWindows, self.nSlaves, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)        
        self.coordinator.startSlaves(None, range(self.nTasks))        
        self.assertEquals(self.nTasks, len(self.coordinator.taskList))
        self.assertEquals(self.nSlaves, len(self.coordinator.currentlyRunning))
        self.assertEquals(0, len(self.coordinator.functionList))
        self.assertTrue(self.coordinator.startAhead > 0)
        self.coordinator.killSlaves()
        
        #
        # Errors
        #
        
        # Too few tasks
        self.coordinator = WindowCoordinator(self.nWindows, self.nSlaves, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.assertRaises(CoordinatorError, self.coordinator.startSlaves, visualizeSet, range(self.nSlaves - 1))
        self.assertEquals(0, len(self.coordinator.taskList))
        self.assertEquals(0, len(self.coordinator.currentlyRunning))
        self.assertEquals(0, len(self.coordinator.functionList))
        
        # No tasks
        self.coordinator = WindowCoordinator(self.nWindows, self.nSlaves, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.assertRaises(CoordinatorError, self.coordinator.startSlaves, None, None)
        self.assertEquals(0, len(self.coordinator.taskList))
        self.assertEquals(0, len(self.coordinator.currentlyRunning))
        self.assertEquals(0, len(self.coordinator.functionList))
        
        # Invalid function
        self.coordinator = WindowCoordinator(self.nWindows, self.nSlaves, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.assertRaises(CoordinatorError, self.coordinator.startSlaves, [], range(self.nTasks))
        self.assertEquals(0, len(self.coordinator.taskList))
        self.assertEquals(0, len(self.coordinator.currentlyRunning))
        self.assertEquals(0, len(self.coordinator.functionList))
        
        # Slaves already running
        self.coordinator = WindowCoordinator(self.nWindows, self.nSlaves, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))        
        self.assertRaises(CoordinatorError, self.coordinator.startSlaves, visualizeSet, range(self.nTasks))        
        self.coordinator.killSlaves()
        
    def testVisualize(self):        
        # Start slaves will have an implicit call to visualize
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i, stateObjects[i]['screen'][5][5])
        
        # Do second visualization
        self.coordinator.visualize(visualizeAdd, range(self.nTasks))
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i * 2, stateObjects[i]['screen'][5][5])
        
        # Third
        self.coordinator.visualize(visualizeAdd, range(self.nTasks))
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i * 3, stateObjects[i]['screen'][5][5])
        
        # Fourth
        self.coordinator.visualize(visualizeSub, range(self.nTasks))
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i * 2, stateObjects[i]['screen'][5][5])
            
        # Close the window of a few slaves then call visualize
        closed = []
        for t in self.coordinator.currentlyRunning:
            if t.taskID % 7 == 0:
                self.coordinator.sendCloseMsg(t.pipe)
                closed.append(t.taskID)
        self.coordinator.visualize(visualizeSub, range(self.nTasks))
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):
            if i in closed:
                continue
            self.assertEquals(i, stateObjects[i]['screen'][5][5])
            
        # Invalid visualize function
        self.assertRaises(CoordinatorError, self.coordinator.visualize, [], range(self.nTasks))
        
        # Slaves not started
        self.coordinator.killSlaves()
        self.assertRaises(CoordinatorError, self.coordinator.visualize, visualizeSub, range(self.nTasks))
        
        # Invalid tasks arg (None)
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i, stateObjects[i]['screen'][5][5])
        self.assertRaises(CoordinatorError, self.coordinator.visualize, visualizeSub, None)
        
        # Invalid task arg (too few elements)
        self.assertRaises(CoordinatorError, self.coordinator.visualize, visualizeSub, range(self.nTasks - 3))
        
        # Cleanup
        self.coordinator.killSlaves()
        
    # Also test visbileFilterOfF()
    def testFilterVisible(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        
        # First filter
        nv = self.coordinator.filterVisible(filter1, [True for i in range(self.nTasks)])
        self.assertEquals(self.nTasks, nv)
        self.assertEquals(self.nTasks, self.countNonFilteredTasks()) 
        
        # Second filter
        nv = self.coordinator.filterVisible(filter1, [True for i in range(self.nTasks)])
        self.assertEquals(self.nTasks, nv)
        self.assertEquals(self.nTasks, self.countNonFilteredTasks())
        
        # Third
        nv = self.coordinator.filterVisible(filter1, [False for i in range(self.nTasks)])
        self.assertEquals(0, nv)
        self.assertEquals(0, self.countNonFilteredTasks())
        
        # Clear filter
        self.coordinator.filterVisibleOff()
        self.assertEquals(self.nTasks, self.countNonFilteredTasks())
                
        self.coordinator.visualize(visualizeSet, range(self.nTasks))
        nv = self.coordinator.filterVisible(filter2, None)
        nExpected = (self.nTasks / 2) + (self.nTasks % 2)
        self.assertEquals(nExpected, nv)
        self.assertEquals(nExpected, self.countNonFilteredTasks())
        
        # Clear filter
        self.coordinator.filterVisibleOff()
        self.assertEquals(self.nTasks, self.countNonFilteredTasks())
        
        # Invalid filter function
        self.assertRaises(CoordinatorError, self.coordinator.filterVisible, [], None)
                
        # Invalid tasks (too many)
        self.assertRaises(CoordinatorError, self.coordinator.filterVisible, filter1, [True for i in range(self.nTasks * 2)])
        
        # Slaves not started
        self.coordinator.killSlaves()
        self.assertRaises(CoordinatorError, self.coordinator.filterVisible, filter1, [True for i in range(self.nTasks)])
        
        # Cleanup: not necessary since slaves ahve already been killed
        # self.coordinator.killSlaves()
    
    def countNonFilteredTasks(self):
        """
        Helper functino to cound the number of visible tasks
        """
        cnt = 0
        for t in self.coordinator.taskList:
            if t.isVisible or t.isShowable():
                cnt += 1
        return cnt 
    
    def countFilterVisibleTasks(self):
        '''
        Helper function to count the number of tasks with visible filter set to True
        '''
        cnt = 0
        for t in self.coordinator.taskList:
            if t.visibleFilter == True:
                cnt += 1
        return cnt
    
    def testPendingFilterVisible(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        
        # First filter (all True)
        self.coordinator.pendingFilterVisible(filter1, [True for i in range(self.nTasks)])
        self.assertEquals(self.nTasks, self.countNonFilteredTasks())        
        for i in range(self.nTasks / self.nWindows): # cycle through all tasks
            self.coordinator.showNext(self.nWindows) 
        self.assertEquals(self.nTasks, self.countFilterVisibleTasks())
        
        # Second filter (all False)
        self.coordinator.pendingFilterVisible(filter1, [False for i in range(self.nTasks)])        
        self.assertEquals(self.nTasks - self.nSlaves, self.countNonFilteredTasks())
        # A single call is enough since there are no windows that can be shown
        # but we do two for testing purposes
        self.coordinator.showNext(self.nWindows) 
        self.coordinator.showNext(self.nWindows)
        self.assertEquals(0, self.countFilterVisibleTasks())   
        
        # Third filter (all True)
        self.coordinator.pendingFilterVisible(filter1, [True for i in range(self.nTasks)])        
        self.assertEquals(self.nSlaves, self.countFilterVisibleTasks())
        for i in range(self.nTasks / self.nWindows): # cycle through all tasks
            self.coordinator.showNext(self.nWindows) 
        self.assertEquals(self.nTasks, self.countFilterVisibleTasks())
                
        # Clear filter
        self.coordinator.filterVisibleOff()
        self.assertEquals(self.nTasks, self.countNonFilteredTasks())
                
        # Fourth filter (even are True)                
        self.coordinator.visualize(visualizeSet, range(self.nTasks))
        # Estimate number of visbleFilter=True after the filter
        nExpected = 0
        for t in self.coordinator.currentlyRunning:
            if t.taskID % 2 == 0:
                nExpected += 1        
        self.coordinator.pendingFilterVisible(filter2, None)        
        nCounted = self.countFilterVisibleTasks()
        # Hard to know exact number of tasks since it depends on which tasks are currently running
        self.assertEquals(nExpected, nCounted)
        for i in range(self.nTasks / self.nWindows): # cycle through all tasks
            self.coordinator.showNext(self.nWindows)
        nExpected = (self.nTasks / 2) + (self.nTasks % 2)     
        self.assertEquals(nExpected, self.countFilterVisibleTasks())
        
        # Clear filter
        self.coordinator.filterVisibleOff()
        self.assertEquals(self.nTasks, self.countNonFilteredTasks())
        
        # Invalid filter function
        self.assertRaises(CoordinatorError, self.coordinator.pendingFilterVisible, [], None)
                
        # Invalid tasks (too many)
        self.assertRaises(CoordinatorError, self.coordinator.pendingFilterVisible, filter1, [True for i in range(self.nTasks * 2)])
        
        # Slaves not started
        self.coordinator.killSlaves()
        self.assertRaises(CoordinatorError, self.coordinator.pendingFilterVisible, filter1, [True for i in range(self.nTasks)])
        
        # Cleanup: not necessary since slaves have already been killed
        # self.coordinator.killSlaves()
        
    
    def testFilterKill(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        
        # First filter (all True)
        nv = self.coordinator.filterKill(filter1, [True for i in range(self.nTasks)])
        self.assertEquals(self.nTasks, nv)
        self.assertEquals(self.nTasks, self.countNonFilteredTasks()) 
        
        # Second filter (all True)
        nv = self.coordinator.filterKill(filter1, [True for i in range(self.nTasks)])
        self.assertEquals(self.nTasks, nv)
        self.assertEquals(self.nTasks, self.countNonFilteredTasks())
        
        # Third (all false)
        nv = self.coordinator.filterKill(filter1, [False for i in range(self.nTasks)])
        self.assertEquals(0, nv)
        self.assertEquals(0, self.countNonFilteredTasks())
        
        # Error: no more running slave, but no exception
        nv = self.coordinator.filterKill(filter1, [True for i in range(self.nTasks)])
        self.assertEquals(0, nv)
        self.assertEquals(0, self.countNonFilteredTasks())
        
        # Restart slaves
        self.coordinator.killSlaves()
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        
        # Fourth (kill odd numbered)
        nv = self.coordinator.filterKill(filter2, None)
        nExpected = (self.nTasks / 2) + (self.nTasks % 2)
        self.assertEquals(nExpected, nv)
        self.assertEquals(nExpected, self.countNonFilteredTasks())
                
        # Invalid filter function
        self.assertRaises(CoordinatorError, self.coordinator.filterKill, [], [True for i in range(self.nTasks)])
                
        # Invalid tasks (too many)
        self.assertRaises(CoordinatorError, self.coordinator.filterKill, filter1, [True for i in range(self.nTasks)])
        
        # Slaves not started
        self.coordinator.killSlaves()
        self.assertRaises(CoordinatorError, self.coordinator.filterKill, filter1, [True for i in range(self.nTasks)])
        
        # Cleanup: not necessary since slaves ahve already been killed
        # self.coordinator.killSlaves()
    
    def testGather(self):
        # Start slaves will have an implicit call to visualize
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i, stateObjects[i]['screen'][5][5])
            
        # Redo gather
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i, stateObjects[i]['screen'][5][5])
            
        # Fewer slaves
        self.coordinator.killSlaves()
        self.coordinator = WindowCoordinator(self.nWindows, self.nWindows, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)        
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i, stateObjects[i]['screen'][5][5])
        self.coordinator.showRandom()
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i, stateObjects[i]['screen'][5][5])
        
        
        # All slaves running
        self.coordinator.killSlaves()
        self.coordinator = WindowCoordinator(self.nWindows, self.nTasks, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        stateObjects = self.getStateObjects()
        for i in range(self.nTasks):                    
            self.assertEquals(i, stateObjects[i]['screen'][5][5])
        
        # Invalid gather function
        self.assertRaises(CoordinatorError, self.coordinator.gather, [], None)
                
        # Invalid tasks (too few)
        self.assertRaises(CoordinatorError, self.coordinator.gather, gatherStateObject, [True])
        
        # Slaves not started
        self.coordinator.killSlaves()
        self.assertRaises(CoordinatorError, self.coordinator.gather, gatherStateObject, None)
    
        # Cleanup: slaves already killed
    
    def testReadLogFiles(self):
        
        # Invalid slave logger (console is default
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        logLines = self.coordinator.readLogFiles()
        self.assertEqual(self.nTasks, len(logLines))
        for lines in logLines:
            self.assertIsNone(lines)
        self.coordinator.killSlaves()        
        
        # Setup file logger
        self.coordinator = WindowCoordinator(self.nWindows, self.nSlaves, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.coordinator.slaveLoggerSetup(setupFileLogger, logging.INFO, {'dir': self.tmpLogDir})        
        
        # Cause at least one info line to be logged
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        # Cause "show frame\n" to be logged on each of the currently running slaves
        self.coordinator.filterVisible(filter1, [True for i in range(self.nTasks)])
        
        logLines = self.coordinator.readLogFiles()
        self.assertEqual(self.nTasks, len(logLines))
        for lines in logLines:        
            self.assertTrue(len(lines) > 0)
            lineFound = False
            for l in lines:
                if l == "Filter: True\n":
                    lineFound = True  
                    break          
            self.assertTrue(lineFound)
        
        # Cleanup    
        self.coordinator.killSlaves()
    
    def testShowNext(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))

        si = 0
        for i in range(self.nTasks): # cycle through windows at least a few times        
            self.assertVisible(range(si, si + self.nWindows))
            self.coordinator.showNext()
            si += self.nWindows
        
        for i in range(self.nTasks): # cycle through windows at least a few times, one window at a time        
            self.assertVisible(range(si, si + self.nWindows))
            self.coordinator.showNext(1)
            si += 1
            
        # Too many too show
        self.assertRaises(CoordinatorError, self.coordinator.showNext, 6)
        
        # Cleanup
        self.coordinator.killSlaves()
        
    def assertVisible(self, visibleIDs):
        '''
        Helper function to check which tasks are visible
        
        @param visibleIDs: tasks that should be visible. All other tasks should be hidden
        '''
        # The tasks ids can be modulo something
        for i in range(len(visibleIDs)):
            visibleIDs[i] = visibleIDs[i] % self.nTasks
                    
        for t in self.coordinator.taskList:
            if t.taskID in visibleIDs:
                self.assertTrue(t.isVisible)
            else:
                self.assertFalse(t.isVisible)
                
    def assertShowable(self, showableIDs):
        '''
        Helper function to check which tasks are showable (running, not-visible, and non-filtered)
        
        @param visibleIDs: tasks that should be showable. All other tasks should be hidden
        '''
        # The tasks ids can be modulo something
        for i in range(len(showableIDs)):
            showableIDs[i] = showableIDs % len(self.nTasks)
                    
        for t in self.coordinator.taskList:
            if t.taskID in showableIDs:
                self.assertTrue(t.isShowable())
            else:
                self.assertFalse(t.isShowable())
                
    def assertNVisible(self, n):
        '''
        Helper function to check the number of tasks that are visible
            
        @param n: number of tasks that should be visible    
        '''                 
        nv = 0   
        for t in self.coordinator.taskList:
            if t.isVisible:
                nv += 1
        
        self.assertEqual(n, nv)
        
    def testShowPrevious(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))

        
        si = 0
        self.assertVisible(range(si + self.nWindows))
        for i in range(self.nTasks): # cycle through windows at least a few times                    
            self.coordinator.showPrevious()
            self.assertVisible(range(si - self.nWindows, si, 1))
            si -= self.nWindows
                 
        # Too many too show
        self.assertRaises(CoordinatorError, self.coordinator.showPrevious, 6)
                          
        # Cleanup
        self.coordinator.killSlaves()
    
    def testShowRandom(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))

        for i in range(self.nTasks): # cycle through windows at least a few times                                
            self.coordinator.showRandom()
            self.assertNVisible(self.nWindows)         
        
        for i in range(self.nTasks): # cycle through windows at least a few times, one window at a time        
            self.coordinator.showRandom(1)
            self.assertNVisible(self.nWindows)
            
        # Too many too show
        self.assertRaises(CoordinatorError, self.coordinator.showRandom, 6)
        
        # Cleanup
        self.coordinator.killSlaves()
    
    def testShowIndex(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))

        si = 0
        for i in range(self.nTasks): # cycle through windows at least a few times        
            self.coordinator.showIndex(si)
            self.assertVisible(range(si, si + self.nWindows))            
            si = (si + self.nWindows) % self.nTasks
        
        for i in range(self.nTasks): # cycle through windows at least a few times, one window at a time
            self.coordinator.showIndex(si)        
            self.assertVisible(range(si, si + self.nWindows))            
            si = (si + 7) % self.nTasks
            
        # Index too small
        self.assertRaises(CoordinatorError, self.coordinator.showIndex, -6)
        # Index too large
        self.assertRaises(CoordinatorError, self.coordinator.showIndex, 600)
        
        # Cleanup
        self.coordinator.killSlaves()
    
    def testShowIndexes(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))

        import random
        for i in range(self.nTasks): # cycle through windows at least a few times
            vids = []
            for j in range(self.nWindows):
                r = random.randint(0, self.nTasks - 1)
                if r not in vids:
                    vids.append(r)                                            
            self.coordinator.showIndexes(vids)
            self.assertVisible(vids)         
        
        for i in range(self.nTasks): # cycle through windows at least a few times, two windows at a time       
            vids = []
            for j in range(2):
                ti = random.randint(0, self.nTasks - 1)
                if ti not in vids:
                    vids.append(ti)                                            
            self.coordinator.showIndexes(vids)
            self.assertVisible(vids)
            
        # Too many indexes
        self.assertRaises(CoordinatorError, self.coordinator.showIndexes, range(12))
        # One index is too small
        self.assertRaises(CoordinatorError, self.coordinator.showIndexes, [-6, 0, 6])
        # One index is too large
        self.assertRaises(CoordinatorError, self.coordinator.showIndexes, [0, 6, 60])
        
        # Cleanup
        self.coordinator.killSlaves()
    
    def testStartNext(self):
        # Assumptions for the below tests
        self.assertEqual(32, self.nTasks)
        self.assertEqual(15, self.nSlaves)
        self.assertEqual(3, self.nWindows)
        
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        self.assertEquals(9, self.coordinator.startAhead)
        self.assertRunning(range(-3, 12))        
                        
        self.coordinator.showNext() # Implicit call to show next, 3-6
        self.assertRunning(range(0, 15))
        
        self.coordinator.showNext() # Implicit call to show next, 6-9
        self.assertRunning(range(3, 18))
        self.coordinator.showNext() # Implicit call to show next, 9-12
        self.assertRunning(range(6, 21))
        self.coordinator.showNext() # Implicit call to show next, 12-15
        self.assertRunning(range(9, 24))
        self.coordinator.showNext() # Implicit call to show next, 15-18
        self.assertRunning(range(12, 27))
        self.coordinator.showNext() # Implicit call to show next, 18-21
        self.assertRunning(range(15, 30))
        self.coordinator.showNext() # Implicit call to show next, 21-24
        self.assertRunning(range(18, 33))
        self.coordinator.showNext() # Implicit call to show next, 24-27
        self.assertRunning(range(21, 36))
        self.coordinator.showNext() # Implicit call to show next, 27-30
        self.assertRunning(range(24, 39))
        self.coordinator.showNext() # Implicit call to show next, 30, 31, 0
        self.assertRunning(range(27, 42))
        self.coordinator.showNext() # Implicit call to show next, 1-4
        self.assertRunning(range(30, 45))
                                           
        #
        # Note! very hard to write test case for the below. Step-through in debug mode
        # and see if it makes sense
        #
        print self.getCids()
        self.coordinator.showIndexes([0, 10, 20])
        print self.getCids()
        self.coordinator.showNext()
        print self.getCids()
        self.coordinator.showNext()        
        print self.getCids()
        self.coordinator.showIndex(12)        
        print self.getCids()
        self.coordinator.showNext()        
        print self.getCids()        
        
        #
        # Only a few additional slaves
        #
        self.coordinator.killSlaves()
        self.coordinator = WindowCoordinator(3, 6, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        self.assertEquals(3, self.coordinator.startAhead)
        self.assertRunning(range(0, 6))        
                        
        self.coordinator.showNext() # Implicit call to show next, 3-6
        self.assertRunning(range(3, 9))
        
        self.coordinator.showNext() # Implicit call to show next, 6-9
        self.assertRunning(range(6, 12))
        self.coordinator.showNext() # Implicit call to show next, 9-12
        self.assertRunning(range(9, 15))
        self.coordinator.showNext() # Implicit call to show next, 12-15
        self.assertRunning(range(12, 18))
        self.coordinator.showNext() # Implicit call to show next, 15-18
        self.assertRunning(range(15, 21))
        self.coordinator.showNext() # Implicit call to show next, 18-21
        self.assertRunning(range(18, 24))
                
        #
        # Note! very hard to write test case for the below. Step-through in debug mode
        # and see if it makes sense
        #
        print self.getCids()
        self.coordinator.showIndexes([0, 10, 20])
        print self.getCids()
        self.coordinator.showNext() # 1, 11, 21
        print self.getCids()
        self.coordinator.showNext() # 2, 12, 22
        print self.getCids()
        self.coordinator.showIndex(12) # 12, 13, 14
        print self.getCids()
        self.coordinator.showNext() # 15, 16, 17
        print self.getCids()
        
        #
        # No extra slaves
        #
        self.coordinator.killSlaves()
        self.coordinator = WindowCoordinator(3, 3, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        self.assertEquals(0, self.coordinator.startAhead)
        self.assertRunning(range(0, 3))        
                        
        self.coordinator.showNext() # Implicit call to show next, 3-6
        self.assertRunning(range(3, 6))
        
        self.coordinator.showNext() # Implicit call to show next, 6-9
        self.assertRunning(range(6, 9))
        self.coordinator.showNext() # Implicit call to show next, 9-12
        self.assertRunning(range(9, 12))
        self.coordinator.showNext() # Implicit call to show next, 12-15
        self.assertRunning(range(12, 15))
        self.coordinator.showNext() # Implicit call to show next, 15-18
        self.assertRunning(range(15, 18))
        self.coordinator.showNext() # Implicit call to show next, 18-21
        self.assertRunning(range(18, 21))
                        
        self.coordinator.showIndexes([0, 10, 20])
        self.assertRunning([0, 10, 20])        
        self.coordinator.showNext()
        self.assertRunning([1, 11, 21])        
        self.coordinator.showNext()        
        self.assertRunning([2, 12, 22])
        self.coordinator.showIndex(12)        
        self.assertRunning([12, 13, 14])
        self.coordinator.showNext()        
        self.assertRunning([15, 16, 17])
        
        # Cleanup
        self.coordinator.killSlaves()
    
    def getCids(self):
        '''
        Get current running task ids
        '''
        cids = [] 
        for t in self.coordinator.currentlyRunning:
            cids.append(t.taskID)
        return cids
    
    def assertRunning(self, taskIDs):
        '''
        Helper function to check that all, and only the, tasks in taskIDs are currently running
        '''
        for i in range(len(taskIDs)):
            taskIDs[i] = taskIDs[i] % self.nTasks       
        
        for t in self.coordinator.taskList:
            tid = t.taskID
            if tid in taskIDs:
                self.assertTrue(t in self.coordinator.currentlyRunning, '%d' % (t.taskID))
            else:
                self.assertFalse(t in self.coordinator.currentlyRunning, '%d' % (t.taskID))
        
    
    def testStartPrev(self):
        # Assumptions for the below tests
        self.assertEqual(32, self.nTasks)
        self.assertEqual(15, self.nSlaves)
        self.assertEqual(3, self.nWindows)
        
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        self.assertEquals(9, self.coordinator.startAhead)
        self.assertRunning(range(-3, 12))        
        
        #
        # Note! very hard to write test case for the below. Step-through in debug mode
        # and see if it makes sense
        #                
        self.coordinator.showPrevious() # Implicit call to show next, 29-32
        print self.getCids()        
        self.coordinator.showPrevious() # Implicit call to show next, 26-29
        print self.getCids()
        self.coordinator.showPrevious() # Implicit call to show next, 23-26
        print self.getCids()
        self.coordinator.showPrevious() # Implicit call to show next, 20-23
        print self.getCids()
        self.coordinator.showPrevious() # Implicit call to show next, 17-20
        print self.getCids()
        
        self.coordinator.showNext() # Implicit call to show next, 20-23
        print self.getCids()
        self.coordinator.showPrevious() # Implicit call to show next, 17-20
        print self.getCids()
        self.coordinator.showPrevious() # Implicit call to show next, 14-17
        print self.getCids()
        self.coordinator.showPrevious() # Implicit call to show next, 11-14
        print self.getCids()
        self.coordinator.showPrevious() # Implicit call to show next, 7-11
        print self.getCids()
        self.coordinator.showPrevious() # Implicit call to show next, 4-7
        self.coordinator.showPrevious() # Implicit call to show next, 1-4
        self.coordinator.showPrevious() # Implicit call to show next, 30, 31, 1
        print self.getCids()
                           
        self.coordinator.showIndexes([0, 10, 20]) # 0, 10, 20
        print self.getCids()
        self.coordinator.showPrevious() # 31, 9, 19
        print self.getCids()
        self.coordinator.showPrevious() # 30, 8, 18
        print self.getCids()

        self.coordinator.showIndex(12)  # 12, 13, 14
        print self.getCids()
        self.coordinator.showPrevious() # 9, 10, 11
        print self.getCids()        
        
        # Cleanup
        self.coordinator.killSlaves()
    
    # Not tested
    #def testResetVisible(self):
    
    def testShowMore(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nTasks))
        
        # All are visible so this should not do anything
        self.coordinator.showMore()
        self.assertVisible(range(self.nWindows))
        
        # Set all task to non showable
        self.coordinator.filterVisible(filter1, [False for i in range(self.nTasks)])
        self.assertEqual(0, len(self.coordinator.currentlyVisible))
        self.coordinator.showMore()
        self.assertEqual(0, len(self.coordinator.currentlyVisible))
        
        # Set all to visible again
        self.coordinator.filterVisibleOff()
        self.assertEqual(self.nTasks, self.countNonFilteredTasks())
        
        # Kill one of the currently visible tasks
        self.coordinator.killSlave(self.coordinator.currentlyVisible[0])
        self.assertEquals(self.nWindows - 1, len(self.coordinator.currentlyVisible))
        self.coordinator.showMore()
        self.assertEquals(self.nWindows, len(self.coordinator.currentlyVisible))
        self.assertEquals(self.nSlaves - 1, len(self.coordinator.currentlyRunning))        
        
        # Set all currently running to non-showable, and kill all currently visible task
        for t in self.coordinator.currentlyRunning:
            t.visibleFilter = False            
        for t in self.coordinator.currentlyVisible[:]:
            self.coordinator.killSlave(t)
        self.assertEquals(0, len(self.coordinator.currentlyVisible))
        self.coordinator.showMore()
        self.assertEquals(self.nWindows, len(self.coordinator.currentlyVisible))
        self.assertEquals(self.nSlaves - 1, len(self.coordinator.currentlyRunning))        
        
        # Cleanup
        self.coordinator.killSlaves()
        
    def testStartKillSlave(self):
        
        self.assertEqual(0, len(self.coordinator.currentlyRunning))        
        task = BealjasTask(312)
        
        self.coordinator.startSlave(task)
        self.assertEqual(1, len(self.coordinator.currentlyRunning))
        self.coordinator.sendSetPosSizeMsg(task.pipe, (False, 1, 2, 3, 4))
        m, x, y, w, h = self.coordinator.sendGetPosSizeMsg(task.pipe)
        self.assertEqual(1, x)
        self.assertEqual(2, y)
        self.coordinator.killSlave(task)
        
        self.assertEqual(0, len(self.coordinator.currentlyRunning))
        
        # 2 slaves
        task = BealjasTask(312)
        task2 = BealjasTask(31278)
        self.coordinator.startSlave(task)
        self.coordinator.startSlave(task2)
        self.assertEqual(2, len(self.coordinator.currentlyRunning))
        self.coordinator.sendSetPosSizeMsg(task.pipe, (False, 1, 2, 3, 4))
        self.coordinator.sendSetPosSizeMsg(task2.pipe, (False, 10, 20, 30, 40))
        m, x, y, w, h = self.coordinator.sendGetPosSizeMsg(task.pipe)
        self.assertEqual(1, x)
        m, x, y, w, h = self.coordinator.sendGetPosSizeMsg(task2.pipe)
        self.assertEqual(20, y)                
        self.coordinator.killSlave(task2)
        self.assertEqual(1, len(self.coordinator.currentlyRunning))
        m, x, y, w, h = self.coordinator.sendGetPosSizeMsg(task.pipe)
        self.assertEqual(1, x)
    
        # Note! task is still running
        
        # Task is already running
        self.assertRaises(CoordinatorError, self.coordinator.startSlave, task)
                
        # Task is not running
        self.assertRaises(CoordinatorError, self.coordinator.killSlave, task2)
        
        # Cleanuo
        self.coordinator.killSlave(task)
        self.assertEqual(0, len(self.coordinator.currentlyRunning))        
    
    def testSelectSlaveToKill(self):
        tasks = []
        for i in range(5):
            tasks.append(BealjasTask(i))
        
        # Start all tasks and set all to visible
        for t in tasks:
            self.coordinator.startSlave(t)
            self.coordinator.showTask(t)
            self.coordinator.sendSetPosSizeMsg(t.pipe, (False, i, i, i, i))
            
        # All slaves are visible, the second task should be selected
        self.assertEqual(tasks[1], self.coordinator.selectSlaveToKill(tasks[0]))
        
        # The non visible slave should be selected
        self.coordinator.hideTask(tasks[3])
        self.assertEqual(tasks[3], self.coordinator.selectSlaveToKill(tasks[0]))
        
        # The filtered slave should be selected
        tasks[4].visibleFilter = False
        self.assertEqual(tasks[4], self.coordinator.selectSlaveToKill(tasks[0]))
            
        # The first running slave should be selected
        self.coordinator.showTask(tasks[3])
        tasks[4].visibleFilter = True
        self.assertEqual(tasks[1], self.coordinator.selectSlaveToKill(tasks[0]))
        
        # No slaves to kill: the first running slave should be selected
        self.coordinator.killSlave(tasks[1])
        self.coordinator.killSlave(tasks[2])
        self.coordinator.killSlave(tasks[3])
        self.coordinator.killSlave(tasks[4])        
        self.assertRaises(CoordinatorError, self.coordinator.selectSlaveToKill, tasks[0])
        
        # Cleanup
        self.coordinator.killSlaves()
    
    def testShowHideTask(self):
        task = BealjasTask(312)
        task.showTimestamp = 3
        
        # Show        
        self.assertEqual(0, len(self.coordinator.currentlyRunning))        
        self.assertEqual(0, len(self.coordinator.currentlyVisible))
        self.assertEqual(True, self.coordinator.showTask(task))
        self.assertTrue(task.isVisible)
        self.assertEqual(1, len(self.coordinator.currentlyRunning))
        self.assertEqual(1, len(self.coordinator.currentlyVisible))
        self.assertEqual(task, self.coordinator.currentlyRunning[0])
        self.assertEqual(task, self.coordinator.currentlyVisible[0])
        self.assertTrue(task.showTimestamp > 3)

        # Hide
        self.assertEqual(True, self.coordinator.hideTask(task))
        self.assertFalse(task.isVisible)
        self.assertEqual(1, len(self.coordinator.currentlyRunning))
        self.assertEqual(0, len(self.coordinator.currentlyVisible))
        
        # Show again
        oldTs = task.showTimestamp
        self.assertEqual(True, self.coordinator.showTask(task))
        self.assertTrue(task.isVisible)
        self.assertEqual(1, len(self.coordinator.currentlyRunning))
        self.assertEqual(1, len(self.coordinator.currentlyVisible))
        self.assertEqual(task, self.coordinator.currentlyRunning[0])
        self.assertEqual(task, self.coordinator.currentlyVisible[0])
        self.assertTrue(task.showTimestamp > oldTs)
        
        # Set pos and hide
        self.coordinator.sendSetPosSizeMsg(task.pipe, (False, 1, 2, 3, 4))
        self.assertEqual(True, self.coordinator.hideTask(task))
        self.assertFalse(task.isVisible)
        self.assertEqual(1, len(self.coordinator.currentlyRunning))
        self.assertEqual(0, len(self.coordinator.currentlyVisible))
        
        # Show and verify pos
        self.assertEqual(True, self.coordinator.showTask(task))
        pos = self.coordinator.sendGetPosSizeMsg(task.pipe)
        self.assertEquals(2, pos[2])
        
        # Show visible task
        self.assertRaises(CoordinatorError, self.coordinator.showTask, task)
        
        # Show with pending visible filter
        self.coordinator.killSlave(task)
        task.pendingVisibleFilter = (filter1, False)
        self.assertEqual(False, self.coordinator.showTask(task))
        
        # Hide killed task
        self.coordinator.killSlaves()
        #self.assertRaises(CoordinatorError, self.coordinator.showTask, task)
        self.assertRaises(CoordinatorError, self.coordinator.hideTask, task)
        
        # Show filtered task
        self.coordinator.startSlave(task)
        task.visibleFilter = False
        self.assertFalse(self.coordinator.showTask(task))
        
        # Already hidden
        task.visibleFilter = True
        self.assertEqual(True, self.coordinator.showTask(task))
        self.assertEqual(True, self.coordinator.hideTask(task))
        self.assertRaises(CoordinatorError, self.coordinator.hideTask, task)
        
        # Not running
        self.coordinator.killSlave(task)                
    
    # Tests selectNextTaskToShow, selectPrevTaskToShow and selectRandomTaskToShow
    def testSelectTaskToShow(self):
        self.coordinator = WindowCoordinator(6, 6, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.coordinator.startSlaves(visualizeSet, range(6))
        
        tasks = self.coordinator.currentlyRunning[:]
        
        for t in self.coordinator.currentlyVisible[:]:
            self.coordinator.hideTask(t)
                        
        # next
        t2s = self.coordinator.selectNextTaskToShow(0, [])
        self.assertEquals(1, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(0, [tasks[1], tasks[2]])
        self.assertEquals(3, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(4, [tasks[1], tasks[2], tasks[5]])
        self.assertEquals(0, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(4, [tasks[0], tasks[1], tasks[2], tasks[3], tasks[5]])
        self.assertEquals(4, t2s.taskID)
        self.assertIsNone(self.coordinator.selectNextTaskToShow(3, [tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5]]))
                
        # prev
        t2s = self.coordinator.selectPrevTaskToShow(0, [])
        self.assertEquals(5, t2s.taskID)
        t2s = self.coordinator.selectPrevTaskToShow(2, [tasks[1], tasks[2]])
        self.assertEquals(0, t2s.taskID)
        t2s = self.coordinator.selectPrevTaskToShow(4, [tasks[1], tasks[2], tasks[5]])
        self.assertEquals(3, t2s.taskID)
        self.assertIsNone(self.coordinator.selectPrevTaskToShow(3, [tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5]]))
        
        # random
        t2s = self.coordinator.selectRandomTaskToShow(3, [tasks[1], tasks[2], tasks[3], tasks[4], tasks[5]])
        self.assertEquals(0, t2s.taskID)
        self.assertIsNone(self.coordinator.selectRandomTaskToShow(3, [tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5]]))
        
        # 1 and 3 are visible, 5 is filtered
        self.coordinator.showTask(self.coordinator.currentlyRunning[1])
        self.coordinator.showTask(self.coordinator.currentlyRunning[3])
        self.coordinator.currentlyRunning[5].visibleFilter = False
        self.coordinator.currentlyRunning[0].visibleFilter = None
        self.coordinator.currentlyRunning[2].visibleFilter = True
        self.coordinator.currentlyRunning[4].visibleFilter = True
        
        # Next
        t2s = self.coordinator.selectNextTaskToShow(0, [])
        self.assertEquals(2, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(0, [tasks[1], tasks[2]])
        self.assertEquals(4, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(4, [tasks[1], tasks[2], tasks[5]])
        self.assertEquals(0, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(2, [tasks[0], tasks[1], tasks[2], tasks[3], tasks[5]])
        self.assertEquals(4, t2s.taskID)
        self.assertIsNone(self.coordinator.selectNextTaskToShow(3, [tasks[0], tasks[2], tasks[4]]))
        
        # Prev
        t2s = self.coordinator.selectPrevTaskToShow(0, [])
        self.assertEquals(4, t2s.taskID)
        t2s = self.coordinator.selectPrevTaskToShow(2, [tasks[1], tasks[2]])
        self.assertEquals(0, t2s.taskID)
        t2s = self.coordinator.selectPrevTaskToShow(4, [tasks[1], tasks[2], tasks[5]])
        self.assertEquals(0, t2s.taskID)
        self.assertIsNone(self.coordinator.selectPrevTaskToShow(3, [tasks[0], tasks[2], tasks[4]]))
        
        # Random
        t2s = self.coordinator.selectRandomTaskToShow(3, [tasks[4]])
        self.assertTrue(t2s.taskID == 0 or t2s.taskID == 2)
        self.assertIsNone(self.coordinator.selectRandomTaskToShow(3, [tasks[0], tasks[2], tasks[4]]))
        
        # Next
        t2s = self.coordinator.selectNextTaskToShow(0, [], True)
        self.assertEquals(1, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(0, [tasks[1], tasks[2]], True)
        self.assertEquals(3, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(4, [tasks[1], tasks[2]], True)
        self.assertEquals(0, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(2, [tasks[0], tasks[1], tasks[2], tasks[3], tasks[5]], True)
        self.assertEquals(4, t2s.taskID)
        t2s = self.coordinator.selectNextTaskToShow(3, [tasks[0], tasks[2], tasks[4]], True)
        self.assertEquals(1, t2s.taskID)
        
        # Prev
        #t2s = self.coordinator.selectPrevTaskToShow(0, [], True)
        #self.assertEquals(4, t2s.taskID)
        #t2s = self.coordinator.selectPrevTaskToShow(2, [tasks[1], tasks[2]], True)
        #self.assertEquals(0, t2s.taskID)
        #t2s = self.coordinator.selectPrevTaskToShow(4, [tasks[1], tasks[2], tasks[5]], True)
        #self.assertEquals(3, t2s.taskID)
        #t2s = self.coordinator.selectPrevTaskToShow(3, [tasks[0], tasks[2], tasks[4]], True)
        #self.assertEquals(1, t2s.taskID)
        
        # Random
        #t2s = self.coordinator.selectRandomTaskToShow(3, [tasks[4]], True)
        #self.assertTrue(t2s.taskID != 4 and t2s.taskID != 5)
        #t2s = self.coordinator.selectRandomTaskToShow(3, [tasks[0], tasks[2], tasks[4]], True)
        #self.assertTrue(t2s.taskID == 1 or t2s.taskID == 3)
        
        # Cleanup
        self.coordinator.killSlaves()
            
    def testShowCommon(self):
        self.coordinator = WindowCoordinator(6, 8, 
                                             createWindow, guiLoop, closeWindow, setWindowVisible, getWindowPosSize, setWindowPosSize,
                                             self.debugWindow)
        self.coordinator.startSlaves(visualizeSet, range(10))
        
        tasks = self.coordinator.currentlyRunning[:]
        for t in self.coordinator.currentlyVisible[:]:
            self.coordinator.hideTask(t)
        
        self.coordinator.showCommon([], [tasks[1], tasks[3]])
        self.assertEquals(2, len(self.coordinator.currentlyVisible))
        self.assertEquals(tasks[1], self.coordinator.currentlyVisible[0])
        
        self.coordinator.showCommon([tasks[1]], [tasks[2], tasks[4]])
        self.assertEquals(3, len(self.coordinator.currentlyVisible))
        self.assertEquals(tasks[4], self.coordinator.currentlyVisible[2])
        
        self.coordinator.showCommon([tasks[2], tasks[3]], [])
        self.assertEquals(1, len(self.coordinator.currentlyVisible))
        
        self.coordinator.showCommon([], [tasks[2]])
        self.assertEquals(2, len(self.coordinator.currentlyVisible))
        
        # Accepted that a task in both toHide and toShow
        self.coordinator.showCommon([tasks[1]], [tasks[5], tasks[1]])
        self.assertEquals(4, len(self.coordinator.currentlyVisible))
        
        # Visible: 1, 2, 4 and 5
        
        # Duplicates in hide
        self.assertRaises(CoordinatorError, self.coordinator.showCommon, [tasks[1], tasks[2], tasks[1]], [tasks[5]])
        
        # Duplicates in show
        self.assertRaises(CoordinatorError, self.coordinator.showCommon, [tasks[1]], [tasks[5], tasks[5]])
                                                                                    
        # Hide non-visible task
        self.assertRaises(CoordinatorError, self.coordinator.showCommon, [tasks[1], tasks[3], tasks[5]], [])
        
        # Show visible task
        self.assertRaises(CoordinatorError, self.coordinator.showCommon, [tasks[1]], [tasks[3], tasks[5]])
        
        self.coordinator.killSlaves()
    
    #def testStartCommon(self):
        # This function is tested in testStartNext and testStartPrev
        
        # Test invalid argument: nothing happens
        #self.assertRaises(CoordinatorError, self.coordinator.startCommon, -3, self.coordinator.selectNextTaskToShow)        
    
    def testGetLeastRecentlyShownTasks(self):
        self.coordinator.startSlaves(visualizeSet, range(self.nSlaves))
        
        tasks = self.coordinator.currentlyRunning[:]
        reverseTasks = tasks[:]
        reverseTasks.reverse()
        
        # Set showTimestamps in the same order as the taskIDs 
        ts = 313
        for t in reverseTasks:
            t.showTimestamp = ts
            ts = ts + 4
            
        lrst = self.coordinator.getLeastRecentlyShownTasks(1)
        self.assertEquals(tasks[-1], lrst[0])
        
        lrst = self.coordinator.getLeastRecentlyShownTasks(3)
        self.assertEquals(tasks[-1], lrst[0])
        self.assertEquals(tasks[-2], lrst[1])
        self.assertEquals(tasks[-3], lrst[2])
            
        lrst = self.coordinator.getLeastRecentlyShownTasks(self.nSlaves)
        self.assertEquals(self.nSlaves - self.nWindows, len(lrst))
        for i in range(len(lrst)):
            self.assertEquals(lrst[i], reverseTasks[i])
        
        # Will just return len(tasks)
        lrst = self.coordinator.getLeastRecentlyShownTasks(self.nSlaves * 2)
        self.assertEquals(self.nSlaves - self.nWindows, len(lrst))
        
        self.coordinator.killSlaves()
    
    def testExecuteVisibleFilterTask(self):
        task = BealjasTask(322)
        
        # Task is not running
        self.assertRaises(CoordinatorError, self.coordinator.executeVisibleFilterTask, task)
        
        # This function is also tested by testPendingFilterVisible since that function calls this function
    
    def testExecuteFunctionAll(self):
        self.assertRaises(CoordinatorError, self.coordinator.executeFunctionAll, 'foobar', gather1, False)
        
        # This function is also tested by testFilterVisible and testGather since those functions call this function
    

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()