'''
Created on Jun 25, 2012

@author: larsab
'''
import unittest

from multiprocessing import Process, Pipe
# Entry point for slave process
from bealjas.plotter.windowPlotter import plotterEntry
# Debug window functions
from bealjas.utils.bealjas2debug import createWindow, guiLoop, closeWindow, setWindowPosSize, getWindowPosSize, setWindowVisible
import logging
from bealjas.utils.setupLogging import setupFileLogger, setupConsoleLogger
from bealjas.network.serializeFunction import serialize
import tempfile

def visualizeSet(stateObject, logger, args):
    '''
    @param: stateObject: the debug screen object
    @param logger: a logger
    @param args: an integer to set the screen elements to
    '''
    logger.info("Set screen elements")
    for i in range(len(stateObject['screen'])):
        for j in range(len(stateObject['screen'][i])):
            stateObject['screen'][i][j] = args
            
def visualizeAdd(stateObject, logger, args):
    '''
    @param: stateObject: the debug screen object
    @param logger: a logger
    @param args: an integer to add to the screen elements
    '''
    logger.info("Add %d to screen elements" % (args))
    for i in range(len(stateObject['screen'])):
        for j in range(len(stateObject['screen'][i])):
            stateObject['screen'][i][j] = stateObject['screen'][i][j] + args
            
def visualizeSub(stateObject, logger, args):
    '''
    @param: stateObject: the debug screen object
    @param logger: a logger
    @param args: an integer to substract from the screen elements
    '''
    logger.info("Subtract %d from screen elements" % (args))
    for i in range(len(stateObject['screen'])):
        for j in range(len(stateObject['screen'][i])):
            stateObject['screen'][i][j] = stateObject['screen'][i][j] - args
            
def gather1(stateObject, logger, args=None):
    '''
    Return the args value   
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: an integer returned by this function
    @return: an object that is sent to the coordinator. It is also possible to return None
    '''
     
    logger.info("Gather arg: %d" % (args))
    return args

def gather2(stateObject, logger, args=None):
    '''
    Return an element in the screen matrix   
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: ignored
    @return: an object that is sent to the coordinator. It is also possible to return None
    '''
     
    rv = stateObject['screen'][2][3]
    logger.info("Gather screen element: %d" % (rv))
    return rv

def gatherStateObject(stateObject, logger, args=None):
    '''
    Return the stateObject  
    
    @param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param args: ignored
    @return: the state Object
    '''    

    logger.info("Gather state object")
    return stateObject

def filter1(stateObject, logger, args=None):
    '''
    Return the boolean in args
    
    param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param taskArg: True or False
    @return: args
    '''
    logger.info("Filter: %s" % (args))
    return args

def filter2(stateObject, logger, args=None):
    '''
    Return True if the screen element values are even and False if they are not (all elements have the
    same value)
    
    param stateObject: plotter state that is sent as argument to all application provided functions
    @param logger: plotter logger
    @param taskArg: True or False
    @return: args
    '''    
    v = stateObject['screen'][2][3]
    logger.info("Filter: is %d even" % (v))
    return (v % 2 == 0)
     
class Test(unittest.TestCase):


    def setUp(self):
        # Tmp dir for logfiles
        self.tmpLogDir = tempfile.mkdtemp()
        
        # This dict holds the debug window state
        self.debugWindow = {}
        
        #
        # The windowPlotter to test is run as a seperate process
        #        
        # Create pipe used to send command to slave and receive return values
        self.pipe, childEnd = Pipe()
        # Fork of a new process using Pythons multiprocessing module
        self.p = Process(target=plotterEntry, args=('unit-plot',
                                                    childEnd, 
                                                    setupFileLogger,
                                                    logging.INFO,
                                                    createWindow,
                                                    guiLoop,       
                                                    closeWindow,                                            
                                                    setWindowVisible,
                                                    getWindowPosSize,
                                                    setWindowPosSize,
                                                    {'dir': self.tmpLogDir},
                                                    self.debugWindow))
        self.p.start()
        # Some tests may explicitely kill the child
        self.childKilled = False

    def tearDown(self):
        if self.childKilled == False:
            self.pipe.send('kill')
            self.p.join()
            
    # Helper function that returns a state object
    def getStateObject(self):
        self.pipe.send("gather")
        self.pipe.send((serialize(gatherStateObject), None))
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.pipe.recv()
        return stateObject

    # First test to run is kill since it is used during tearDown
    def testKill(self):
        self.childKilled = True # Prevent another kill in tearDown
        self.pipe.send('kill')
        self.assertEquals('processing', self.pipe.recv())
        self.p.join()

    def testConstructor(self):
        stateObject = self.getStateObject()         
        self.assertTrue(stateObject.has_key('screen'))
        self.assertFalse(stateObject['isVisible'])
        
    def testVisualize(self):
        # Do first visualization
        self.pipe.send("visualize")
        self.pipe.send((serialize(visualizeSet), 0))
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject() 
        self.assertEquals(0, stateObject['screen'][5][5])
        
        # Do second visualization
        self.pipe.send("visualize")
        self.pipe.send((serialize(visualizeAdd), 3))
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject() 
        self.assertEquals(3, stateObject['screen'][5][5])
        
        # Third
        self.pipe.send("visualize")
        self.pipe.send((serialize(visualizeAdd), 12))
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject() 
        self.assertEquals(15, stateObject['screen'][5][5])
        
        # Fourth
        self.pipe.send("visualize")
        self.pipe.send((serialize(visualizeSub), 10))
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject()
        self.assertEquals(5, stateObject['screen'][5][5])
        
    def testGather(self):
        # Do first gather
        self.pipe.send("gather")
        self.pipe.send((serialize(gather1), 12))
        self.assertEquals('processing', self.pipe.recv())
        rv = self.pipe.recv()
        self.assertEqual(12, rv)
        
        # Do some visualizations
        self.testVisualize()
        
        # Do second gather
        self.pipe.send("gather")
        self.pipe.send((serialize(gather1), 13))
        self.assertEquals('processing', self.pipe.recv())
        rv = self.pipe.recv()
        self.assertEqual(13, rv)
        
        # Do third gather
        self.pipe.send("gather")
        self.pipe.send((serialize(gather2), None))
        self.assertEquals('processing', self.pipe.recv())
        rv = self.pipe.recv()
        self.assertEqual(5, rv)        
    
    def testFilterF(self):
        self.pipe.send("filter")
        self.pipe.send((serialize(filter1), True))
        self.assertEquals('processing', self.pipe.recv())
        self.assertTrue(self.pipe.recv())
        
        self.pipe.send("filter")
        self.pipe.send((serialize(filter1), False))
        self.assertEquals('processing', self.pipe.recv())
        self.assertFalse(self.pipe.recv())
        
        self.pipe.send("visualize")
        self.pipe.send((serialize(visualizeSet), 0))
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject() 
        self.assertEquals(0, stateObject['screen'][5][5])
        
        self.pipe.send("filter")
        self.pipe.send((serialize(filter2), None))
        self.assertEquals('processing', self.pipe.recv())
        self.assertTrue(self.pipe.recv())
        
        self.pipe.send("visualize")
        self.pipe.send((serialize(visualizeAdd), 3))
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject() 
        self.assertEquals(3, stateObject['screen'][5][5])
        
        self.pipe.send("filter")
        self.pipe.send((serialize(filter2), None))
        self.assertEquals('processing', self.pipe.recv())
        self.assertFalse(self.pipe.recv())
    
    # Test for hide, show, getPosSize, and setPosSize
    def testHideShowPos(self):        
        # This does not actually do anything
        self.pipe.send("visualize")
        self.pipe.send((serialize(visualizeSet), 3))
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject() 
        self.assertEquals(3, stateObject['screen'][5][5])
        
        # Show and hide
        self.pipe.send("show")
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject()        
        self.assertTrue(stateObject['isVisible'])
        
        self.pipe.send("show")
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject() 
        self.assertTrue(stateObject['isVisible'])
        
        self.pipe.send("hide")
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject() 
        self.assertFalse(stateObject['isVisible'])
        
        self.pipe.send("hide")
        self.assertEquals('processing', self.pipe.recv())
        stateObject = self.getStateObject() 
        self.assertFalse(stateObject['isVisible'])
        
        # Set and get pos
        self.pipe.send("setPosSize")
        self.pipe.send((False, 10, 20, 30, 40))
        self.assertEquals('processing', self.pipe.recv())
        self.pipe.send("getPosSize")
        self.assertEquals('processing', self.pipe.recv())
        m, x, y, w, h = self.pipe.recv()
        self.assertEquals(False, m)
        self.assertEquals(10, x)
        self.assertEquals(20, y)
        self.assertEquals(30, w)
        self.assertEquals(40, h)
        
        # Maximize
        self.pipe.send("setPosSize")
        self.pipe.send((True, 10, 20, 30, 40))
        self.assertEquals('processing', self.pipe.recv())
        self.pipe.send("getPosSize")
        self.assertEquals('processing', self.pipe.recv())
        m, x, y, w, h = self.pipe.recv()
        self.assertEquals(True, m)
        
        # Set and get pos (2)
        self.pipe.send("setPosSize")
        self.pipe.send((False, 11, 21, 31, 41))
        self.assertEquals('processing', self.pipe.recv())
        self.pipe.send("getPosSize")
        self.assertEquals('processing', self.pipe.recv())
        m, x, y, w, h = self.pipe.recv()
        self.assertEquals(False, m)
        self.assertEquals(11, x)
        self.assertEquals(21, y)
        self.assertEquals(31, w)
        self.assertEquals(41, h)
    
    def testReadLogfileLines(self):
        # Cause an info line to be logged
        self.pipe.send("show")
        self.assertEquals('processing', self.pipe.recv())
        
        self.pipe.send("readLogfileLines")        
        self.assertEquals('processing', self.pipe.recv())
        lines = self.pipe.recv()
        self.assertTrue(len(lines) > 0)
        lineFound = False
        for l in lines:
            if l == "Show frame\n":
                lineFound = True            
        self.assertTrue(lineFound)

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()